/* 
 * popen_secure, a variant of popen that uses 
 * execl_secure to transition to a different SID.
 *
 * Based on iopopen.c from glibc, which has the following copyright
 * and license:
 */

/* Copyright (C) 1993,1997,1998,1999,2000,2001 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Written by Per Bothner <bothner@cygnus.com>.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.

   As a special exception, if you link the code in this file with
   files compiled with a GNU compiler to produce an executable,
   that does not cause the resulting executable to be covered by
   the GNU Lesser General Public License.  This exception does not
   however invalidate any other reasons why the executable file
   might be covered by the GNU Lesser General Public License.
   This exception applies to code released by its copyright holders
   in files containing the exception.  */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/wait.h>
#include <proc_secure.h>

struct proc_file
{
	FILE *fp; /* cannot embed the FILE object, unlike ioopen.c */
	pid_t pid;
	struct proc_file *next;
};

static struct proc_file *proc_file_chain = NULL;

FILE *popen_secure(const char *command, const char *mode, security_id_t sid)
{
	int parent_end, child_end;
	int pipe_fds[2];
	pid_t child_pid;
	FILE *fp;
	struct proc_file *newp;

	newp = malloc(sizeof(struct proc_file));
	if (!newp) 
		return NULL;

	if (pipe(pipe_fds) < 0) {
		free(newp);
		return NULL;
	}

	if (mode[0] == 'r' && mode[1] == '\0') {
		parent_end = pipe_fds[0];
		child_end = pipe_fds[1];
	} else if (mode[0] == 'w' && mode[1] == '\0') {
		parent_end = pipe_fds[1];
		child_end = pipe_fds[0];
	} else {
		close(pipe_fds[0]);
		close(pipe_fds[1]);
		free(newp);
		errno = -EINVAL;
		return NULL;
	}

	child_pid = fork();
	if (child_pid == 0) {
		int child_std_end = mode[0] == 'r' ? 1 : 0;
		struct proc_file *p;

		close(parent_end);
		if (child_end != child_std_end)	{
			dup2(child_end, child_std_end);
			close(child_end);
		}
		/* POSIX.2:  "popen() shall ensure that any streams from previous
		   popen() calls that remain open in the parent process are closed
		   in the new child process." */
		for (p = proc_file_chain; p; p = p->next)
			fclose(p->fp);

		execl_secure ("/bin/sh", sid, "sh", "-c", command, (char *) 0);
		exit (127);
	}
	close (child_end);
	if (child_pid < 0)
	  {
	    close (parent_end);
	    free(newp);
	    return NULL;
	  }

	fp = fdopen(parent_end, mode);
	if (!fp) {
		close(parent_end);
		free(newp);
		return NULL;
	}

	newp->fp = fp;
	newp->pid = child_pid;

	newp->next = proc_file_chain;
	proc_file_chain = newp;
	return fp;
}

int pclose_secure(FILE *fp)
{
	int wstatus;
	struct proc_file *prev, *cur;
	pid_t pid, wait_pid;

	for (prev = NULL, cur = proc_file_chain; cur; prev = cur, cur = cur->next) {
		if (cur->fp == fp) 
			break;
	}

	if (!cur) 
		return -1;
	
	if (prev) {
		prev->next = cur->next;
	} else {
		proc_file_chain = cur->next;
	}
	
	pid = cur->pid;
	free(cur);
	
	if (fclose(fp) < 0)
		return -1;

	/* POSIX.2 Rationale:  "Some historical implementations either block
	   or ignore the signals SIGINT, SIGQUIT, and SIGHUP while waiting
	   for the child process to terminate.  Since this behavior is not
	   described in POSIX.2, such implementations are not conforming." */
	do {
		wait_pid = waitpid(pid, &wstatus, 0);
	} while (wait_pid == -1 && errno == EINTR);
	if (wait_pid == -1)
		return -1;
	return wstatus;
}
