/* Copyright (C) 2003 Tresys Technology, LLC
 * see file 'COPYING' for use and warranty information */

/* 
 * Author: mayerf@tresys.com 
 */

/* analysis.c
 *
 * Analysis routines for libapol
 */
#include <stdlib.h>
#include <assert.h>
#include "policy.h"
#include "util.h"
#include "analysis.h"


/* all the "free" fns below have a prototype just like free() so that
 * ll_free() in util.c can use them.  This makes us have to cast the
 * pointer, which can also cause run-time errors since someone could
 * mistakenly pass the wrong data type!  BE CAREFUL!.
 */
void free_entrypoint_type(void *t)
{
	entrypoint_type_t *p = (entrypoint_type_t *)t;
	if(p == NULL)
		return;
	if(p->ep_rules != NULL) 
		free(p->ep_rules);
	if(p->ex_rules != NULL) 
		free(p->ex_rules);
	free(p);
	return;
}

void free_tgt_trans_domain(void *t)
{
	tgt_trans_domain_t *p = (tgt_trans_domain_t *)t;
	if(p == NULL)
		return;
	ll_free(p->entry_types, free_entrypoint_type);
	if(p->pt_rules != NULL) 
		free(p->pt_rules);
	free(p);
	return;
}

void free_domain_trans_analysis(domain_trans_analysis_t *p)
{
	if(p == NULL)
		return;
	ll_free(p->tgt_domains, free_tgt_trans_domain);
	free(p);
	return;
}

entrypoint_type_t *new_entry_point_type(void)
{
	entrypoint_type_t *t;
	t = (entrypoint_type_t *)malloc(sizeof(entrypoint_type_t));
	if(t == NULL) {
		fprintf(stderr, "out of memory");
		return NULL;
	}
	memset(t, 0, sizeof(entrypoint_type_t));
	return t;
}

tgt_trans_domain_t *new_tgt_trans_domain(void)
{
	tgt_trans_domain_t *t;
	t = (tgt_trans_domain_t *)malloc(sizeof(tgt_trans_domain_t));
	if(t == NULL) {
		fprintf(stderr, "out of memory");
		return NULL;
	}
	memset(t, 0, sizeof(tgt_trans_domain_t));
	t->entry_types = ll_new();
	return t;
}

domain_trans_analysis_t *new_domain_trans_analysis(void)
{
	domain_trans_analysis_t *t;
	t = (domain_trans_analysis_t *)malloc(sizeof(domain_trans_analysis_t));
	if(t == NULL) {
		fprintf(stderr, "out of memory");
		return NULL;
	}
	memset(t, 0, sizeof(domain_trans_analysis_t));
	t->tgt_domains = ll_new();

	return t;
}


/* INTERNAL */
static int dta_add_rule_to_tgt_type(int src_idx, int tgt_idx, int rule_idx, 
		domain_trans_analysis_t *dta)
{	
	llist_node_t *t;
	tgt_trans_domain_t *t_data = NULL;
	/* 1. find the type in the dta->tgt_domains list */
	/*TODO: Need to fix the list; right now unsorted so this will can become painful*/
	for(t = dta->tgt_domains->head; t != NULL; t = t->next) {
		t_data = (tgt_trans_domain_t *) t->data;
		assert(t_data->src_type == src_idx);
		if(t_data->tgt_type == tgt_idx)
			break;
	}
	if(t == NULL)
		return -1; /* tgt_idx doesn't currently exist in the dta! */
	assert(t_data != NULL);
	
	/* 2. add the rule to pt_rules list for that tgt type */
	return add_i_to_a(rule_idx ,&(t_data->num_pt_rules), &(t_data->pt_rules));
}

/* INTERNAL */
static int dta_add_tgt_type(int src_idx, int tgt_idx, int rule_idx, 
		domain_trans_analysis_t *dta)
{
	tgt_trans_domain_t *t;
	
	/* allocate and initialize new target type struct (we may undo this later) */
	t = new_tgt_trans_domain();
	if(t == NULL) 
		return -1;
	t->src_type = src_idx;
	t->tgt_type = tgt_idx;

	/* add the rule to the new target type */
	if(add_i_to_a(rule_idx ,&(t->num_pt_rules), &(t->pt_rules)) != 0) {
		free_tgt_trans_domain(t);
		return -1;
	}
	/* and link the target into the dta struct */
	/* TODO: need to do an insertion sort */
	if(ll_append_data(dta->tgt_domains, t) != 0 ) {
		free_tgt_trans_domain(t);
		return -1;
	}
			
	return 0;
}

/* INTERNAL: add process trans allowed target types to dta result */
static int dta_add_process_trans_rule(int src_idx, int rule_idx, bool_t *b_tgt, domain_trans_analysis_t *dta, 
		policy_t *policy)
{
	int *types = NULL, num_types, rt, i, idx;
	assert(b_tgt != NULL && dta != NULL && policy != NULL && is_valid_av_rule_idx(rule_idx, 1, policy));
	rt = extract_types_from_te_rule(rule_idx, RULE_TE_ALLOW, TGT_LIST, &types, &num_types, policy);
	if(rt < 0)
		return -1;
	if(rt == 2) {
		/* all all types */
		/* NOTE: Start from i = 1 since we know that type index 0 is 'self' and
		 * 	we don't want to include the pdeudo type self
		 */
		for(i = 1; i < policy->num_types; i++) {
			if(!b_tgt[i]) {
				/* add new target type and record its rules */
				if(dta_add_tgt_type(src_idx, i, rule_idx, dta) != 0) 
					return -1;
				b_tgt[i] = TRUE;
			}
			else {
				/* type already added just added, include this pt rule */
				if(dta_add_rule_to_tgt_type(src_idx, i, rule_idx, dta) != 0)
					return -1;
			}
		}
	} 
	else {
		/* add types and rules returned in list to target domains list*/
		for(i = 0; i < num_types; i++) {
			/* NOTE: We have a special case if types[i] == 0.  This is the pseudo
			 *	type 'self'.  In this case we really don't want to add self, but
			 *	rather the src_idx.  So in that case we'll change the idx
			 * 	the src_idx.
			 */
			if(types[i] == 0)
				idx = src_idx;
			else
				idx = types[i];

			if(!b_tgt[idx]) {
				/* add new target type and record its rules */
				if(dta_add_tgt_type(src_idx, idx, rule_idx, dta) != 0) {
					if(types != NULL) free(types);
					return -1;
				}
				b_tgt[idx] = TRUE;
			}
			else {
				/* type already added just added, include this pt rule */
				if(dta_add_rule_to_tgt_type(src_idx, idx, rule_idx, dta) != 0) {
					if(types != NULL) free(types);
					return -1;
				}
			}
		}
		if(types != NULL) free(types);
	}
	
	return 0;
}



/* INTERNAL */
static int dta_add_file_execute_rule(int rule_idx, entrypoint_type_t *ep)
{
	if(ep != NULL) 
		return add_i_to_a(rule_idx, &(ep->num_ex_rules), &(ep->ex_rules));
	else 
		return -1;
}

/* INTERNAL */
static int dta_add_rule_to_ep_file_type(int file_idx, int rule_idx, tgt_trans_domain_t *tgt)
{	
	llist_node_t *t;
	entrypoint_type_t *t_data = NULL;
	/* 1. find the file type in the tgt */
	/*TODO: Need to fix the list; right now unsorted so this will can become painful*/
	for(t = tgt->entry_types->head; t != NULL; t = t->next) {
		t_data = (entrypoint_type_t *) t->data;
		if(t_data->file_type == file_idx)
			break;
	}
	if(t == NULL)
		return -1; /* file_idx doesn't currently exist in the tgt! */
	assert(t_data != NULL);
	
	/* 2. add the rule  */
	return add_i_to_a(rule_idx ,&(t_data->num_ep_rules), &(t_data->ep_rules));
}

/* INTERNAL */
static int dta_add_ep_type(int file_idx, int rule_idx, tgt_trans_domain_t *tgt)
{
	entrypoint_type_t *t;
	
	/* allocate and initialize new target type struct (we may undo this later) */
	t = new_entry_point_type();
	if(t == NULL) 
		return -1;
	t->src_type = tgt->src_type;
	t->tgt_type = tgt->tgt_type;
	t->file_type = file_idx;

	/* add the rule to the new target type */
	if(add_i_to_a(rule_idx, &(t->num_ep_rules), &(t->ep_rules)) != 0) {
		free_entrypoint_type(t);
		return -1;
	}
	/* link in new file type */
	/* TODO: need to do an insertion sort */
	if(ll_append_data(tgt->entry_types, t) != 0 ) {
		free_entrypoint_type(t);
		return -1;
	}
			
	return 0;
}


/* INTERNAL */ 
/* TODO: This is very similar to dta_add_process_trans_rule(); should consolidate */
static int dta_add_file_entrypoint_rule(int rule_idx, bool_t *b_types, tgt_trans_domain_t *tgt, policy_t *policy)
{
	int rt, i, idx, *types, num_types; 
	assert(policy != NULL &&is_valid_av_rule_idx(rule_idx,1,policy) && b_types != NULL && tgt != NULL);
	
	/* get all the rule target types */
	rt = extract_types_from_te_rule(rule_idx, RULE_TE_ALLOW, TGT_LIST, &types, &num_types, policy);
	if(rt < 0)
		return -1;
	if(rt == 2) {
		/* all all types */
		/* NOTE: Start from i = 1 since we know that type index 0 is 'self' and
		 * 	we don't want to include the pdeudo type self */
		for(i = 1; i < policy->num_types; i++) {
			if(!b_types[i]) {
				/* new */
				if(dta_add_ep_type(i, rule_idx, tgt) != 0)
					return -1;
				b_types[i] = TRUE;
			}
			else {
				/* existing; add rule to existing one */
				if(dta_add_rule_to_ep_file_type(i, rule_idx, tgt) != 0)
					return -1;
			}
		}
	}
	else {
		/* adding new file type */
		/* add types and rules returned in list to target domains list*/
		for(i = 0; i < num_types; i++) {
			/* NOTE: We have a special case if types[i] == 0.  This is the pseudo
			 *	type 'self'.  In this case we really don't want to add self, but
			 *	rather the target's index (which is the source for these rules).
			 *	So in that case we'll change the idx the tgt->tgt_type.
			 */
			if(types[i] == 0)
				idx = tgt->tgt_type;
			else
				idx = types[i];	
			if(!b_types[idx]) {
				/* new */
				if(dta_add_ep_type(idx, rule_idx, tgt) != 0) {
					if(types != NULL) free(types);
					return -1;
				}
				b_types[idx] = TRUE;
			}
			else {
				/* existing; add rule to existing one */
				if(dta_add_rule_to_ep_file_type(idx, rule_idx, tgt) != 0) {
					if(types != NULL) free(types);
					return -1;
				}
			}
		}
		if(types != NULL) free(types);
	}				
				

	return 0;
}


/* main domain trans analysis function.
 * 	dta must be allocated and initialized
 *
 *	returns:	
 *		-1 general error
 *		-2 source invalid type
 */

int determine_domain_trans(char *source, domain_trans_analysis_t **dta, policy_t *policy)
{
	int src_idx, i, classes[1], perms[1], perms2[1], rt;
	rules_bool_t b_s, b_t;
	bool_t *b_tgt;	/* scratch arrays to keep track of types */
	tgt_trans_domain_t *tgt;
	entrypoint_type_t *ep;
	llist_node_t *ll_node, *ll_node2;
	
	if(policy == NULL || dta == NULL)
		return -1;
	if((src_idx = get_type_idx(source, policy)) < 0)
		return -2;
	*dta = NULL;
	
	/* initialize our bool rule structures...free before leaving function */
	b_tgt = (bool_t *)malloc(sizeof(bool_t) * policy->num_types);
	if(b_tgt == NULL) {
		fprintf(stderr, "out of memory");
		return -1;
	}
	memset(b_tgt, 0, policy->num_types * sizeof(bool_t));
	/* b_s (all rules that have src as SOURCE) & b_t (similar but used as tgt as SOURCE) */
	if(init_rules_bool(0, &b_s, policy) != 0) 
		goto err_return;
	if(init_rules_bool(0, &b_t, policy) != 0) 
		goto err_return;		
	
	/* initialize the results structure (caller must free if successful) */
	*dta = new_domain_trans_analysis();
	if(*dta == NULL) {
		fprintf(stderr, "out of memory");
		goto err_return;
	}
	(*dta)->src_type = src_idx;
	if((*dta)->tgt_domains == NULL)
		goto err_return;
		
	/* populate dta with candidate target domains by collecting
	 * all the rules that allow src_type process transition access, and 
	 * then selecting all the target types from those rules.
	 */
 
	/* 1. select all rules that allow have source as in its SOURCE list (keep this around
	 	we use it later when down-selecting candidate entry point file types ) */
	if(match_te_rules(0, NULL, 0, src_idx, IDX_TYPE, 0, SRC_LIST, 1, &b_s, policy) != 0)
		goto err_return;
	
	/* 2. extract the target domain types for process transition perm, and add to our result 
	      keeping track if type already added in b_tgt */
	classes[0] = get_obj_class_idx("process", policy);
	assert(classes[0] >= 0);
	perms[0] = get_perm_idx("transition", policy);
	assert(perms[0] >= 0);
	for(i = 0; i < policy->num_av_access; i++) {
		if(b_s.access[i] && (policy->av_access)[i].type == RULE_TE_ALLOW && 
				does_av_rule_use_classes(i, 1, classes, 1, policy) &&
				does_av_rule_use_perms(i, 1, perms, 1, policy)) {
			/* 2.a we have a rule that allows process tran access, add it for now */
			rt = dta_add_process_trans_rule(src_idx, i, b_tgt, *dta, policy);
			if(rt != 0)
				goto err_return;
		}
	}
	
	/* At this point, we have a list of all target types (and assoicated list of rules) that
	 * allow process transition permission to the src.  Now we need to take each tgt type,
	 * and look for file types that provide source execute and target entrypoint access.
	 */
	 
	/* 3. get all the file types for the candidate target types */
	
	/* set up some temporary structure for our searched */
	classes[0] = get_obj_class_idx("file", policy);
	assert(classes[0] >= 0);
	perms[0] = get_perm_idx("entrypoint", policy);
	perms2[0] = get_perm_idx("execute", policy);
	assert(perms[0] >= 0);
	assert(perms2[0] >= 0);
	
	for(ll_node = (*dta)->tgt_domains->head; ll_node != NULL; ) {
		tgt = (tgt_trans_domain_t *)ll_node->data;
		assert(tgt != NULL);
		all_false_rules_bool(&b_t, policy);
		memset(b_tgt, 0, policy->num_types * sizeof(bool_t));
		/* 3.a all rules that provide target access as SOURCE */
		if(match_te_rules(0, NULL, 0, tgt->tgt_type, IDX_TYPE, 0, SRC_LIST, 1, &b_t, policy) != 0)
			goto err_return;
			
		/* 3.b extract candidate files types from rules that allow target file entrypoint access */
		for(i = 0; i < policy->num_av_access; i++) {
			if(b_t.access[i] && (policy->av_access)[i].type == RULE_TE_ALLOW && 
					does_av_rule_use_classes(i, 1, classes, 1, policy) &&
					does_av_rule_use_perms(i, 1, perms, 1, policy)) {
				rt = dta_add_file_entrypoint_rule(i, b_tgt,tgt, policy);
				if(rt != 0)
					goto err_return;
			}
		}
		/* 3.c for each candidate file type, now look for rules that allow the source
		 *	execute access to the file type. */
		for(ll_node2 = tgt->entry_types->head; ll_node2 != NULL; ) {
			ep = (entrypoint_type_t *) ll_node2->data;
			assert(ep != NULL);
			for(i = 0; i < policy->num_av_access; i++) {
				/* to be of interest, rule must have source as src (b_s), be an allow
				 * rule, provide access to the current entrypoint file type, relate
				 * to file class objects, and provide execute permission. */
				if(b_s.access[i] && policy->av_access[i].type == RULE_TE_ALLOW && 
				  does_av_rule_idx_use_type(i, 0, ep->src_type, IDX_TYPE, SRC_LIST, TRUE, policy) &&
				  does_av_rule_idx_use_type(i, 0, ep->file_type, IDX_TYPE, TGT_LIST, TRUE, policy) &&
				  does_av_rule_use_classes(i, 1, classes, 1, policy) &&
				  does_av_rule_use_perms(i, 1, perms2, 1, policy)) {	
				rt = dta_add_file_execute_rule(i, ep);
				if(rt != 0)
					goto err_return;
				}		
			}
			/* 3.d at this point if a candidate file type does not have any execute rules
			 * 	than it fails all 3 criteria and we remove it from the target type. 
			 	We don't have to check for entrypoint rules, because the file type
			 	would not even be in the list if it didn't already have at least one
			 	entrypoint rule */
			if(ep->num_ex_rules < 1) {
				assert(ep->ex_rules == NULL);
				if(ll_unlink_node(tgt->entry_types, ll_node2) != 0) 
					goto err_return;
				ll_node2 = ll_node_free(ll_node2, free_entrypoint_type);
			}
			else {
				/* interate */
				ll_node2 = ll_node2->next;
			}
		}
		/* 3.e at this point, if a candidate target types does not have any file types,
		 *	remove it since it fails the criteria */
		if(tgt->entry_types->num < 1) {
			if(ll_unlink_node((*dta)->tgt_domains, ll_node) !=0)
				goto err_return;
			ll_node = ll_node_free(ll_node, free_tgt_trans_domain);
		}
		else {
			/* interate */
			ll_node = ll_node->next;
		}
		
	}
	

	if(b_tgt != NULL) free(b_tgt);
	free_rules_bool(&b_t);	
	free_rules_bool(&b_s);	
	return 0;	
err_return:	
	free_domain_trans_analysis(*dta);
	if(b_tgt != NULL) free(b_tgt);
	free_rules_bool(&b_t);	
	free_rules_bool(&b_s);
	return -1;
}


