/* Copyright (C) 2002-2003 Tresys Technology, LLC
 * see file 'COPYING' for use and warranty information */

/* 
 * Author: mayerf@tresys.com 
 */

/* apol_tcl.c
 *
 */
 
/* The tcl functions to support the GUI using TK */

#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#include <tk.h>
#include <assert.h>
#include <unistd.h>
#include <regex.h>
#include "policy.h"
#include "util.h"
#include "queue.h"
#include "apol_tcl.h"
#include "render.h"
#include "analysis.h"
#ifdef APOL_PERFORM_TEST
#include <time.h>
#endif

/* APOL stuff */
extern unsigned int policydb_lineno;
extern policy_t *policy; /*apolicy_parse.y, our policy DB */
extern queue_t id_queue;
extern FILE *yyin;
extern int yyparse(void);
extern void yyrestart(FILE *);
extern unsigned int pass;
extern char *rulenames[]; /* in render.c*/



/**************************************************************************
 * work functions
 **************************************************************************/
 
 /* some internal prototypes */
static int append_common_perm_str(bool_t do_perms, bool_t do_classes, bool_t newline, int idx, Tcl_DString *buf,
				policy_t *policy);
static int append_perm_str(bool_t do_common_perms, bool_t do_classes, bool_t newline, int idx, Tcl_DString *buf,
				policy_t *policy);
static int append_class_str(bool_t do_perms, bool_t do_cps, bool_t expand_cps, bool_t newline, int idx, Tcl_DString *buf,
				policy_t *policy);


 
/* We look for the TCL files in the following order:
 * 	1. If we find apol.tcl in the cur directory, we then assume
 *	   the TCL files are there, else
 * 	2. We look for the environment variable APOL_SCRIPT_DIR and if
 * 	   exists, look for apol.tcl there, else
 *	3. We then look for in APOL_INSTALL_DIR for apol.tcl.
 * Otherwise we report an installation error. 
 */
/* global used to keep track of the script directory, set by Apol_GetScriptDir */
static char *script_dir = NULL;
 
 
/* find the provided TCL script file according to the algorithm
 * described above.  This function returns a string of the directory.
 */
char* find_tcl_script(char *script_name)
{
	char *script, *var = NULL, *dir = NULL;
	int scriptsz;
	int rt;
	
	if(script_name == NULL)
		return NULL;
		
	/* 1. check current directory */
	scriptsz = strlen(script_name) + 4;
	script = (char *)malloc(scriptsz);
	if(script == NULL) {
		fprintf(stderr, "out of memory");
		return NULL;
	}	
	sprintf(script, "./%s", script_name);
	rt = access(script, R_OK);
	if(rt == 0) {
		dir = (char *)malloc(4);
		if(dir == NULL) {
			fprintf(stderr, "out of memory");
			return NULL;
		}
		sprintf(dir, "./");
		free(script);
		return dir;
	}
	free(script);
	
	/* 2. check environment variable */
	var = getenv(APOL_ENVIRON_VAR_NAME);
	if(!(var == NULL)) {
		scriptsz = strlen(var) + strlen(script_name) + 2;
		script = (char *)malloc(scriptsz);
		if(script == NULL) {
			fprintf(stderr, "out of memory");
			return NULL;
		}	
		sprintf(script, "%s/%s", var, script_name);	
		rt = access(script, R_OK);
		if(rt == 0) {
			dir = (char *)malloc(strlen(var) + 1);
			if(dir == NULL) {
				fprintf(stderr, "out of memory");
				return NULL;
			}
			sprintf(dir, var);
			free(script);
			return dir;		
		}
	}
	
	/* 3. installed directory */
	scriptsz = strlen(APOL_INSTALL_DIR) + strlen(script_name) + 2;
	script = (char *)malloc(scriptsz);
	if(script == NULL) {
		fprintf(stderr, "out of memory");
		return NULL;
	}	
	sprintf(script, "%s%s", APOL_INSTALL_DIR, script_name);
	rt = access(script, R_OK);
	if(rt == 0) {
		dir = (char *)malloc(strlen(APOL_INSTALL_DIR) +1);
		if(dir == NULL) {
			fprintf(stderr, "out of memory");
			return NULL;
		}
		sprintf(dir, APOL_INSTALL_DIR);
		free(script);
		return dir;	
	}
	
	/* 4. Didn't find it! */
	free(script);		
	return NULL;		
}

static int append_common_perm_str(bool_t do_perms, bool_t do_classes, bool_t newline, int idx, Tcl_DString *buf,
				policy_t *policy)
{
	int i;
	char tbuf[APOL_STR_SZ + 64];
	
	if(idx >= policy->num_common_perms|| buf == NULL) {
		return -1;
	}	
	Tcl_DStringAppend(buf, policy->common_perms[idx].name, -1);
	if(do_perms) {
		sprintf(tbuf, "   (%d permissions)\n", policy->common_perms[idx].num_perms);
		Tcl_DStringAppend(buf, tbuf, -1);
		for(i = 0; i < policy->common_perms[idx].num_perms; i++) {
			Tcl_DStringAppend(buf, "     ", -1);
			append_perm_str(0, 0, 1, policy->common_perms[idx].perms[i], buf, policy);
		}
	}
	/* determine which classes use this common perm */
	if(do_classes) {
		Tcl_DStringAppend(buf, "\n   Object classes that use this common permission\n", -1);
		for(i = 0; i < policy->num_obj_classes; i++) {
			if(does_class_use_common_perm(i, idx, policy)) {
				Tcl_DStringAppend(buf, "     ", -1);
				append_class_str(0,0,0,1,i,buf,policy);
			}
		}
	}
	
	if(newline)
		Tcl_DStringAppend(buf, "\n", -1);		
	return TCL_OK;
}

static int append_perm_str(bool_t do_common_perms, bool_t do_classes, bool_t newline, int idx, Tcl_DString *buf,
				policy_t *policy)
{
	int i;
	bool_t used;
	
	if(idx >= policy->num_perms|| buf == NULL) {
		return -1;
	}
	
	Tcl_DStringAppend(buf, policy->perms[idx], -1);
	if(do_classes || do_common_perms) {
		Tcl_DStringAppend(buf, "\n", -1);
	}
	/* find the classes that use this perm */
	if(do_classes) {
		used = FALSE;
		Tcl_DStringAppend(buf, "   object classes:\n", -1);
		for(i = 0; i < policy->num_obj_classes; i++) {
			if(does_class_use_perm(i, idx, policy)) {
				used = TRUE;
				Tcl_DStringAppend(buf, "        ", -1);
				append_class_str(0,0,0,1,i,buf,policy);
			}
			else if(does_class_indirectly_use_perm(i, idx, policy)) {
				used = TRUE;
				Tcl_DStringAppend(buf, "        ", -1);
				append_class_str(0,0,0,0,i,buf,policy);
				/* we "star" those included via common perm */
				Tcl_DStringAppend(buf, "*\n", -1);
			}
		}
		if(!used) {
			Tcl_DStringAppend(buf, "        <none>\n", -1);
		}
	}	
	/* find the common perms that use this perm */
	if(do_common_perms) {
		used = FALSE;
		Tcl_DStringAppend(buf, "   common permissions:\n", -1);
		for(i = 0; i < policy->num_common_perms; i++) {
			if(does_common_perm_use_perm(i, idx, policy)) {
				used = TRUE;
				Tcl_DStringAppend(buf, "        ", -1);
				append_common_perm_str(0,0,1,i,buf,policy);
			}
		}
		if(!used) {
			Tcl_DStringAppend(buf, "        <none>\n", -1);
		}
	}

	if(newline)
		Tcl_DStringAppend(buf, "\n", -1);		
	return TCL_OK;	
}


static int append_class_str(bool_t do_perms, bool_t do_cps, bool_t expand_cps, bool_t newline, int idx, Tcl_DString *buf,
				policy_t *policy)
{
	int i, cp_idx;
	
	if(idx >= policy->num_obj_classes|| buf == NULL) {
		return -1;
	}
	
	Tcl_DStringAppend(buf, policy->obj_classes[idx].name, -1);
	
	if(do_cps || do_perms) {
		Tcl_DStringAppend(buf, "\n", -1);
	}
	/* class-specific perms */
	if(do_perms) {
		for(i = 0; i < policy->obj_classes[idx].num_u_perms; i++) {
			assert(i < policy->num_perms);
			Tcl_DStringAppend(buf, "     ", -1);
			append_perm_str(0, 0, 1, policy->obj_classes[idx].u_perms[i], buf, policy);
		}
	}
	
	/* common perms */
	if(do_cps) {
		if(policy->obj_classes[idx].common_perms >= 0) {
			cp_idx = policy->obj_classes[idx].common_perms;
			Tcl_DStringAppend(buf, "     ", -1);
			append_common_perm_str(0, 0, 0, cp_idx, buf, policy);
			Tcl_DStringAppend(buf, "  (common perm)\n", -1);
			if(expand_cps) {
				for(i = 0; i < policy->common_perms[cp_idx].num_perms; i++) {
					assert(i < policy->num_perms);
					Tcl_DStringAppend(buf, "          ", -1);
					append_perm_str(0, 0, 1, i, buf, policy);
				}
			}
		}
	}
	
	if(newline)
		Tcl_DStringAppend(buf, "\n", -1);		
	return TCL_OK;
}


static int append_type_str(bool_t do_attribs, bool_t do_aliases, bool_t newline, int idx, 
	policy_t *policy, Tcl_DString *buf)
{
	int j;
	char tbuf[APOL_STR_SZ + 64];
	if(idx >= policy->num_types || buf == NULL) {
		return -1;
	}
	Tcl_DStringAppend(buf, policy->types[idx].name, -1);
	
	if(do_aliases) {
		if(policy->types[idx].aliases != NULL) {
			name_item_t *ptr;
			Tcl_DStringAppend(buf, " alias {", -1);
			for(ptr = policy->types[idx].aliases; ptr != NULL; ptr = ptr->next) {
				Tcl_DStringAppend(buf, ptr->name, -1);
				if(ptr->next != NULL)
					Tcl_DStringAppend(buf, " ", -1);
			}
			Tcl_DStringAppend(buf, "} ", -1);
		}
	}
	
	if(do_attribs) {
		sprintf(tbuf, " (%d attributes)\n", policy->types[idx].num_attribs);
		Tcl_DStringAppend(buf, tbuf, -1);
		for(j = 0; j < policy->types[idx].num_attribs; j++) {
			sprintf(tbuf, "\t%s\n", policy->attribs[policy->types[idx].attribs[j]].name);
			Tcl_DStringAppend(buf, tbuf, -1);
		}
	}
	if(newline)
		Tcl_DStringAppend(buf, "\n", -1);	

	return 0;
}

static int append_attrib_str(bool_t do_types, bool_t do_type_attribs, bool_t use_aliases, 
	bool_t newline, bool_t upper, int idx, policy_t *policy, Tcl_DString *buf)
{
	int j, k;
	char tbuf[APOL_STR_SZ+64];

	if(idx >= policy->num_attribs || buf == NULL) {
		return -1;
	}	

	if(upper) {
		Tcl_DStringAppend(buf, uppercase(policy->attribs[idx].name, tbuf), -1);
	}
	else
		Tcl_DStringAppend(buf, policy->attribs[idx].name, -1);
		
	if(do_types) {
		sprintf(tbuf, " (%d types)\n", policy->attribs[idx].num_types);
		Tcl_DStringAppend(buf, tbuf, -1);
		for(j = 0; j < policy->attribs[idx].num_types; j++) {
			Tcl_DStringAppend(buf, "\t", -1);
			Tcl_DStringAppend(buf, policy->types[policy->attribs[idx].types[j]].name, -1);
			/* aliases */
			if(use_aliases && policy->types[policy->attribs[idx].types[j]].aliases != NULL) {
				name_item_t *ptr;
				Tcl_DStringAppend(buf, ":", -1);
				for(ptr = policy->types[policy->attribs[idx].types[j]].aliases; ptr != NULL; ptr = ptr->next) {
					Tcl_DStringAppend(buf, " ", -1);
					Tcl_DStringAppend(buf, ptr->name, -1);
					if(ptr->next != NULL)
						Tcl_DStringAppend(buf, ",", -1);
				}
			}			
			if(do_type_attribs) {
				Tcl_DStringAppend(buf, " { ", -1);
				for(k = 0; k < policy->types[policy->attribs[idx].types[j]].num_attribs; k++) {
					if(strcasecmp(policy->attribs[idx].name, policy->attribs[policy->types[policy->attribs[idx].types[j]].attribs[k]].name) != 0)
						Tcl_DStringAppend(buf, policy->attribs[policy->types[policy->attribs[idx].types[j]].attribs[k]].name, -1);
						Tcl_DStringAppend(buf, " ", -1);
				}
				Tcl_DStringAppend(buf, "}", -1);
			}
			Tcl_DStringAppend(buf, "\n", -1);
		}
	}
	if(newline)
		Tcl_DStringAppend(buf, "\n", -1);		

	return 0;
}


/* searches using regular expressions */
static int append_all_ta_using_regex(regex_t *preg, const char *regexp, bool_t do_types, bool_t do_attribs, bool_t use_aliases,
			bool_t type_attribs, bool_t attrib_types, bool_t attrib_type_attribs, policy_t *policy,
			Tcl_DString *buf)
{
	int i, rt;
	char tbuf[APOL_STR_SZ+64];

	if(!(do_types || do_attribs)) {
		return 0; /* nothing to do */
	}
	if(buf == NULL) {
		return -1;
	}
	
	if(do_types) {
		Tcl_DStringAppend(buf, "Type ", -1);
		if(do_attribs) {
			Tcl_DStringAppend(buf, "and Type Attribute ", -1);
		}
	}
	else {
		Tcl_DStringAppend(buf, "Type Attribute ", -1);
	}
	sprintf(tbuf, "Search using regular expressions: \'%s\'\n", regexp);
	Tcl_DStringAppend(buf, tbuf, -1);
	
	if(do_types) {
		Tcl_DStringAppend(buf, "\n\nTYPES:\n", -1);
		for(i = 0; i < policy->num_types; i++) {
			rt = regexec(preg, policy->types[i].name, 0, NULL, 0);
			if(rt == 0) {
				rt = append_type_str(type_attribs, use_aliases, 1, i, policy, buf);
				if(rt != 0) {
					return -1;
				}
			}
			else if(use_aliases) {
				name_item_t *ptr;
				for(ptr = policy->types[i].aliases; ptr != NULL; ptr = ptr->next) {
					rt = regexec(preg, ptr->name, 0, NULL, 0);
					if(rt == 0) {
						rt = append_type_str(type_attribs, use_aliases, 1, i, policy, buf);
						if(rt != 0) {
							return -1;
						}
					}
				}
			}		
		}
	}
	
	if(do_attribs) {
		Tcl_DStringAppend(buf, "\n\nTYPE ATTRIBUTES:\n", -1);
		for(i = 0; i < policy->num_attribs; i++) {
			if(regexec(preg, policy->attribs[i].name, 0,NULL,0) == 0) {
				rt = append_attrib_str(attrib_types, attrib_type_attribs, use_aliases, 1, 0, i, policy, buf);
				if(rt != 0) {
					return -1;
				}

			}
									
		}
	}

	return 0;
}


static int append_tt_rule(bool_t addnl, bool_t addlineno, int idx, policy_t *policy, Tcl_DString *buf) 
{
	char *rule;
	
	if(buf == NULL) {
		return -1;
	}
	
	rule = re_render_tt_rule(addlineno, idx, policy);
	if(rule == NULL)
		return -1;
	Tcl_DStringAppend(buf, rule, -1);
	free(rule);
	
	if(addnl) {
		Tcl_DStringAppend(buf, "\n", -1);
	}
	return 0;	

}

static int append_clone_rule(bool_t addnl, bool_t addlineno, cln_item_t *item, policy_t *policy, Tcl_DString *buf) 
{
	char tbuf[APOL_STR_SZ+64];
	
	if(buf == NULL) {
		return -1;
	}
	
	Tcl_DStringAppend(buf, rulenames[RULE_CLONE], -1);
	Tcl_DStringAppend(buf, " ", -1);
	if(append_type_str(0,0, 0, item->src, policy, buf) != 0)
		return -1;	
	Tcl_DStringAppend(buf, " ", -1);		
	if(append_type_str(0,0, 0, item->tgt, policy, buf) != 0)
		return -1;
	Tcl_DStringAppend(buf, ";", -1);

	if(addlineno) {
		sprintf(tbuf, "       (%lu)", item->lineno);
		Tcl_DStringAppend(buf, tbuf, -1);
	}
	if(addnl) {
		Tcl_DStringAppend(buf, "\n", -1);
	}
	return 0;
}


static int append_av_rule(bool_t addnl, bool_t addlineno, int idx, bool_t is_au, policy_t *policy, Tcl_DString *buf)
{ 
	char *rule;
	
	if(buf == NULL) {
		return -1;
	}
	
	rule = re_render_av_rule(addlineno, idx, is_au, policy);
	if(rule == NULL)
		return -1;
	Tcl_DStringAppend(buf, rule, -1);
	free(rule);

	if(addnl) {
		Tcl_DStringAppend(buf, "\n", -1);
	}
	return 0;
}

static int append_role(int idx, bool_t name_only, int numperline, policy_t *policy, Tcl_DString *buf)
{
	int j;
	div_t x;	
	char tmpbuf[APOL_STR_SZ+64];
	if(numperline < 1)
		numperline = 1;
	
	Tcl_DStringAppend(buf, policy->roles[idx].name, -1);
	if(!name_only) {
		sprintf(tmpbuf, " (%d types)\n     ", policy->roles[idx].num_types);
		Tcl_DStringAppend(buf, tmpbuf, -1);
		for(j = 0; j < policy->roles[idx].num_types; j++) {
			/* control # of types per line */
			if(j != 0) {
				x = div(j, numperline);
				if(x.rem == 0) {
					sprintf(tmpbuf, "\n     ");
					Tcl_DStringAppend(buf, tmpbuf, -1);
				}
			}
			sprintf(tmpbuf, "%s  ", policy->types[policy->roles[idx].types[j]].name);
			Tcl_DStringAppend(buf, tmpbuf, -1);
		}
		Tcl_DStringAppend(buf, "\n", -1); /* extra line if we're exploding role types */	
	}
	Tcl_DStringAppend(buf, "\n", -1);
	return 0;	
}

static int append_user_str(user_item_t *user, bool_t name_only,  policy_t *policy, Tcl_DString *buf)
{
	char *name;
	ta_item_t *ptr;
	int rt;
		
	if(user == NULL || buf == NULL || policy == NULL)
		return -1;
	
	Tcl_DStringAppend(buf, user->name, -1);
	if(!name_only) {
		Tcl_DStringAppend(buf, " { ", -1);
		for(ptr = user->roles; ptr != NULL; ptr = ptr->next) {
			assert(ptr->type == IDX_ROLE);
			rt = get_role_name(ptr->idx, &name, policy);
			if(rt != 0) {
				return -1;
			}
			Tcl_DStringAppend(buf, name, -1);
			free(name);
			Tcl_DStringAppend(buf, " ", -1);
		}
		Tcl_DStringAppend(buf, "};", -1);
	}
	Tcl_DStringAppend(buf, "\n", -1);
	return 0;
}

static int append_role_allow_rule(role_allow_t *rule, policy_t *policy, Tcl_DString *buf)
{
	ta_item_t *tptr;	
	int multiple = 0;
	if(buf == NULL) {
		return -1;
	}
		
	Tcl_DStringAppend(buf, rulenames[RULE_ROLE_ALLOW], -1);

	/* source roles */
	if(rule->flags & AVFLAG_SRC_TILDA) 
		Tcl_DStringAppend(buf, " ~", -1);
	else
		Tcl_DStringAppend(buf, " ", -1);
	if(rule->src_roles != NULL && rule->src_roles->next != NULL) {
		multiple = 1;
		Tcl_DStringAppend(buf, "{", -1);
	}
	if(rule->flags & AVFLAG_SRC_STAR)
		Tcl_DStringAppend(buf, "*", -1);
			
	for(tptr = rule->src_roles; tptr != NULL; tptr = tptr->next) {
		assert(tptr->type == IDX_ROLE);
		Tcl_DStringAppend(buf, " ", -1);
		Tcl_DStringAppend(buf, policy->roles[tptr->idx].name, -1);
		Tcl_DStringAppend(buf, " ", -1);
	}
	if(multiple) {
		Tcl_DStringAppend(buf, "}", -1);
		multiple = 0;
	}
	/* target roles */
	if(rule->flags & AVFLAG_TGT_TILDA) 
		Tcl_DStringAppend(buf, " ~", -1);
	else
		Tcl_DStringAppend(buf, " ", -1);
	if(rule->tgt_roles != NULL && rule->tgt_roles->next != NULL) {
		multiple = 1;
		Tcl_DStringAppend(buf, "{", -1);
	}
	if(rule->flags & AVFLAG_TGT_STAR)
		Tcl_DStringAppend(buf, "*", -1);

	
	for(tptr = rule->tgt_roles; tptr != NULL; tptr = tptr->next) {
		assert(tptr->type == IDX_ROLE);
		Tcl_DStringAppend(buf, " ", -1);
		Tcl_DStringAppend(buf, policy->roles[tptr->idx].name, -1);
		Tcl_DStringAppend(buf, " ", -1);
	}
	if(multiple) {
		Tcl_DStringAppend(buf, "}", -1);
		multiple = 0;
	}
	
	Tcl_DStringAppend(buf, ";\n", -1);
		
	return 0;
}


static int append_role_trans_rule(rt_item_t *rule, policy_t *policy, Tcl_DString *buf)
{
	ta_item_t *tptr;	
	int multiple = 0;
	if(buf == NULL) {
		return -1;
	}
		
	Tcl_DStringAppend(buf, rulenames[RULE_ROLE_TRANS], -1);
	
	/* source roles */
	if(rule->flags & AVFLAG_SRC_TILDA) 
		Tcl_DStringAppend(buf, " ~", -1);
	else
		Tcl_DStringAppend(buf, " ", -1);
	if(rule->src_roles != NULL && rule->src_roles->next != NULL) {
		multiple = 1;
		Tcl_DStringAppend(buf, "{", -1);
	}
	if(rule->flags & AVFLAG_SRC_STAR)
		Tcl_DStringAppend(buf, "*", -1);
			
	for(tptr = rule->src_roles; tptr != NULL; tptr = tptr->next) {
		assert(tptr->type == IDX_ROLE);
		Tcl_DStringAppend(buf, " ", -1);
		Tcl_DStringAppend(buf, policy->roles[tptr->idx].name, -1);
		Tcl_DStringAppend(buf, " ", -1);
	}
	if(multiple) {
		Tcl_DStringAppend(buf, "}", -1);
		multiple = 0;
	}
	/* target types/attributes */
	if(rule->flags & AVFLAG_TGT_TILDA) 
		Tcl_DStringAppend(buf, " ~", -1);
	else
		Tcl_DStringAppend(buf, " ", -1);
	if(rule->tgt_types != NULL && rule->tgt_types->next != NULL) {
		multiple = 1;
		Tcl_DStringAppend(buf, "{", -1);
	}
	if(rule->flags & AVFLAG_TGT_STAR)
		Tcl_DStringAppend(buf, "*", -1);

	for(tptr = rule->tgt_types; tptr != NULL; tptr = tptr->next) {
		if(tptr->type == IDX_TYPE) {
			Tcl_DStringAppend(buf, " ", -1);
			if(append_type_str(0, 0, 0, tptr->idx, policy, buf) != 0)
				return -1;
		}
		else if(tptr->type == IDX_ATTRIB) {
			Tcl_DStringAppend(buf, " ", -1);
			if(append_attrib_str(0, 0, 0, 0, 0, tptr->idx, policy, buf) != 0)
				return -1;
		}			
		else {
			fprintf(stderr, "Invalid index type: %d\n", tptr->type);
			return -1;
		}
	}
	if(multiple) {
		Tcl_DStringAppend(buf, "}", -1);
		multiple = 0;
	}
	
	/* default role */
	Tcl_DStringAppend(buf, " ", -1);
	assert(rule->trans_role.type == IDX_ROLE);
	Tcl_DStringAppend(buf, policy->roles[rule->trans_role.idx].name, -1);
	
	Tcl_DStringAppend(buf, ";\n", -1);
		
	return 0;
}



static int read_policy(void)
{
	/* assumed yyin is opened to policy file */
	id_queue = queue_create();
	if (!id_queue) {
		fprintf(stderr, "out of memory\n");
		queue_destroy(id_queue);
		fclose(yyin);
		return -1;
	}
	policydb_lineno = 1;
	pass = 1;
	if (yyparse()) {
		fprintf(stderr, "error(s) encountered while parsing configuration (first pass)\n");
		queue_destroy(id_queue);;
		fclose(yyin);
		return -1;
	}
	policydb_lineno = 1;
	pass = 2;
	rewind(yyin);
	yyrestart(yyin);	
	if (yyparse()) {
		fprintf(stderr, "error(s) encountered while parsing configuration (second pass)\n");
		fclose(yyin);
		queue_destroy(id_queue);
		return -1;
	}
		
	queue_destroy(id_queue);
	fclose(yyin);
	return 0;		
}

static int open_policy(void) 
{
	int rt;
	rt = init_policy(&policy);
	if(rt != 0) {
		fprintf(stderr, "error initializing policy\n");
		return -1;
	}
	rt = read_policy();
	if(rt != 0) {
		fprintf(stderr, "error reading policy\n");
		return -1;	
	}
	return 0;
}




/**************************************************************************
 * TCL interface functions
 **************************************************************************/


/* Get the directory where the TCL scripts are located.  This function
 * simply returns the value of the script_dir GLOBAL variable defined above 
 * if has been set previously.  Otherwise it calls
 * find_tcl_script() and then returns the variable.  Someone needs to call
 * this function during or prior to running scripts that use these commands.
 *
 * There is one argument, the file name of the top-level TCL script (e.g.,
 * apol.tcl) which is located according to find_tcl_script().  The presumption
 * is that any other TCL script will be in the same directory.
 */
int Apol_GetScriptDir(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	
	if(script_dir == NULL) {
		script_dir = find_tcl_script(argv[1]);
		if(script_dir == NULL) {
			Tcl_AppendResult(interp, "problem locating TCL startup script", (char *) NULL);
			return TCL_ERROR;
		}
	}
	assert(script_dir != NULL);
	Tcl_AppendResult(interp, script_dir, (char *) NULL);
	return TCL_OK;		
}

/* open a policy.conf file */
int Apol_OpenPolicy(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	char tbuf[APOL_STR_SZ+64];
	int rt;
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char*)NULL);
		return TCL_ERROR;
	}
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "File name string too large", (char *) NULL);
		return TCL_ERROR;
	}
	if((yyin = fopen(argv[1], "r")) == NULL) {
		Tcl_AppendResult(interp, "cannot open policy file ", argv[1], (char *) NULL);
		return TCL_ERROR;
	}	
	free_policy(&policy);
#ifdef APOL_PERFORM_TEST
/*  test policy load performance; it's an undocumented feature only in test builds */
	{
	clock_t start,  stop;
	double time;
	start = clock();	
	rt = open_policy();
	stop = clock();
	time = ((double) (stop - start)) / CLOCKS_PER_SEC;
	fprintf(stdout, "\nTime to load policy %s: %f\n\n", argv[1], time);
	}
#else
	rt = open_policy();
#endif
	if(rt != 0) {
		free_policy(&policy);
		sprintf(tbuf, "open_policy error (%d)", rt);
		Tcl_AppendResult(interp, tbuf, (char *) NULL);
		return TCL_ERROR;
	}
	return TCL_OK;	
}

int Apol_ClosePolicy(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	free_policy(&policy);
	policy = NULL;
	return TCL_OK;
}

int Apol_GetVersion(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	Tcl_AppendResult(interp, VERSION_STRING, (char *) NULL);
	return TCL_OK;
}

/* Return flags indicating what data is in the current policy.  Following data types:
 *     	classes		object classes
 *	perms		permissions (including common perms)
 *	types		types and attributes
 *	te_rules	type enforcement rules, including allow, type_trans, audit_*, etc.
 *	roles		roles
 *	rbac		role rules
 *	users		user definitions
 */
int Apol_GetPolicyContents(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}

/* FIX: This is a place-holder function to be used by the GUI...need to have the policy.c
 * stuff control this via flags. */
	Tcl_AppendElement(interp, "classes 1");
	Tcl_AppendElement(interp, "perms 1");
	Tcl_AppendElement(interp, "types 1");
	Tcl_AppendElement(interp, "te_rules 1");
	Tcl_AppendElement(interp, "roles 1");
	Tcl_AppendElement(interp, "rbac 1");
	Tcl_AppendElement(interp, "users 1");
	return TCL_OK;	
}

int Apol_GetPolicyVersion(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	switch (policy->version) {
	case POL_VER_PRE_11:
		Tcl_AppendResult(interp, POL_VER_STRING_PRE_11, (char *) NULL);
		break;
	case POL_VER_11:
	/* case POL_VER_12: */ /* (currently synonmous with v.11 */
		Tcl_AppendResult(interp, POL_VER_STRING_11, (char *) NULL);
		break;
	default:
		Tcl_AppendResult(interp, "Unkown version", (char *) NULL);
		break;
	}
	return TCL_OK;
}

/* return statics about the policy */
int Apol_GetStats(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	char buf[128];
	if(argc != 1) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	/* elements */	
	sprintf(buf, "types %d", policy->num_types);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "attribs %d", policy->num_attribs);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "roles %d", policy->num_roles);
	Tcl_AppendElement(interp, buf);
/* FIX: Rather than this type of code, we need to add TCL flags that indicate
 *      what was collected so that tcl/tk GUIs will know what to expect
 */
	sprintf(buf, "classes %d", policy->num_obj_classes);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "common_perms %d", policy->num_common_perms);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "perms %d", policy->num_perms);
	Tcl_AppendElement(interp, buf);
		
	/* rules */
	sprintf(buf, "teallow %d", policy->rule_cnt[RULE_TE_ALLOW]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "neverallow %d", policy->rule_cnt[RULE_NEVERALLOW]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "auditallow %d", policy->rule_cnt[RULE_AUDITALLOW]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "auditdeny %d", policy->rule_cnt[RULE_AUDITDENY]);
	Tcl_AppendElement(interp, buf);	
	sprintf(buf, "dontaudit %d", policy->rule_cnt[RULE_DONTAUDIT]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "tetrans %d", policy->rule_cnt[RULE_TE_TRANS]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "temember %d", policy->rule_cnt[RULE_TE_MEMBER]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "techange %d", policy->rule_cnt[RULE_TE_CHANGE]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "clone %d", policy->rule_cnt[RULE_CLONE]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "roleallow %d", policy->rule_cnt[RULE_ROLE_ALLOW]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "roletrans %d", policy->rule_cnt[RULE_ROLE_TRANS]);
	Tcl_AppendElement(interp, buf);
	sprintf(buf, "users %d", policy->rule_cnt[RULE_USER]);
	Tcl_AppendElement(interp, buf);
			
	return TCL_OK;
}

/* 
 * Get information about object classes, permissions, and common perms
 * args are:
 *
 * 1	do_classes	(bool)
 * 2	classes_perms 	(bool, ignored if !do_classes)
 * 3	classes_cps	(bool, ignored if !do_classes || !classes_cps)
 * 4	do_comm_perms	(bool)
 * 5	cp_perms	(bool, ignored if !do_comm_perms)
 * 6	cp_classes	(bool, ignored if !do_comm_perms)
 * 7	do_perms	(bool)
 * 8	perm_classes	(bool, ignored if !do_perms)
 * 9	perm_cps	(bool, ignored if !do_perms)
 * 10	use_srchstr	(bool, regex)
 * 11	srch_str	(string, ignored if !use_srch_str)
 *
 */
int Apol_GetClassPermInfo(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int i, sz, rt;
	char *err;
	Tcl_DString buffer, *buf = &buffer;
	bool_t do_classes, classes_perms, classes_cps, do_common_perms, cp_perms, cp_classes, do_perms,
		perm_classes, perm_cps, use_srchstr;
	regex_t reg;
	
	if(argc != 12) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	do_classes = getbool(argv[1]);
	if(do_classes) {
		classes_perms = getbool(argv[2]);
		if(classes_perms)
			classes_cps = getbool(argv[3]);
		else
			classes_cps = FALSE;
	}
	else {
		classes_perms = FALSE;
		classes_cps = FALSE;
	}
	do_common_perms = getbool(argv[4]);
	if(do_common_perms) {
		cp_perms = getbool(argv[5]);
		cp_classes = getbool(argv[6]);
	}
	else {
		cp_perms = FALSE;
		cp_classes = FALSE;
	}
	do_perms = getbool(argv[7]);
	if(do_perms) {
		perm_classes = getbool(argv[8]);
		perm_cps = getbool(argv[9]);
	}
	else {
		perm_classes = FALSE;
		perm_cps = FALSE;
	}
	if(!do_classes && !do_common_perms && !do_perms) {
		return TCL_OK; /* nothing to do! */
	}
	use_srchstr = getbool(argv[10]);
	if(use_srchstr) {
		if(!is_valid_str_sz(argv[11])) {
			Tcl_AppendResult(interp, "Regular expression string is too large", (char *) NULL);
			return TCL_ERROR;
		}
		rt = regcomp(&reg, argv[11], REG_ICASE|REG_EXTENDED|REG_NOSUB);
		if(rt != 0) {
			sz = regerror(rt, &reg, NULL, 0);
			if((err = (char *)malloc(++sz)) == NULL) {
				Tcl_AppendResult(interp, "out of memory", (char *) NULL);
				return TCL_ERROR;
			}
			regerror(rt, &reg, err, sz);
			Tcl_AppendResult(interp, "Invalid regular expression:\n\n     ", (char*) NULL);
			Tcl_AppendResult(interp, argv[11], (char*) NULL);
			Tcl_AppendResult(interp, "\n\n", (char*) NULL);
			Tcl_AppendResult(interp, err, (char *) NULL);
			Tcl_DStringFree(buf);
			regfree(&reg);
			free(err);
			return TCL_ERROR;
			
		}
	}

	
	Tcl_DStringInit(buf);

	/* FIX: Here and elsewhere, need to use sorted traversal using AVL trees */	
	if(do_classes) {
		Tcl_DStringAppend(buf, "OBJECT CLASSES \n\n", -1);
		for(i = 0; i < policy->num_obj_classes; i++) {
			if(use_srchstr && (regexec(&reg, policy->obj_classes[i].name, 0,NULL,0) != 0)) {
				continue;
			}
			append_class_str(classes_perms, classes_perms, classes_cps, 1, i, buf, policy);
		}
		Tcl_DStringAppend(buf, "\n", -1);
	}
	if(do_common_perms) {
		Tcl_DStringAppend(buf, "COMMON PERMISSIONS\n\n", -1);
		for(i = 0; i < policy->num_common_perms; i++) {
			if(use_srchstr && (regexec(&reg, policy->common_perms[i].name, 0,NULL,0) != 0)) {
				continue;
			}
			append_common_perm_str(cp_perms, cp_classes, 1, i, buf, policy);
		}
		Tcl_DStringAppend(buf, "\n", -1);
	}
	if(do_perms) {
		Tcl_DStringAppend(buf, "PERMISSIONS", -1);
		if(perm_classes) {
			Tcl_DStringAppend(buf,  "  (* means class uses permission via a common permission)\n\n", -1);
		}
		else {
			Tcl_DStringAppend(buf, "\n\n", -1);
		}
		for(i = 0; i < policy->num_perms; i++) {
			if(use_srchstr && (regexec(&reg, policy->perms[i], 0,NULL,0) != 0)) {
				continue;
			}
			append_perm_str(perm_cps, perm_classes, 1, i, buf, policy);
		}
		Tcl_DStringAppend(buf, "\n", -1);
	}
	
	Tcl_DStringResult(interp, buf);
	return TCL_OK;
}

/* get information for a single class/perm/common perm 
 * argv[1]	name
 * argv[2]	which ("class", "perm", or "common_perm")
 */
int Apol_GetSingleClassPermInfo(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int rt, idx;
	char tbuf[APOL_STR_SZ+64];
	Tcl_DString buffer, *buf = &buffer;
	
	if(argc != 3) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}	

	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "Class/Perm name is too large", (char *) NULL);
		return TCL_ERROR;
	}
	if(!is_valid_str_sz(argv[2])) {
		Tcl_AppendResult(interp, "Which option string is too large", (char *) NULL);
		return TCL_ERROR;
	}
	
	Tcl_DStringInit(buf);
	if(strcasecmp(argv[2], "class") == 0) {
		idx = get_obj_class_idx(argv[1], policy);
		if(idx < 0) {
			Tcl_DStringFree(buf);
			sprintf(tbuf, "%s is an invalid class name", argv[1]);
			Tcl_AppendResult(interp, tbuf, (char *) NULL);
			return TCL_ERROR;
		}
		rt = append_class_str(1, 1, 0, 0, idx, buf, policy);
		if(rt != 0) {
			Tcl_DStringFree(buf);
			Tcl_AppendResult(interp, "error appending class info", (char *) NULL);
			return TCL_ERROR;
		}
	}
	else if(strcasecmp(argv[2], "common_perm") == 0) {
		idx = get_common_perm_idx(argv[1], policy);
		if(idx < 0) {
			Tcl_DStringFree(buf);
			sprintf(tbuf, "%s is an invalid common permission name", argv[1]);
			Tcl_AppendResult(interp, tbuf, (char *) NULL);
			return TCL_ERROR;
		}
		rt = append_common_perm_str(1, 0, 0, idx, buf, policy);
		if(rt != 0) {
			Tcl_DStringFree(buf);
			Tcl_AppendResult(interp, "error appending common perm info", (char *) NULL);
			return TCL_ERROR;
		}
	}
	else if(strcasecmp(argv[2], "perm") == 0) {
		idx = get_perm_idx(argv[1], policy);
		if(idx < 0) {
			Tcl_DStringFree(buf);
			sprintf(tbuf, "%s is an invalid permission name", argv[1]);
			Tcl_AppendResult(interp, tbuf, (char *) NULL);
			return TCL_ERROR;
		}
		rt = append_perm_str(1,1, 0, idx, buf, policy);
		if(rt != 0) {
			Tcl_DStringFree(buf);
			Tcl_AppendResult(interp, "error appending permmission info", (char *) NULL);
			return TCL_ERROR;
		}
	}
	else {
		Tcl_DStringFree(buf);
		sprintf(tbuf, "%s is an invalid which options", argv[2]);
		Tcl_AppendResult(interp, tbuf, (char *) NULL);
		return TCL_ERROR;
	}
	
	Tcl_DStringResult(interp, buf);
	return TCL_OK;	
}


/* Get list of permission that are associated with given list of object classes 
 *
 * 1	classes (list)
 * 2	union (bool)	indicates whether union (1) or intersection (0) desired
 *
 */
int Apol_GetPermsByClass(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	int i, rt, num_classes, num_perms, *perms;
	bool_t p_union;
	char **classes, buf[128], *name;
	
	if(argc != 3) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	rt = Tcl_SplitList(interp, argv[1], &num_classes, &classes);
	if(rt != TCL_OK)
		return rt;
	if(num_classes < 1)  {
		Tcl_AppendResult(interp, "No object classes were provided!", (char *) NULL);
		return TCL_ERROR;
	}	
	p_union = getbool(argv[2]);
	
	rt = get_perm_list_by_classes(p_union, num_classes, classes, &num_perms, &perms, policy);
	if(rt == -2) {
		sprintf(buf, "Error with class names (%d)", num_perms);
		Tcl_AppendResult(interp, buf, (char *) NULL);
		Tcl_Free((char *) classes);
		return TCL_ERROR;
	}
	else if(rt != 0) {
		Tcl_AppendResult(interp, "Unspecified error getting permissions", (char *) NULL);
		Tcl_Free((char *) classes);
		return TCL_ERROR;
	}
	for(i = 0; i < num_perms; i++) {
		assert(is_valid_perm_idx(perms[i], policy));
		rt = get_perm_name(perms[i], &name, policy);
		if(rt != 0) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "Problem getting permission name", (char *) NULL);
			Tcl_Free((char *) classes);
			free(perms);
		}
		Tcl_AppendElement(interp, name);
		free(name);
	}
	
	free(perms);
	Tcl_Free((char *) classes);	
	return TCL_OK;
}


/* get a list of policy resource names names */
/* 
 * argv[1] indicates which name list to return, possible values are:
 * 	types
 * 	attrib
 *	roles
 * 	users
 * 	classes
 *	perms
 *	common_perms
 */
int Apol_GetNames(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	int i;
	char *name;
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	if(strcasecmp("types", argv[1]) == 0) {
		for(i = 0; get_type_name(i, &name, policy) == 0; i++) {
			Tcl_AppendElement(interp, name);
			free(name);
		}
	}
	else if(strcasecmp("attribs", argv[1]) == 0) {
		for(i = 0; get_attrib_name(i, &name, policy) == 0; i++) {
			Tcl_AppendElement(interp, name);
			free(name);
		}
	}
	else if(strcasecmp("roles", argv[1]) == 0) {
		for(i = 0; get_role_name(i, &name, policy) == 0; i++) {
			Tcl_AppendElement(interp, name);
			free(name);
		}
	}
	/* user list is a linked list and not an array like the other lists */
	else if(strcasecmp("users", argv[1]) == 0) {
		user_item_t *ptr;
		for(ptr = policy->users.head; get_user_name(ptr, &name) == 0; ptr = ptr->next) {
			Tcl_AppendElement(interp, name);
			free(name);
		}
	}
	else if(strcasecmp("classes", argv[1]) == 0) {
		for(i = 0; get_obj_class_name(i, &name, policy) == 0; i++) {
			Tcl_AppendElement(interp, name);
			free(name);
		}
	}
	else if(strcasecmp("perms", argv[1]) == 0) {
		for(i = 0; get_perm_name(i, &name, policy) == 0; i++) 
			Tcl_AppendElement(interp, name);
			free(name);
		}
	else if(strcasecmp("common_perms", argv[1]) == 0) {
		for(i = 0; get_common_perm_name(i, &name, policy) == 0; i++)
			Tcl_AppendElement(interp, name);
			free(name);
		}		
	else {
		Tcl_AppendResult(interp, "invalid name class (types, attribs, roles, users, classes, perms, or common_perms)", (char *) NULL);
		return TCL_ERROR;
	}		
	return TCL_OK;
}

/* search and return type enforcement rules */

/* This is a newer function that replaces the legacy Apol_GetTErules() function (below).  The 
 * latter is deprecated.  This function returns the search results in a tcl list.  This list
 * is organized as follows:
 *	index		contexts
 *	0		# of returned rules
 *	1		first rule (if any)
 *	2		first rule's lineno ref into policy.co
 *	3		2nd rule (if any)
 *	4		2nd rule lineno ref
 *	n-1		last rule
 *	n		last rule's lineno ref
 *
 * arg ordering; argv[x] where x is:
 *  1		te_allow
 *  2		neverallow
 *  3		clone
 *  4		auallow
 *  5		audeny
 *  6		dontaudit
 *  7		ttrans
 *  8		tmember
 *  9		tchange
 * 10		use_1
 * 11		indirect_1
 * 12		ta1	(first type/attrib search parameter)
 * 13		which	(indicates whether ta1 is used for source, or any location
 * 14		use_2
 * 15		indirect_2
 * 16		ta2	(second type/attrib search parameter, always as target)
 * 17		use_3
 * 18		indirect_3
 * 19		ta3
 * 20		classes (list)
 * 21		perms (list)
 * 22		allow_regex (bool, indicate whether ta* are regexp or not)
 * 23		ta1_opt (indicates whether ta1 is a TYPES, ATTRIBS, or BOTH)
 * 24		ta2_opt (same for ta2; NOTE ta3 is always a TYPES)
 */
int Apol_SearchTErules(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	
	int i, j, rt, sz, ta1 = -1, ta2 = -1, ta3 = -1, ta1_type, ta2_type, ta3_type, num_classes, num_perms, *cls_idxs, *perm_idxs,
		ta1_opt, ta2_opt, cntr = 0;
	Tcl_DString buffer, *buf = &buffer;
	char tmpbuf[APOL_STR_SZ+64], *err;
	char **classes, **perms;
	rules_bool_t rules_src, rules_tgt, rules_default;
	bool_t te_allow, neverallow, clone, auallow, audeny, dontaudit, ttrans, tmember, 
		tchange, use_1, use_2, use_3, any = FALSE, include_audit,
		indirect_1, indirect_2, indirect_3, allow_regex;
	regex_t reg[3];

	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	if(argc != 25) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}

	te_allow = getbool(argv[1]);
	neverallow = getbool(argv[2]);
	clone = getbool(argv[3]);
	auallow = getbool(argv[4]);
	audeny = getbool(argv[5]);
	dontaudit = getbool(argv[6]);
	ttrans = getbool(argv[7]);
	tmember = getbool(argv[8]);
	tchange = getbool(argv[9]);
	use_1 = getbool(argv[10]);
	indirect_1 = getbool(argv[11]);
	indirect_2 = getbool(argv[15]);
	indirect_3 = getbool(argv[18]);
	allow_regex = getbool(argv[22]);
	
	/* for ta*_opt, following values
	 * 	0 = types
	 *	1 = attribs
	 *	2 = both (aka either) */
	if(strcasecmp("types", argv[23])  == 0) 
		ta1_opt = 0;
	else	if(strcasecmp("attribs", argv[23]) == 0) 
		ta1_opt = 1;
	else if((strcasecmp("both", argv[23]) == 0) ||( strcasecmp("either", argv[23]) == 0)) 
		ta1_opt = 2;
	else {
		sprintf(tmpbuf, "ta1_opt value invalid: %s", argv[23]);
		Tcl_AppendResult(interp, tmpbuf, (char*) NULL);
		return TCL_ERROR;
	}
	if(strcasecmp("types", argv[24])  == 0) 
		ta2_opt = 0;
	else if(strcasecmp("attribs", argv[24]) == 0) 
		ta2_opt = 1;
	else if((strcasecmp("both", argv[24]) == 0) || ( strcasecmp("either", argv[24]) == 0)) 
		ta2_opt = 2;
	else {
		sprintf(tmpbuf, "ta2_opt value invalid: %s", argv[24]);
		Tcl_AppendResult(interp, tmpbuf, (char*) NULL);		
		return TCL_ERROR;
	}
	
	include_audit = auallow || audeny || dontaudit;

	if(use_1) {
		if(strcmp(argv[13], "source") == 0) 
			any = FALSE;
		else if(strcmp(argv[13], "either") == 0)
			any = TRUE;
		else {
			Tcl_AppendResult(interp, "Invalid which option for source parameter", (char *) NULL);
			return TCL_ERROR;			
		}
		if(!is_valid_str_sz(argv[12])) {
			Tcl_AppendResult(interp, "Source type/attrib string too large", (char *) NULL);
			return TCL_ERROR;
		}
		if(allow_regex) {
			rt = regcomp(&(reg[0]), argv[12], REG_ICASE|REG_EXTENDED|REG_NOSUB);
			if(rt != 0) {
				sz = regerror(rt, &(reg[0]), NULL, 0);
				if((err = (char *)malloc(++sz)) == NULL) {
					Tcl_AppendResult(interp, "out of memory", (char *) NULL);
					return TCL_ERROR;
				}
				regerror(rt, &(reg[0]), err, sz);
				sprintf(tmpbuf, "Invalid regular expression (source):\n\n     %s\n\n", argv[8]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				Tcl_AppendResult(interp, err, (char *) NULL);
				regfree(&(reg[0]));
				free(err);
				return TCL_ERROR;
				
			}
		}
		else {
			ta1 = get_type_or_attrib_idx(argv[12], &ta1_type, policy);
			if(ta1 < 0) {
				Tcl_AppendResult(interp, "Invalid Source type or attribute", (char *) NULL);
				return TCL_ERROR;			
			}
		}
	}

	use_2 = getbool(argv[14]) && !any;
	if(use_2) {
		if(!is_valid_str_sz(argv[16])) {
			Tcl_AppendResult(interp, "Target type/attrib string too large", (char *) NULL);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			return TCL_ERROR;
		}
		if(allow_regex) {
			rt = regcomp(&(reg[1]), argv[16], REG_ICASE|REG_EXTENDED|REG_NOSUB);
			if(rt != 0) {
				sz = regerror(rt, &(reg[1]), NULL, 0);
				if((err = (char *)malloc(++sz)) == NULL) {
					Tcl_AppendResult(interp, "out of memory", (char *) NULL);
					return TCL_ERROR;
				}
				regerror(rt, &(reg[1]), err, sz);
				sprintf(tmpbuf, "Invalid regular expression (target):\n\n     %s\n\n", argv[8]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				Tcl_AppendResult(interp, err, (char *) NULL);
				if(use_1){regfree(&(reg[1]));}
				free(err);
				return TCL_ERROR;
				
			}
		}
		else {
			ta2 = get_type_or_attrib_idx(argv[16], &ta2_type, policy);
			if(ta2 < 0) {
				Tcl_AppendResult(interp, "Invalid Target type or attribute", (char *) NULL);
				return TCL_ERROR;			
			}
		}
	}
	use_3 = getbool(argv[17]) && !any;
	if(use_3) {
		if(!is_valid_str_sz(argv[19])) {
			Tcl_AppendResult(interp, "Third type/attrib string too large", (char *) NULL);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			return TCL_ERROR;
		}
		if(allow_regex) {
			rt = regcomp(&(reg[2]), argv[19], REG_ICASE|REG_EXTENDED|REG_NOSUB);
			if(rt != 0) {
				sz = regerror(rt, &(reg[2]), NULL, 0);
				if((err = (char *)malloc(++sz)) == NULL) {
					Tcl_AppendResult(interp, "out of memory", (char *) NULL);
					return TCL_ERROR;
				}
				regerror(rt, &(reg[2]), err, sz);
				sprintf(tmpbuf, "Invalid regular expression (target):\n\n     %s\n\n", argv[8]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				Tcl_AppendResult(interp, err, (char *) NULL);
				if(use_1){regfree(&(reg[0]));}
				if(use_2){regfree(&(reg[1]));}
				free(err);
				return TCL_ERROR;
				
			}
		}
		else {
			ta3 = get_type_or_attrib_idx(argv[19], &ta3_type, policy);
			if(ta3 < 0) {
				sprintf(tmpbuf, "Invalid Default type or attribute (%s)", argv[19]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;			
			}
		}
	}

	/* classes */
	rt = Tcl_SplitList(interp, argv[20], &num_classes, &classes);
	if(rt != TCL_OK) {
		Tcl_AppendResult(interp, "error splitting classes", (char *) NULL);
		return rt;
	}
	if(num_classes < 1) {
		cls_idxs = NULL;
	}
	else {
		cls_idxs = (int *)malloc(sizeof(int)*num_classes );
		if(cls_idxs == NULL) {
			Tcl_Free((char *) classes);
			Tcl_Free((char *) perms);
			Tcl_AppendResult(interp, "out of memory", (char *) NULL);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}
			return TCL_ERROR;
		}
		for(i = 0; i < num_classes; i++) {
			cls_idxs[i] = get_obj_class_idx(classes[i], policy);
			if(cls_idxs[i] < 0) {
				sprintf(tmpbuf, "%s is not a valid object class name", classes[i]);
				Tcl_Free((char *) classes);
				Tcl_Free((char *) perms);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				free(cls_idxs);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;
			}
		}
		Tcl_Free((char *) classes);
	}
	/* perms */
	rt = Tcl_SplitList(interp, argv[21], &num_perms, &perms);
	if(rt != TCL_OK) {
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		Tcl_AppendResult(interp, "error splitting perms", (char *) NULL);
		return rt;
	}
	if(num_perms < 1) {
		perm_idxs = NULL;
	}
	else {
		perm_idxs = (int *)malloc(sizeof(int)*num_perms);
		if(perm_idxs == NULL) {
			Tcl_Free((char *) perms);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}
			Tcl_AppendResult(interp, "out of memory", (char *) NULL);
			return TCL_ERROR;
		}
		for(i = 0; i < num_perms; i++) {
			perm_idxs[i] = get_perm_idx(perms[i], policy);
			if(perm_idxs[i] < 0) {
				sprintf(tmpbuf, "%s is not a permission name", perms[i]);
				Tcl_Free((char *) perms);
				if(cls_idxs != NULL)
					free(cls_idxs);
				free(perm_idxs);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;
			}
		}
		Tcl_Free((char *) perms);
	}
		
	/* set up match structures */
	if(init_rules_bool(include_audit, &rules_src, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing src rules bool", (char *) NULL);
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(perm_idxs != NULL)
			free(perm_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		return TCL_ERROR;
	}
	if(init_rules_bool(include_audit, &rules_tgt, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing tgt rules bool", (char *) NULL);
		free_rules_bool(&rules_src);
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(perm_idxs != NULL)
			free(perm_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		return TCL_ERROR;
	}
	if(init_rules_bool(include_audit, &rules_default, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing default rules bool", (char *) NULL);
		free_rules_bool(&rules_src);	
		free_rules_bool(&rules_tgt);
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(perm_idxs != NULL)
			free(perm_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		return TCL_ERROR;
	}

	if(use_1) {
		if(match_te_rules(allow_regex, &(reg[0]), ta1_opt, ta1, ta1_type, include_audit, SRC_LIST, indirect_1, &rules_src, policy) != 0) {
			Tcl_AppendResult(interp, "error matching source type/attrib", (char *) NULL);
			free_rules_bool(&rules_src);	
			free_rules_bool(&rules_tgt);	
			free_rules_bool(&rules_default);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(perm_idxs != NULL)
				free(perm_idxs);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}	
			return TCL_ERROR;
		}
	}
	else {
		all_true_rules_bool(&rules_src, policy);
	}
	
	if(use_1 && any) { 
		/* since "any", need to check target and default lists too */
		if(match_te_rules(allow_regex, &(reg[0]), ta1_opt, ta1, ta1_type, include_audit, TGT_LIST, indirect_1, &rules_tgt, policy) != 0) {
			Tcl_AppendResult(interp, "error matching target type/attrib", (char *) NULL);
			free_rules_bool(&rules_src);	
			free_rules_bool(&rules_tgt);	
			free_rules_bool(&rules_default);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(perm_idxs != NULL)
				free(perm_idxs);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}	
			return TCL_ERROR;
		}
		if(match_te_rules(allow_regex, &(reg[0]), 0, ta1, ta1_type, include_audit, DEFAULT_LIST, indirect_1, &rules_default, policy) != 0) {
			Tcl_AppendResult(interp, "error matching default type/attrib", (char *) NULL);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(perm_idxs != NULL)
				free(perm_idxs);
			free_rules_bool(&rules_src);	
			free_rules_bool(&rules_tgt);	
			free_rules_bool(&rules_default);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}
			return TCL_ERROR;
		}	
	}
	else {
		if(use_2) {
			if(match_te_rules(allow_regex, &(reg[1]), ta2_opt, ta2, ta2_type, include_audit, TGT_LIST, indirect_2, &rules_tgt, policy) != 0) {
				Tcl_AppendResult(interp, "error matching target type/attrib", (char *) NULL);
				if(cls_idxs != NULL)
					free(cls_idxs);
				if(perm_idxs != NULL)
					free(perm_idxs);
				free_rules_bool(&rules_src);	
				free_rules_bool(&rules_tgt);	
				free_rules_bool(&rules_default);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				return TCL_ERROR;
			}	
		}
		else {
			all_true_rules_bool(&rules_tgt, policy);
		}
		if(use_3) {
			if(match_te_rules(allow_regex, &(reg[2]), 0, ta3, ta3_type, include_audit, DEFAULT_LIST, indirect_3, &rules_default, policy) != 0) {
				Tcl_AppendResult(interp, "error matching default type/attrib", (char *) NULL);
				if(cls_idxs != NULL)
					free(cls_idxs);
				if(perm_idxs != NULL)
					free(perm_idxs);
				free_rules_bool(&rules_src);	
				free_rules_bool(&rules_tgt);	
				free_rules_bool(&rules_default);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				return TCL_ERROR;
			}	
		}
		else {
			all_true_rules_bool(&rules_default, policy);		}
	}
	
	if(use_1 && allow_regex){regfree(&(reg[0]));}
	if(use_2 && allow_regex){regfree(&(reg[1]));}
	if(use_3 && allow_regex){regfree(&(reg[2]));}	
	
	/* further select and output matching rules */
	Tcl_DStringInit(buf);
	if(te_allow || neverallow) {
		for(i = 0, j = 0; i < policy->num_av_access; i++) {
			if(
			   ((te_allow && (policy->av_access[i].type == RULE_TE_ALLOW)) ||
			   (neverallow && (policy->av_access[i].type == RULE_NEVERALLOW))) 
			   &&			   
			   ((!any && (rules_src.access[i] && rules_tgt.access[i])) ||
			   (any && (rules_src.access[i] || rules_tgt.access[i])))
			  &&
			   ( (num_classes < 1) || ((num_classes > 0) && does_av_rule_use_classes(i, 1, cls_idxs, num_classes, policy)) )
			  &&
			   ( (num_perms < 1) || ((num_perms > 0) && does_av_rule_use_perms(i, 1, perm_idxs, num_perms, policy)) )
			  )	{
				/* Have a matching AV rule */
				cntr++;
				rt = append_av_rule(0, 0, i, FALSE, policy, buf);
				if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);
					return TCL_ERROR;
				}
				Tcl_AppendElement(interp, buf->string);
				Tcl_DStringFree(buf);
				sprintf(tmpbuf, "%lu", policy->av_access[i].lineno);
				Tcl_AppendElement(interp, tmpbuf);
			}
		}
	}
	if(clone) {
		for(i = 0, j = 0; i < policy->rule_cnt[RULE_CLONE]; i++) {
			if(
			   ((!any && (rules_src.clone[i] && rules_tgt.clone[i])) ||
			   (any && (rules_src.clone[i] || rules_tgt.clone[i])))
			  ) {
				rt = append_clone_rule(0, 0, &(policy->clones[i]), policy, buf);
				if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);
					return TCL_ERROR;
				}
				Tcl_AppendElement(interp, buf->string);
				Tcl_DStringFree(buf);
				sprintf(tmpbuf, "%lu", policy->clones[i].lineno);
				Tcl_AppendElement(interp, tmpbuf);			  	
			}
		}
	}
	if(ttrans || tmember || tchange) { 
		/* only case to use third (default) list */
		for(i = 0, j = 0; i < policy->num_te_trans; i++) {
			if(
			   ((ttrans && (policy->te_trans[i].type == RULE_TE_TRANS)) ||
			    (tmember && (policy->te_trans[i].type == RULE_TE_MEMBER)) ||
			    (tchange && (policy->te_trans[i].type == RULE_TE_CHANGE)))
			   &&
			   ((!any && (rules_src.ttrules[i] && rules_tgt.ttrules[i] && rules_default.ttrules[i])) ||
			    (any && (rules_src.ttrules[i] || rules_tgt.ttrules[i] || rules_default.ttrules[i])))
	  		   &&
			   ( (num_classes < 1) || ((num_classes > 0) && does_tt_rule_use_classes(i, cls_idxs, num_classes, policy)) )
			  ) {
			  	rt = append_tt_rule(0, 0, i, policy, buf);
			  	if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);
					return TCL_ERROR;
				}
				Tcl_AppendElement(interp, buf->string);
				Tcl_DStringFree(buf);
				sprintf(tmpbuf, "%lu", policy->te_trans[i].lineno);
				Tcl_AppendElement(interp, tmpbuf);
			}			
		}
	}
	if(include_audit) {
		for(i = 0, j = 0; i < policy->num_av_audit; i++) {
			if(
			   ((auallow && (policy->av_audit[i].type == RULE_AUDITALLOW)) ||
			   (audeny  && (policy->av_audit[i].type == RULE_AUDITDENY)) ||
			   (dontaudit  && (policy->av_audit[i].type == RULE_DONTAUDIT))) 
			   &&
			   ((!any && (rules_src.audit[i] && rules_tgt.audit[i])) ||
			   (any && (rules_src.audit[i] || rules_tgt.audit[i])))
	  		   &&
			    ( (num_classes < 1) || ((num_classes > 0) && does_av_rule_use_classes(i, 0, cls_idxs, num_classes, policy)) )
			   &&
			    ( (num_perms < 1) || ((num_perms > 0) && does_av_rule_use_perms(i, 0, perm_idxs, num_perms, policy)) )
			  )	{
				rt = append_av_rule(0, 0, i, TRUE, policy, buf);
				if(rt != 0){
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);	
					return TCL_ERROR;
				}
				Tcl_AppendElement(interp, buf->string);
				Tcl_DStringFree(buf);
				sprintf(tmpbuf, "%lu", policy->av_audit[i].lineno);
				Tcl_AppendElement(interp, tmpbuf);
			}
		}
	}
	if(cls_idxs != NULL)
		free(cls_idxs);
	if(perm_idxs != NULL)
		free(perm_idxs);
	free_rules_bool(&rules_src);	
	free_rules_bool(&rules_tgt);	
	free_rules_bool(&rules_default);	
	
//	Tcl_DStringResult(interp, buf);
	return TCL_OK;	
}

/* NOTE: The below funtion is the OLD function that is being deprecated.  It's a legacy functions
 * whose purpose was to essentially build one huge string that is returned.  This evolved from 
 * an early command line version that simply outputed to stdout.  Apol_SearchTErules(above) replaces
 * this function with one that returns list of rules and lineno refs to the GUI, which can render
 * this information any way it wants.
 */
/* arg ordering; argv[x] where x is:
 *  1		te_allow
 *  2		neverallow
 *  3		clone
 *  4		auallow
 *  5		audeny
 *  6		dontaudit
 *  7		ttrans
 *  8		tmember
 *  9		tchange
 * 10		use_1
 * 11		indirect_1
 * 12		ta1	(first type/attrib search parameter)
 * 13		which	(indicates whether ta1 is used for source, or any location
 * 14		use_2
 * 15		indirect_2
 * 16		ta2	(second type/attrib search parameter, always as target)
 * 17		use_3
 * 18		indirect_3
 * 19		ta3
 * 20		classes (list)
 * 21		perms (list)
 * 22		allow_regex (bool, indicate whether ta* are regexp or not)
 * 23		ta1_opt (indicates whether ta1 is a TYPES, ATTRIBS, or BOTH)
 * 24		ta2_opt (same for ta2; NOTE ta3 is always a TYPES)
 */
int Apol_GetTErules(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	
	int i, j, rt, sz, ta1 = -1, ta2 = -1, ta3 = -1, ta1_type, ta2_type, ta3_type, num_classes, num_perms, *cls_idxs, *perm_idxs,
		ta1_opt, ta2_opt;
	Tcl_DString buffer, *buf = &buffer;
	char tmpbuf[APOL_STR_SZ+64], *err;
	char **classes, **perms;
	rules_bool_t rules_src, rules_tgt, rules_default;
	bool_t te_allow, neverallow, clone, auallow, audeny, dontaudit, ttrans, tmember, 
		tchange, use_1, use_2, use_3, any = FALSE, include_audit,
		indirect_1, indirect_2, indirect_3, allow_regex;
	regex_t reg[3];

	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	if(argc != 25) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}

	te_allow = getbool(argv[1]);
	neverallow = getbool(argv[2]);
	clone = getbool(argv[3]);
	auallow = getbool(argv[4]);
	audeny = getbool(argv[5]);
	dontaudit = getbool(argv[6]);
	ttrans = getbool(argv[7]);
	tmember = getbool(argv[8]);
	tchange = getbool(argv[9]);
	use_1 = getbool(argv[10]);
	indirect_1 = getbool(argv[11]);
	indirect_2 = getbool(argv[15]);
	indirect_3 = getbool(argv[18]);
	allow_regex = getbool(argv[22]);
	
	/* for ta*_opt, following values
	 * 	0 = types
	 *	1 = attribs
	 *	2 = both (aka either) */
	if(strcasecmp("types", argv[23])  == 0) 
		ta1_opt = 0;
	else	if(strcasecmp("attribs", argv[23]) == 0) 
		ta1_opt = 1;
	else if((strcasecmp("both", argv[23]) == 0) ||( strcasecmp("either", argv[23]) == 0)) 
		ta1_opt = 2;
	else {
		sprintf(tmpbuf, "ta1_opt value invalid: %s", argv[23]);
		Tcl_AppendResult(interp, tmpbuf, (char*) NULL);
		return TCL_ERROR;
	}
	if(strcasecmp("types", argv[24])  == 0) 
		ta2_opt = 0;
	else if(strcasecmp("attribs", argv[24]) == 0) 
		ta2_opt = 1;
	else if((strcasecmp("both", argv[24]) == 0) || ( strcasecmp("either", argv[24]) == 0)) 
		ta2_opt = 2;
	else {
		sprintf(tmpbuf, "ta2_opt value invalid: %s", argv[24]);
		Tcl_AppendResult(interp, tmpbuf, (char*) NULL);		
		return TCL_ERROR;
	}
	
	include_audit = auallow || audeny || dontaudit;

	if(use_1) {
		if(strcmp(argv[13], "source") == 0) 
			any = FALSE;
		else if(strcmp(argv[13], "either") == 0)
			any = TRUE;
		else {
			Tcl_AppendResult(interp, "Invalid which option for source parameter", (char *) NULL);
			return TCL_ERROR;			
		}
		if(!is_valid_str_sz(argv[12])) {
			Tcl_AppendResult(interp, "Source type/attrib string too large", (char *) NULL);
			return TCL_ERROR;
		}
		if(allow_regex) {
			rt = regcomp(&(reg[0]), argv[12], REG_ICASE|REG_EXTENDED|REG_NOSUB);
			if(rt != 0) {
				sz = regerror(rt, &(reg[0]), NULL, 0);
				if((err = (char *)malloc(++sz)) == NULL) {
					Tcl_AppendResult(interp, "out of memory", (char *) NULL);
					return TCL_ERROR;
				}
				regerror(rt, &(reg[0]), err, sz);
				sprintf(tmpbuf, "Invalid regular expression (source):\n\n     %s\n\n", argv[8]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				Tcl_AppendResult(interp, err, (char *) NULL);
				regfree(&(reg[0]));
				free(err);
				return TCL_ERROR;
				
			}
		}
		else {
			ta1 = get_type_or_attrib_idx(argv[12], &ta1_type, policy);
			if(ta1 < 0) {
				Tcl_AppendResult(interp, "Invalid Source type or attribute", (char *) NULL);
				return TCL_ERROR;			
			}
		}
	}

	use_2 = getbool(argv[14]) && !any;
	if(use_2) {
		if(!is_valid_str_sz(argv[16])) {
			Tcl_AppendResult(interp, "Target type/attrib string too large", (char *) NULL);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			return TCL_ERROR;
		}
		if(allow_regex) {
			rt = regcomp(&(reg[1]), argv[16], REG_ICASE|REG_EXTENDED|REG_NOSUB);
			if(rt != 0) {
				sz = regerror(rt, &(reg[1]), NULL, 0);
				if((err = (char *)malloc(++sz)) == NULL) {
					Tcl_AppendResult(interp, "out of memory", (char *) NULL);
					return TCL_ERROR;
				}
				regerror(rt, &(reg[1]), err, sz);
				sprintf(tmpbuf, "Invalid regular expression (target):\n\n     %s\n\n", argv[8]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				Tcl_AppendResult(interp, err, (char *) NULL);
				if(use_1){regfree(&(reg[1]));}
				free(err);
				return TCL_ERROR;
				
			}
		}
		else {
			ta2 = get_type_or_attrib_idx(argv[16], &ta2_type, policy);
			if(ta2 < 0) {
				Tcl_AppendResult(interp, "Invalid Target type or attribute", (char *) NULL);
				return TCL_ERROR;			
			}
		}
	}
	use_3 = getbool(argv[17]) && !any;
	if(use_3) {
		if(!is_valid_str_sz(argv[19])) {
			Tcl_AppendResult(interp, "Third type/attrib string too large", (char *) NULL);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			return TCL_ERROR;
		}
		if(allow_regex) {
			rt = regcomp(&(reg[2]), argv[19], REG_ICASE|REG_EXTENDED|REG_NOSUB);
			if(rt != 0) {
				sz = regerror(rt, &(reg[2]), NULL, 0);
				if((err = (char *)malloc(++sz)) == NULL) {
					Tcl_AppendResult(interp, "out of memory", (char *) NULL);
					return TCL_ERROR;
				}
				regerror(rt, &(reg[2]), err, sz);
				sprintf(tmpbuf, "Invalid regular expression (target):\n\n     %s\n\n", argv[8]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				Tcl_AppendResult(interp, err, (char *) NULL);
				if(use_1){regfree(&(reg[0]));}
				if(use_2){regfree(&(reg[1]));}
				free(err);
				return TCL_ERROR;
				
			}
		}
		else {
			ta3 = get_type_or_attrib_idx(argv[19], &ta3_type, policy);
			if(ta3 < 0) {
				sprintf(tmpbuf, "Invalid Default type or attribute (%s)", argv[19]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;			
			}
		}
	}

	/* classes */
	rt = Tcl_SplitList(interp, argv[20], &num_classes, &classes);
	if(rt != TCL_OK) {
		Tcl_AppendResult(interp, "error splitting classes", (char *) NULL);
		return rt;
	}
	if(num_classes < 1) {
		cls_idxs = NULL;
	}
	else {
		cls_idxs = (int *)malloc(sizeof(int)*num_classes );
		if(cls_idxs == NULL) {
			Tcl_Free((char *) classes);
			Tcl_Free((char *) perms);
			Tcl_AppendResult(interp, "out of memory", (char *) NULL);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}
			return TCL_ERROR;
		}
		for(i = 0; i < num_classes; i++) {
			cls_idxs[i] = get_obj_class_idx(classes[i], policy);
			if(cls_idxs[i] < 0) {
				sprintf(tmpbuf, "%s is not a valid object class name", classes[i]);
				Tcl_Free((char *) classes);
				Tcl_Free((char *) perms);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				free(cls_idxs);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;
			}
		}
		Tcl_Free((char *) classes);
	}
	/* perms */
	rt = Tcl_SplitList(interp, argv[21], &num_perms, &perms);
	if(rt != TCL_OK) {
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		Tcl_AppendResult(interp, "error splitting perms", (char *) NULL);
		return rt;
	}
	if(num_perms < 1) {
		perm_idxs = NULL;
	}
	else {
		perm_idxs = (int *)malloc(sizeof(int)*num_perms);
		if(perm_idxs == NULL) {
			Tcl_Free((char *) perms);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}
			Tcl_AppendResult(interp, "out of memory", (char *) NULL);
			return TCL_ERROR;
		}
		for(i = 0; i < num_perms; i++) {
			perm_idxs[i] = get_perm_idx(perms[i], policy);
			if(perm_idxs[i] < 0) {
				sprintf(tmpbuf, "%s is not a permission name", perms[i]);
				Tcl_Free((char *) perms);
				if(cls_idxs != NULL)
					free(cls_idxs);
				free(perm_idxs);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;
			}
		}
		Tcl_Free((char *) perms);
	}
		
	/* set up match structures */
	if(init_rules_bool(include_audit, &rules_src, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing src rules bool", (char *) NULL);
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(perm_idxs != NULL)
			free(perm_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		return TCL_ERROR;
	}
	if(init_rules_bool(include_audit, &rules_tgt, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing tgt rules bool", (char *) NULL);
		free_rules_bool(&rules_src);
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(perm_idxs != NULL)
			free(perm_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		return TCL_ERROR;
	}
	if(init_rules_bool(include_audit, &rules_default, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing default rules bool", (char *) NULL);
		free_rules_bool(&rules_src);	
		free_rules_bool(&rules_tgt);
		if(cls_idxs != NULL)
			free(cls_idxs);
		if(perm_idxs != NULL)
			free(perm_idxs);
		if(use_1 && allow_regex){regfree(&(reg[0]));}
		if(use_2 && allow_regex){regfree(&(reg[1]));}
		if(use_3 && allow_regex){regfree(&(reg[2]));}
		return TCL_ERROR;
	}

	if(use_1) {
		if(match_te_rules(allow_regex, &(reg[0]), ta1_opt, ta1, ta1_type, include_audit, SRC_LIST, indirect_1, &rules_src, policy) != 0) {
			Tcl_AppendResult(interp, "error matching source type/attrib", (char *) NULL);
			free_rules_bool(&rules_src);	
			free_rules_bool(&rules_tgt);	
			free_rules_bool(&rules_default);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(perm_idxs != NULL)
				free(perm_idxs);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}	
			return TCL_ERROR;
		}
	}
	else {
		all_true_rules_bool(&rules_src, policy);
	}
	
	if(use_1 && any) { 
		/* since "any", need to check target and default lists too */
		if(match_te_rules(allow_regex, &(reg[0]), ta1_opt, ta1, ta1_type, include_audit, TGT_LIST, indirect_1, &rules_tgt, policy) != 0) {
			Tcl_AppendResult(interp, "error matching target type/attrib", (char *) NULL);
			free_rules_bool(&rules_src);	
			free_rules_bool(&rules_tgt);	
			free_rules_bool(&rules_default);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(perm_idxs != NULL)
				free(perm_idxs);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}	
			return TCL_ERROR;
		}
		if(match_te_rules(allow_regex, &(reg[0]), 0, ta1, ta1_type, include_audit, DEFAULT_LIST, indirect_1, &rules_default, policy) != 0) {
			Tcl_AppendResult(interp, "error matching default type/attrib", (char *) NULL);
			if(cls_idxs != NULL)
				free(cls_idxs);
			if(perm_idxs != NULL)
				free(perm_idxs);
			free_rules_bool(&rules_src);	
			free_rules_bool(&rules_tgt);	
			free_rules_bool(&rules_default);
			if(use_1 && allow_regex){regfree(&(reg[0]));}
			if(use_2 && allow_regex){regfree(&(reg[1]));}
			if(use_3 && allow_regex){regfree(&(reg[2]));}
			return TCL_ERROR;
		}	
	}
	else {
		if(use_2) {
			if(match_te_rules(allow_regex, &(reg[1]), ta2_opt, ta2, ta2_type, include_audit, TGT_LIST, indirect_2, &rules_tgt, policy) != 0) {
				Tcl_AppendResult(interp, "error matching target type/attrib", (char *) NULL);
				if(cls_idxs != NULL)
					free(cls_idxs);
				if(perm_idxs != NULL)
					free(perm_idxs);
				free_rules_bool(&rules_src);	
				free_rules_bool(&rules_tgt);	
				free_rules_bool(&rules_default);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				return TCL_ERROR;
			}	
		}
		else {
			all_true_rules_bool(&rules_tgt, policy);
		}
		if(use_3) {
			if(match_te_rules(allow_regex, &(reg[2]), 0, ta3, ta3_type, include_audit, DEFAULT_LIST, indirect_3, &rules_default, policy) != 0) {
				Tcl_AppendResult(interp, "error matching default type/attrib", (char *) NULL);
				if(cls_idxs != NULL)
					free(cls_idxs);
				if(perm_idxs != NULL)
					free(perm_idxs);
				free_rules_bool(&rules_src);	
				free_rules_bool(&rules_tgt);	
				free_rules_bool(&rules_default);
				if(use_1 && allow_regex){regfree(&(reg[0]));}
				if(use_2 && allow_regex){regfree(&(reg[1]));}
				if(use_3 && allow_regex){regfree(&(reg[2]));}
				return TCL_ERROR;
			}	
		}
		else {
			all_true_rules_bool(&rules_default, policy);		}
	}
	
	if(use_1 && allow_regex){regfree(&(reg[0]));}
	if(use_2 && allow_regex){regfree(&(reg[1]));}
	if(use_3 && allow_regex){regfree(&(reg[2]));}	
	
	/* further select and output matching rules */
	Tcl_DStringInit(buf);
	if(te_allow || neverallow) {
		Tcl_DStringAppend(buf, "\n\nTE Access Rules \n", -1);
		for(i = 0, j = 0; i < policy->num_av_access; i++) {
			if(
			   ((te_allow && (policy->av_access[i].type == RULE_TE_ALLOW)) ||
			   (neverallow && (policy->av_access[i].type == RULE_NEVERALLOW))) 
			   &&			   
			   ((!any && (rules_src.access[i] && rules_tgt.access[i])) ||
			   (any && (rules_src.access[i] || rules_tgt.access[i])))
			  &&
			   ( (num_classes < 1) || ((num_classes > 0) && does_av_rule_use_classes(i, 1, cls_idxs, num_classes, policy)) )
			  &&
			   ( (num_perms < 1) || ((num_perms > 0) && does_av_rule_use_perms(i, 1, perm_idxs, num_perms, policy)) )
			  )	{
				sprintf(tmpbuf, "%d: ", ++j);
				Tcl_DStringAppend(buf, tmpbuf, -1);
				rt = append_av_rule(1, 1, i, FALSE, policy, buf);
				if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);
					return TCL_ERROR;
				}
			}
		}
	}
	if(clone) {
		Tcl_DStringAppend(buf, "\n\nClone Rules \n", -1);
		for(i = 0, j = 0; i < policy->rule_cnt[RULE_CLONE]; i++) {
			if(
			   ((!any && (rules_src.clone[i] && rules_tgt.clone[i])) ||
			   (any && (rules_src.clone[i] || rules_tgt.clone[i])))
			  ) {
				sprintf(tmpbuf, "%d: ", ++j);
				Tcl_DStringAppend(buf, tmpbuf, -1);
				rt = append_clone_rule(1, 1, &(policy->clones[i]), policy, buf);
				if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);
					return TCL_ERROR;
				}			  	
			}
		}
	}
	if(ttrans || tmember || tchange) { 
		/* only case to use third (default) list */
		Tcl_DStringAppend(buf, "\n\nTE Transition/Member/Change Rules \n", -1);
		for(i = 0, j = 0; i < policy->num_te_trans; i++) {
			if(
			   ((ttrans && (policy->te_trans[i].type == RULE_TE_TRANS)) ||
			    (tmember && (policy->te_trans[i].type == RULE_TE_MEMBER)) ||
			    (tchange && (policy->te_trans[i].type == RULE_TE_CHANGE)))
			   &&
			   ((!any && (rules_src.ttrules[i] && rules_tgt.ttrules[i] && rules_default.ttrules[i])) ||
			    (any && (rules_src.ttrules[i] || rules_tgt.ttrules[i] || rules_default.ttrules[i])))
	  		   &&
			   ( (num_classes < 1) || ((num_classes > 0) && does_tt_rule_use_classes(i, cls_idxs, num_classes, policy)) )
			  ) {
			  	sprintf(tmpbuf, "%d: ", ++j);
			  	Tcl_DStringAppend(buf, tmpbuf, -1);
			  	rt = append_tt_rule(1, 1, i, policy, buf);
			  	if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);
					return TCL_ERROR;
				}
			}			
		}
	}
	if(include_audit) {
		Tcl_DStringAppend(buf, "\n\nTE Audit Rules \n", -1);
		for(i = 0, j = 0; i < policy->num_av_audit; i++) {
			if(
			   ((auallow && (policy->av_audit[i].type == RULE_AUDITALLOW)) ||
			   (audeny  && (policy->av_audit[i].type == RULE_AUDITDENY)) ||
			   (dontaudit  && (policy->av_audit[i].type == RULE_DONTAUDIT))) 
			   &&
			   ((!any && (rules_src.audit[i] && rules_tgt.audit[i])) ||
			   (any && (rules_src.audit[i] || rules_tgt.audit[i])))
	  		   &&
			    ( (num_classes < 1) || ((num_classes > 0) && does_av_rule_use_classes(i, 0, cls_idxs, num_classes, policy)) )
			   &&
			    ( (num_perms < 1) || ((num_perms > 0) && does_av_rule_use_perms(i, 0, perm_idxs, num_perms, policy)) )
			  )	{
				sprintf(tmpbuf, "%d: ", ++j);
				Tcl_DStringAppend(buf, tmpbuf, -1);
				rt = append_av_rule(1, 1, i, TRUE, policy, buf);
				if(rt != 0){
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					if(cls_idxs != NULL)
						free(cls_idxs);
					if(perm_idxs != NULL)
						free(perm_idxs);
					free_rules_bool(&rules_src);	
					free_rules_bool(&rules_tgt);	
					free_rules_bool(&rules_default);	
					return TCL_ERROR;
				}
			}
		}
	}
	if(cls_idxs != NULL)
		free(cls_idxs);
	if(perm_idxs != NULL)
		free(perm_idxs);
	free_rules_bool(&rules_src);	
	free_rules_bool(&rules_tgt);	
	free_rules_bool(&rules_default);	
	
	Tcl_DStringResult(interp, buf);
	return TCL_OK;	
}

/* get information about a single role  
 * args ordering:
 * argv[1]	name of role
 * argv[2]	# of role types per line to do
 */
int Apol_GetSingleRoleInfo(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int rt, idx, numperline;
	Tcl_DString *buf, buffer;
	char tmpbuf[APOL_STR_SZ +64];

	if(argc != 3) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	
	rt = Tcl_GetInt(interp, argv[2], &numperline);
	if(rt == TCL_ERROR) {
		Tcl_AppendResult(interp,"argv[2] apparently not an integer", (char *) NULL);
		return TCL_ERROR;
	}
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "Role string too large", (char *) NULL);
		return TCL_ERROR;
	}
	idx = get_role_idx(argv[1], policy);
	if(idx < 0) {
		sprintf(tmpbuf, "Invalid role name (%s)", argv[1]);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;			
	}
	buf = &buffer;
	Tcl_DStringInit(buf);
	
	rt = append_role(idx, FALSE, 1, policy, buf);
	if(rt != 0){
		Tcl_DStringFree(buf);
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
		return TCL_ERROR;
	}	
	Tcl_DStringResult(interp, buf);
	return TCL_OK;	
}

/* get a list of types for a give role
 * argv[1] role name
 */
int Apol_RoleTypes(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	char tmpbuf[APOL_STR_SZ+64], *name;
	int i, idx, rt;
	
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}	
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "Role string too large", (char *) NULL);
		return TCL_ERROR;
	}
	idx = get_role_idx(argv[1], policy);
	if(idx < 0) {
		sprintf(tmpbuf, "Invalid role: %s", argv[1]);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;
	}
	assert(strcmp(argv[1], policy->roles[idx].name) == 0);
	for(i = 0; i < policy->roles[idx].num_types; i++) {
		rt = get_type_name(policy->roles[idx].types[i], &name, policy);
		if(rt != 0) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "Problem finding a role name", (char *) NULL);
			return TCL_ERROR;
		}
		Tcl_AppendElement(interp, name);
		free(name);
	}
	
	return TCL_OK;
}


/* get roles for a user */
int Apol_UserRoles(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	char tmpbuf[APOL_STR_SZ+64], *name;
	user_item_t *user;
	ta_item_t *ptr;
	int rt;
	
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "User string too large", (char *) NULL);
		return TCL_ERROR;
	}	
	rt = get_user_by_name(argv[1], &user, policy);
	if(rt != 0) {
		sprintf(tmpbuf, "Invalid user name (%s)", argv[1]);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;		
	}
	
	for(ptr = user->roles; ptr != NULL; ptr = ptr->next) {
		assert(ptr->type == IDX_ROLE);
		rt = get_role_name(ptr->idx, &name, policy);
		if(rt != 0) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "error getting role name", (char *) NULL);			
			return TCL_ERROR;
		}
		Tcl_AppendElement(interp, name);
		free(name);
	}
		
	return TCL_OK;
} 

/* get a list of users that contain a given role */
/* args ordering for argv[x]:
 * 1	(bool) name_only (whether to show names only, or all role information)
 * 2	(bool) use_type (whether to only return roles that include provided type)
 * 3	(string, opt) type (type to use if use_type)
 */
int Apol_GetUsersByRole(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int rt, idx = -1;
	char tmpbuf[APOL_STR_SZ+64];
	bool_t name_only, use_role;
	Tcl_DString *buf, buffer;
	user_item_t *user;
	
	if(argc < 3 || argc > 4) {
		Tcl_AppendResult(interp, "apol_GetUsersByRole: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}	
	name_only = getbool(argv[1]);
	use_role = getbool(argv[2]);
	if(use_role) {
		if(argc != 4) {
			Tcl_AppendResult(interp, "apol_GetUsersByRole: wrong # of args", (char *) NULL);
			return TCL_ERROR;
		}
		if(!is_valid_str_sz(argv[3])) {
			Tcl_AppendResult(interp, "Type string too large", (char *) NULL);
			return TCL_ERROR;
		}
		idx = get_role_idx(argv[3], policy);
		if(idx < 0) {
			sprintf(tmpbuf, "Invalid role name (%s)", argv[3]);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			return TCL_ERROR;			
		}	
	}

	buf = &buffer;	
	Tcl_DStringInit(buf);
	
	for(user = get_first_user_ptr(policy); user != NULL; user = get_next_user_ptr(user)) {
		if(!use_role || does_user_have_role(user, idx, policy)) {
			rt = append_user_str(user, name_only, policy, buf);
			if(rt != 0){
				Tcl_DStringFree(buf);
				Tcl_ResetResult(interp);
				Tcl_AppendResult(interp, "error appending user", (char *) NULL);
				return rt;
			}	
		}
	}
	
	Tcl_DStringResult(interp, buf);
	return TCL_OK;	
}



/* Search role rules */
/* arg ordering for argv[x]:
 * 1	allow (bool)		get allow rules
 * 2	trans (bool)		get role_transition rules
 * 3	use_src (bool)		whether to search by source role
 * 4	source			the source role
 * 5	which			whether source used for source or any (if any, others ignored)
 *					possible values: "source", "any"
 * 6	use_tgt (bool)		whether to search by target role (allow) or type (trans)
 * 7	target			the target role/type
 * 8	tgt_is_role (bool) 	whther target is role (allow only) or type (trans only)
 * 9	use_default (bool) 	search using default role (trans only)
 * 10	default			the default role
 */
int Apol_GetRoleRules(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int i, rt, src_idx = -1, tgt_idx = -1, tgt_type = IDX_ROLE, dflt_idx = -1;
	Tcl_DString buffer, *buf = &buffer;
	char tmpbuf[APOL_STR_SZ+64];
	bool_t allow, trans, any = FALSE, use_src, use_tgt, tgt_is_role, use_dflt;
	rbac_bool_t src_b, tgt_b, dflt_b;
	
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	if(argc != 11) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}

	allow = getbool(argv[1]);
	trans = getbool(argv[2]);
	use_src = getbool(argv[3]);
	tgt_is_role = getbool(argv[8]);

	if(use_src) {
		if(strcmp(argv[5], "source") == 0)
			any = FALSE;
		else if(strcmp(argv[5], "any") == 0)
			any = TRUE;
		else {
			Tcl_AppendResult(interp, "Invalid which option for source ", (char *) NULL);
			return TCL_ERROR;			
		}
		if(!is_valid_str_sz(argv[4])) {
			Tcl_AppendResult(interp, "Source string is too large", (char *) NULL);
			return TCL_ERROR;
		}		
		src_idx = get_role_idx(argv[4], policy);
		if(src_idx < 0) {
			sprintf(tmpbuf, "Invalid source role name (%s)", argv[4]);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			return TCL_ERROR;			
		}
		
	}
	use_tgt = getbool(argv[6]) && !any;
	if(use_tgt) {
		if(allow && trans) {
			Tcl_AppendResult(interp, "Invalid option, target option may only be used if EITHER allow or role_trans is selected, but not both", (char *) NULL);
			return TCL_ERROR;
		}
		if(tgt_is_role && (!allow || trans)) {
			Tcl_AppendResult(interp, "Invalid option, target option may be a ROLE when allow, and only allow, is selected", (char *) NULL);
			return TCL_ERROR;
		}
		if(!tgt_is_role && (allow || !trans)) {
			Tcl_AppendResult(interp, "Invalid option, target option may be a TYPE when role_trans, and only role_trans, is selected", (char *) NULL);
			return TCL_ERROR;
		}
		if(tgt_is_role) {
			if(!is_valid_str_sz(argv[7])) {
				Tcl_AppendResult(interp, "Target string is too large", (char *) NULL);
				return TCL_ERROR;
			}
			tgt_idx = get_role_idx(argv[7], policy);
			if(tgt_idx < 0) {
				sprintf(tmpbuf, "Invalid target role name (%s)", argv[7]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;			
			}
			tgt_type = IDX_ROLE;
		}
		else {
			if(!is_valid_str_sz(argv[7])) {
				Tcl_AppendResult(interp, "Target string is too large", (char *) NULL);
				return TCL_ERROR;
			}
			tgt_idx = get_type_or_attrib_idx(argv[7], &tgt_type, policy);
			if(tgt_idx < 0) {
				sprintf(tmpbuf, "Invalid target type or attribute (%s)", argv[7]);
				Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
				return TCL_ERROR;			
			}
		}
		
	}
	use_dflt = getbool(argv[9]) && !any;
	if(use_dflt) {
		if(allow || !trans) {
			Tcl_AppendResult(interp, "Invalid option, default may use when role_trans, and only role_trans, is selected", (char *) NULL);
			return TCL_ERROR;
		}
		if(!is_valid_str_sz(argv[10])) {
			Tcl_AppendResult(interp, "Default string is too large", (char *) NULL);
			return TCL_ERROR;
		}
		dflt_idx = get_role_idx(argv[10], policy);
		if(dflt_idx < 0) {
			sprintf(tmpbuf, "Invalid default role name (%s)", argv[10]);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			return TCL_ERROR;			
		}
	}
	
	Tcl_DStringInit(buf);
	
	if(init_rbac_bool(&src_b, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing src rules bool", (char *) NULL);
		return TCL_ERROR;
	}
	if(init_rbac_bool(&tgt_b, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing tgt rules bool", (char *) NULL);
		free_rbac_bool(&src_b);	
		return TCL_ERROR;
	}
	if(init_rbac_bool(&dflt_b, policy) != 0) {
		Tcl_AppendResult(interp, "error initializing default rules bool", (char *) NULL);
		free_rbac_bool(&src_b);	
		free_rbac_bool(&tgt_b);	
		return TCL_ERROR;
	}
	
	if(use_src) {
		if(match_rbac_rules(src_idx, IDX_ROLE, SRC_LIST, FALSE, tgt_is_role, &src_b, policy) != 0) {
			Tcl_AppendResult(interp, "error matching source", (char *) NULL);
			free_rbac_bool(&src_b);	
			free_rbac_bool(&tgt_b);	
			free_rbac_bool(&dflt_b);	
			return TCL_ERROR;			
		}
	}
	else {
		all_true_rbac_bool(&src_b, policy);
	}
	if(use_src && any) {
		if(match_rbac_rules(src_idx, IDX_ROLE, TGT_LIST, FALSE, TRUE, &tgt_b, policy) != 0) {
			Tcl_AppendResult(interp, "error matching target", (char *) NULL);
			free_rbac_bool(&src_b);	
			free_rbac_bool(&tgt_b);	
			free_rbac_bool(&dflt_b);	
			return TCL_ERROR;			
		}
		if(match_rbac_rules(src_idx, IDX_ROLE, DEFAULT_LIST, FALSE, TRUE, &dflt_b, policy) != 0) {
			Tcl_AppendResult(interp, "error matching default", (char *) NULL);
			free_rbac_bool(&src_b);	
			free_rbac_bool(&tgt_b);	
			free_rbac_bool(&dflt_b);	
			return TCL_ERROR;			
		}
	}
	else {
		
		if(use_tgt && tgt_is_role) {
			if(match_rbac_rules(tgt_idx, IDX_ROLE, TGT_LIST, FALSE, TRUE, &tgt_b, policy) != 0) {
				Tcl_AppendResult(interp, "error matching target", (char *) NULL);
				free_rbac_bool(&src_b);	
				free_rbac_bool(&tgt_b);	
				free_rbac_bool(&dflt_b);	
				return TCL_ERROR;			
			}
		}
		else if(use_tgt && !tgt_is_role) {
			if(match_rbac_rules(tgt_idx, tgt_type, TGT_LIST, FALSE, FALSE, &tgt_b, policy) != 0) {
				Tcl_AppendResult(interp, "error matching target", (char *) NULL);
				free_rbac_bool(&src_b);	
				free_rbac_bool(&tgt_b);	
				free_rbac_bool(&dflt_b);	
				return TCL_ERROR;			
			}			
		}
		else {
			all_true_rbac_bool(&tgt_b, policy);
		}
		if(use_dflt) {
			if(match_rbac_rules(dflt_idx, IDX_ROLE, DEFAULT_LIST, FALSE, FALSE, &dflt_b, policy) != 0) {
				Tcl_AppendResult(interp, "error matching default", (char *) NULL);
				free_rbac_bool(&src_b);	
				free_rbac_bool(&tgt_b);	
				free_rbac_bool(&dflt_b);	
				return TCL_ERROR;			
			}
		}
		else {
			all_true_rbac_bool(&dflt_b, policy);
		}
	}
	
	if(allow) {
		for(i = 0; i < policy->num_role_allow; i++) {
			if((!any && (src_b.allow[i] && tgt_b.allow[i])) ||
			   (any && (src_b.allow[i] || tgt_b.allow[i]))) {
				rt = append_role_allow_rule(&(policy->role_allow[i]), policy, buf);
				if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending role allow rule", (char *) NULL);
					free_rbac_bool(&src_b);	
					free_rbac_bool(&tgt_b);	
					free_rbac_bool(&dflt_b);
					return TCL_ERROR;
				}
			}
		}
	}
	if(trans) {
		for(i =0; i < policy->num_role_trans; i++) {
			if((!any && (src_b.trans[i] && tgt_b.trans[i] && dflt_b.trans[i])) ||
			   (any && (src_b.trans[i] || tgt_b.trans[i] || dflt_b.trans[i]))) {
				rt = append_role_trans_rule(&(policy->role_trans[i]), policy, buf);
				if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending role_transition rule", (char *) NULL);
					free_rbac_bool(&src_b);	
					free_rbac_bool(&tgt_b);	
					free_rbac_bool(&dflt_b);
					return TCL_ERROR;
				}
			}
		}
	}
	
	Tcl_DStringResult(interp, buf);
	free_rbac_bool(&src_b);	
	free_rbac_bool(&tgt_b);	
	free_rbac_bool(&dflt_b);				
	return TCL_OK;	
}

/* get a list of role who contain a given type */
/* args ordering for argv[x]:
 * 1	(bool) name_only (whether to show names only, or all role information)
 * 2	(bool) use_type (whether to only return roles that include provided type)
 * 3	(string, opt) type (type to use if use_type)
 */
int Apol_GetRolesByType(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int idx = -1, i, rt;
	char tmpbuf[APOL_STR_SZ+64];
	bool_t name_only, use_type;
	Tcl_DString *buf, buffer;
	
	if(argc < 3 || argc > 4) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}	
	name_only = getbool(argv[1]);
	use_type = getbool(argv[2]);
	if(use_type) {
		if(argc != 4) {
			Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
			return TCL_ERROR;
		}
		if(!is_valid_str_sz(argv[3])) {
			Tcl_AppendResult(interp, "Type string is too large", (char *) NULL);
			return TCL_ERROR;
		}
		idx = get_type_idx(argv[3], policy);
		if(idx < 0) {
			sprintf(tmpbuf, "Invalid type name (%s)", argv[3]);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			return TCL_ERROR;			
		}	
	}

	buf = &buffer;	
	Tcl_DStringInit(buf);
	
	for(i = 0; i < policy->num_roles; i++) {
		if(!use_type || does_role_use_type(i, idx, policy)) {
			rt = append_role(i, name_only, 1, policy, buf);
			if(rt != 0){
				Tcl_DStringFree(buf);
				Tcl_ResetResult(interp);
				Tcl_AppendResult(interp, "error appending role", (char *) NULL);
				return TCL_ERROR;
			}	
		}
	}
	Tcl_DStringResult(interp, buf);
	
	return TCL_OK;	
}

/* get types for a given attribute, returns a TCL list */
/* args ordering:
 * argv[1]	attrib name
 */
int Apol_GetAttribTypesList(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int idx, j;
	char tmpbuf[APOL_STR_SZ];
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	if(!is_valid_str_sz(argv[3])) {
		Tcl_AppendResult(interp, "Type/attrib string is too large", (char *) NULL);
		return TCL_ERROR;
	}
	
	idx = get_attrib_idx(argv[1], policy);
	if(idx < 0) {
		sprintf(tmpbuf, "Invalid  attribute (%s)", argv[1]);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;			
	}	
	
	for(j = 0; j < policy->attribs[idx].num_types; j++) {
		Tcl_AppendElement(interp, policy->types[policy->attribs[idx].types[j]].name);
	}	
	
	return TCL_OK;
}

/* get information about a single type/attrib */
/* args ordering:
 * argv[1]	(bool) name only (otherwise show all information for type/attrib
 * argv[2]	(bool) attrib type's attribs (only used if argv[3] is an attrib)
 * argv[3]	name of type/attrib to get info for
 */
int Apol_GetSingleTypeInfo(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int rt, ta_idx, ta_type;
	bool_t name_only, attr_type_attribs;
	char tmpbuf[APOL_STR_SZ+64];
	Tcl_DString *buf, buffer;
	
	if(argc != 4) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}	
	name_only = getbool(argv[1]);
	attr_type_attribs = getbool(argv[2]);
	buf = &buffer;
	
	if(!is_valid_str_sz(argv[3])) {
		Tcl_AppendResult(interp, "Type/attrib string is too large", (char *) NULL);
		return TCL_ERROR;
	}
	ta_idx = get_type_or_attrib_idx(argv[3], &ta_type, policy);
	if(ta_idx < 0) {
		sprintf(tmpbuf, "Invalid type or attribute (%s)", argv[3]);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;			
	}	
	Tcl_DStringInit(buf);
	
	if(ta_type == IDX_TYPE) {
		rt = append_type_str(!name_only, !name_only, 0, ta_idx, policy, buf);
		if(rt != 0) {
			Tcl_DStringFree(buf);
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "error appending type info", (char *) NULL);
			return TCL_ERROR;
		}		
	}
	else if (ta_type == IDX_ATTRIB) {
		rt = append_attrib_str(!name_only, (!name_only ? attr_type_attribs : FALSE),
			!name_only, 0, 0, ta_idx, policy, buf);
		if(rt != 0) {
			Tcl_DStringFree(buf);
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "error appending attrib info", (char *) NULL);
			return TCL_ERROR;			
		}
	}
	else {
		Tcl_DStringFree(buf);
		Tcl_AppendResult(interp, "Invalid type from get_type_or_attrib_idx()!!", (char *) NULL);
		return TCL_ERROR;			

	}
	
	Tcl_DStringResult(interp, buf);
	return TCL_OK;	
}

/* gets information about types/attribs, based on option */
/* args ordering, each expected to be a "boolean" string 
 * argv[1]	do_types
 * argv[2] 	type_attribs
 * argv[3] 	do_attribs
 * argv[4]	attrib_types
 * argv[5]	attrib_type_attribs 
 * argv[6]	use_aliases
 * argv[7]	use search string (regexp)
 * argv[8]	search string
 */
int Apol_GetTypeInfo(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int i, rt, sz;
	Tcl_DString *buf, buffer;
	char tmpbuf[APOL_STR_SZ+64], *err;
	regex_t reg;
	
	bool_t do_types, type_attribs, do_attribs, attrib_types, attrib_type_attribs, 
		use_aliases, use_srchstr;

	buf = &buffer;
	if(argc != 9) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}
	
	do_types = getbool(argv[1]);
	type_attribs = getbool(argv[2]);
	do_attribs = getbool(argv[3]);
	attrib_types = getbool(argv[4]);
	attrib_type_attribs = getbool(argv[5]);
	use_aliases = getbool(argv[6]);
	use_srchstr = getbool(argv[7]);
		
	Tcl_DStringInit(buf);	
	if(!use_srchstr) {	
		if(do_types) {
			sprintf(tmpbuf, "\n\nTYPES (%d):\n", policy->num_types);
			Tcl_DStringAppend(buf, tmpbuf, -1);
			for(i = 0; i < policy->num_types; i++) {
				sprintf(tmpbuf, "%d: ", i+1);
				Tcl_DStringAppend(buf, tmpbuf, -1);
				rt = append_type_str(type_attribs, use_aliases, 1, i, policy, buf);
				if(rt != 0) {
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending types", (char *) NULL);
					return TCL_ERROR;
				}
			}
		}
		
		if(do_attribs) {
			sprintf(tmpbuf, "\n\nTYPE ATTRIBUTES (%d):\n", policy->num_attribs);
			Tcl_DStringAppend(buf, tmpbuf, -1);
			for(i = 0; i < policy->num_attribs; i++) {
				sprintf(tmpbuf, "%d: ", i+1);
				Tcl_DStringAppend(buf, tmpbuf, -1);
				rt = append_attrib_str(attrib_types, attrib_type_attribs, use_aliases, 1, 0, i,
					policy, buf);
				if(rt != 0){
					Tcl_DStringFree(buf);
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "error appending attributes", (char *) NULL);
					return TCL_ERROR;
				}
	
			}
		}	
	}
	else {   /* use search string; search string was provided */
		if(!is_valid_str_sz(argv[8])) {
			Tcl_AppendResult(interp, "regular expression string is too large", (char *) NULL);
			return TCL_ERROR;
		}
		rt = regcomp(&reg, argv[8], REG_ICASE|REG_EXTENDED|REG_NOSUB);
		if(rt != 0) {
			sz = regerror(rt, &reg, NULL, 0);
			if((err = (char *)malloc(++sz)) == NULL) {
				Tcl_AppendResult(interp, "out of memory", (char *) NULL);
				return TCL_ERROR;
			}
			regerror(rt, &reg, err, sz);
			sprintf(tmpbuf, "Invalid regular expression:\n\n     %s\n\n", argv[8]);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			Tcl_AppendResult(interp, err, (char *) NULL);
			Tcl_DStringFree(buf);
			regfree(&reg);
			free(err);
			return TCL_ERROR;
			
		}
		rt = append_all_ta_using_regex(&reg, argv[8], do_types, do_attribs, use_aliases, type_attribs, attrib_types, attrib_type_attribs, policy, buf);
		if(rt != 0) {
			Tcl_DStringFree(buf);
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "error searching with regex", (char *) NULL);
			regfree(&reg);
			return TCL_ERROR;
		}
		regfree(&reg);
	}
	
	Tcl_DStringResult(interp, buf);
	return TCL_OK;
}

/* Given a domain type, this function determines what new domains the given domain can transition to.
 * and returns those domains, as well as the entrypoint type that allows that transition and all
 * the rules that provide the required access.
 */
/* TODO: We're at the point where our simple tcl/tk interface is becoming a burden...probably time
 * to seriously consider native C GUIs, or maybe PerlTK (or at least use tcl v. 8 objects rather
 * than strings)....however, for now.....
 *
 * we return a list orgainzed to represent the tree structure that results from a domain transition
 * analysis.  The conceptual tree result looks like:
 *	source_type (provided)
 *		+ target_type1
 *			(list of rules providing source target:process transition)
 *			+ entry_file_type1
 *				(list of rules providing target file_type:file entrypoint)
 *				(list of rules providing source file_type:file execute)
 *			+ entry_file_type2
 *				(...entrypoint)
 *				(...execute)
 *			...
 *			+ entry_file_typeN
 *				...
 *		+ target_type2
 *			...
 *		...
 *		+ target_typeN
 *			...
 *
 * Since we're lazily using a tcl list to return the entire result, we take the above conceptual tree
 * and encoded it as follows:
 *	INDEX		CONTENTS
 *	0		source type name
 *	1		N, # of target domain types (if none, then no other results returned)
 *	  2		name first target type (if any)
 *	  3		X, # of process transition rules
 *	  next X*2	pt rule1, lineno1, ....
 *	  next		Y, # of entry point file types for first target type
 *	    next	first file type
 *	    next	A, # of file entrypoint rules
 *	    next A*2	ep rule 1, lineno1,....
 *	    next	B, # of file execute rules
 *	    next B*2	ex rule1, lineno1, ...
 *
 *	    next	(repeat next file type record as above Y times)
 *
 *	next		(repeat target type record N times)
 */

int Apol_DomainTransitionAnalysis(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	domain_trans_analysis_t *dta;
	int rt;
	
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(policy == NULL) {
		Tcl_AppendResult(interp,"No current policy file is opened!", (char *) NULL);
		return TCL_ERROR;
	}	
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "provided soruce string too large", (char *) NULL);
		return TCL_ERROR;
	}
	
	/*perform the analysis */
	rt = determine_domain_trans(argv[1], &dta, policy);
	if(rt == -2) {
		Tcl_AppendResult(interp, "invalid source type name", (char *) NULL);
		return TCL_ERROR;
	}
	else if(rt < 0) {
		Tcl_AppendResult(interp, "error with domain transition anaysis", (char *) NULL);
		return TCL_ERROR;
	}
	
	/* format the result into a tcl list (yuk!) */
	{
	llist_node_t *x, *y;
	char *tmp, tbuf[64];
	int i;
	tgt_trans_domain_t *t;
	entrypoint_type_t *ep;
	/* source type */
	rt = get_type_name(dta->src_type, &tmp, policy);
	if(rt != 0) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "analysis error (looking up source name)", (char *) NULL);
		return TCL_ERROR;
	}
	Tcl_AppendElement(interp, tmp);
	free(tmp);
	/* # of target types */
	sprintf(tbuf, "%d", dta->tgt_domains->num);
	Tcl_AppendElement(interp, tbuf);
	
	/* all target types */
	for(x = dta->tgt_domains->head; x != NULL; x = x->next) {
		t = (tgt_trans_domain_t *)x->data;
		/* target type */
		assert(dta->src_type == t->src_type);
		rt = get_type_name(t->tgt_type, &tmp, policy);
		if(rt != 0) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "analysis error (looking up target name)", (char *) NULL);
			return TCL_ERROR;
		}
		Tcl_AppendElement(interp, tmp);
		free(tmp);
		/* # of pt rules */
		sprintf(tbuf, "%d", t->num_pt_rules);
		Tcl_AppendElement(interp, tbuf);
		/* all the pt rules */
		for(i = 0; i < t->num_pt_rules; i++) {
			tmp = re_render_av_rule(0,t->pt_rules[i], 0, policy);
			if(tmp == NULL) {
				Tcl_ResetResult(interp);
				Tcl_AppendResult(interp, "analysis error (rendering process transition rule)",  (char *) NULL);
				return TCL_ERROR;
			}
			Tcl_AppendElement(interp, tmp);
			free(tmp);
			sprintf(tbuf, "%d", get_rule_lineno(t->pt_rules[i],RULE_TE_ALLOW, policy));
			Tcl_AppendElement(interp, tbuf);
		}
		/* # of entrypoint file types */
		sprintf(tbuf, "%d", t->entry_types->num);
		Tcl_AppendElement(interp, tbuf);
		/* all the entrypoint file types */
		for(y = t->entry_types->head; y != NULL; y = y->next) {
			ep = (entrypoint_type_t *)y->data;
			assert(t->tgt_type == ep->tgt_type);
			/* file type */
			rt = get_type_name(ep->file_type, &tmp, policy);
			if(rt != 0) {
				Tcl_ResetResult(interp);
				Tcl_AppendResult(interp, "analysis error (looking up entry file name)", (char *) NULL);
				return TCL_ERROR;
			}
			Tcl_AppendElement(interp, tmp);
			free(tmp);			
			/* # of file entrypoint rules */
			sprintf(tbuf, "%d", ep->num_ep_rules);
			Tcl_AppendElement(interp, tbuf);
			/* all entrypoint rules */
			for(i = 0; i < ep->num_ep_rules; i++) {
				tmp = re_render_av_rule(0, ep->ep_rules[i], 0, policy);
				if(tmp == NULL) {
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "analysis error (rendering file entrypoint rule)",  (char *) NULL);
					return TCL_ERROR;
				}
				Tcl_AppendElement(interp, tmp);
				free(tmp);
				sprintf(tbuf, "%d", get_rule_lineno(ep->ep_rules[i],RULE_TE_ALLOW, policy));
				Tcl_AppendElement(interp, tbuf);				
			}
			/* # of file execute rules */
			sprintf(tbuf, "%d", ep->num_ex_rules);
			Tcl_AppendElement(interp, tbuf);
			/* all execute rules */
			for(i = 0; i < ep->num_ex_rules; i++) {
				tmp = re_render_av_rule(0,ep->ex_rules[i], 0, policy);
				if(tmp == NULL) {
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "analysis error (rendering file execute rule)",  (char *) NULL);
					return TCL_ERROR;
				}
				Tcl_AppendElement(interp, tmp);
				free(tmp);
				sprintf(tbuf, "%d", get_rule_lineno(ep->ex_rules[i],RULE_TE_ALLOW, policy));
				Tcl_AppendElement(interp, tbuf);
			}
		}
	}
	
	} /* formal result block */
	
	free_domain_trans_analysis(dta);
	return TCL_OK;
	
}



/* Package initialization */
int Apol_Init(Tcl_Interp *interp) 
{
	Tcl_CreateCommand(interp, "apol_GetScriptDir", Apol_GetScriptDir, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_OpenPolicy", Apol_OpenPolicy, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_ClosePolicy", Apol_ClosePolicy, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetVersion", Apol_GetVersion, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetNames", Apol_GetNames, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetStats", Apol_GetStats, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetSingleTypeInfo", Apol_GetSingleTypeInfo, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetTypeInfo", Apol_GetTypeInfo, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetTErules", Apol_GetTErules, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_SearchTErules", Apol_SearchTErules, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetSingleRoleInfo", Apol_GetSingleRoleInfo, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetRolesByType", Apol_GetRolesByType, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetRoleRules", Apol_GetRoleRules, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_RoleTypes", Apol_RoleTypes, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_UserRoles", Apol_UserRoles, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetUsersByRole", Apol_GetUsersByRole, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetPolicyVersion", Apol_GetPolicyVersion, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetPolicyContents", Apol_GetPolicyContents, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetPermsByClass", Apol_GetPermsByClass, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetClassPermInfo", Apol_GetClassPermInfo, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetSingleClassPermInfo", Apol_GetSingleClassPermInfo, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_DomainTransitionAnalysis", Apol_DomainTransitionAnalysis, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "apol_GetAttribTypesList", Apol_GetAttribTypesList, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	
	Tcl_PkgProvide(interp, "apol", VERSION_STRING);

	return TCL_OK;
}



