/* Copyright (C) 2002 Tresys Technology, LLC
 * see file 'COPYING' for use and warranty information */

/*
 * Author: mayerf@tresys.com
 */

/* policy-avl.c
 *
 * AVL binary tree functions that are aware of the policy database structure.
 */

#ifdef USE_SORTING

#include "policy.h"
#include "util.h"
#include <stdlib.h>
#include <assert.h>
#include "avl-util.h"

/* These prototypes are for non-exported functions in policy.c */
int check_type_array(policy_t *policy);
int check_attrib_array(policy_t *policy);
int check_class_array(policy_t *policy);
int check_perm_array(policy_t *policy);

int init_avl_trees(policy_t *policy)
{
	int i;
	
	if(policy == NULL)
		return -1;
	
	for(i = 0; i < AVL_NUM_TREES; i++) {
		policy->tree[i].head = -1;
		policy->tree[i].list_type = i;
		policy->tree[i].ptrs = (avl_ptrs_t *)malloc(sizeof(avl_ptrs_t) * LIST_SZ);
		if(policy->tree[i].ptrs == NULL) {
			fprintf(stderr, "out of memory\n");
			return -1;
		}
	}
	return 0;
}

int free_avl_trees(policy_t *policy)
{
	int i;
	if(policy == NULL)
		return -1;
	for(i = 0; i < AVL_NUM_TREES; i++) {
		free(policy->tree[i].ptrs);
		}
	return 0;
}
 
/* Compare support function for TYPES to use with avl-utils */
int type_compare(const char *key, int idx, policy_t *policy)
{
	int cmpval;
	name_item_t *ptr;
	
	assert(!(key == NULL || policy == NULL || !is_valid_type_idx(idx, policy)));
	cmpval = strcasecmp(key, policy->types[idx].name);
	if(cmpval != 0) {
		/* FIX: for now aliases are just unsorted attached linked list;
		 * low priority since they are rare */		
		/* check aliases */
		for(ptr = policy->types[idx].aliases; ptr != NULL; ptr = ptr->next) {
			if(strcasecmp(ptr->name, key) == 0)
				return 0; /* found via alias */
		}
	}
	return cmpval;
}
/* ATTRIB comp function*/
int attrib_compare(const char *key, int idx, policy_t *policy)
{
	assert(!(key == NULL || policy == NULL || !is_valid_attrib_idx(idx, policy)));
	return strcasecmp(key, policy->attribs[idx].name);
}
/* CLASSES comp function */
int class_compare(const char *key, int idx, policy_t *policy)
{
	assert(!(key == NULL || policy == NULL || !is_valid_obj_class_idx(idx, policy)));
	return strcasecmp(key, policy->obj_classes[idx].name);
}

/* PERMS comp function */
int perm_compare(const char *key, int idx, policy_t *policy)
{
	assert(!(key == NULL || policy == NULL || !is_valid_perm_idx(idx, policy)));
	return strcasecmp(key, policy->perms[idx]);
}

/* Subtree support function for TYPES to use with avl-utils */
int get_type_subtree(int idx, int dir, policy_t *policy)
{
	assert(is_valid_type_idx(idx, policy) && policy != NULL);
	assert(dir == RIGHT || dir == LEFT);
	if(dir == LEFT) 
		return policy->tree[AVL_TYPES].ptrs[idx].left;
	else
		return policy->tree[AVL_TYPES].ptrs[idx].right;
}

/* Subtree support function for ATTRIBS to use with avl-utils */
int get_attrib_subtree(int idx, int dir, policy_t *policy)
{
	assert(is_valid_attrib_idx(idx, policy) && policy != NULL);
	assert(dir == RIGHT || dir == LEFT);
	if(dir == LEFT) 
		return policy->tree[AVL_ATTRIBS].ptrs[idx].left;
	else
		return policy->tree[AVL_ATTRIBS].ptrs[idx].right;
}

/* Subtree support function for CLASSES to use with avl-utils */
int get_class_subtree(int idx, int dir, policy_t *policy)
{
	assert(is_valid_obj_class_idx(idx, policy) && policy != NULL);
	assert(dir == RIGHT || dir == LEFT);
	if(dir == LEFT) 
		return policy->tree[AVL_CLASSES].ptrs[idx].left;
	else
		return policy->tree[AVL_CLASSES].ptrs[idx].right;
}
/* Subtree support function for CLASSES to use with avl-utils */
int get_perm_subtree(int idx, int dir, policy_t *policy)
{
	assert(is_valid_perm_idx(idx, policy) && policy != NULL);
	assert(dir == RIGHT || dir == LEFT);
	if(dir == LEFT) 
		return policy->tree[AVL_PERMS].ptrs[idx].left;
	else
		return policy->tree[AVL_PERMS].ptrs[idx].right;
}


/* checks whether a given subtree is out of balance */
const bool_t avl_check_balance(int idx, int dir, int tree_type, policy_t *policy)
{
	int l, r;
	assert(idx >= 0 && policy != NULL);
	assert((dir == RIGHT || dir == LEFT) && tree_type >= 0 && tree_type < AVL_NUM_TREES);
	l = avl_height(policy->tree[tree_type].ptrs[idx].left, policy->tree[tree_type].ptrs);
	r = avl_height(policy->tree[tree_type].ptrs[idx].right, policy->tree[tree_type].ptrs);
	if(dir == LEFT) {
		return (l - r == 2);
	}
	else {
		return (r - l == 2);
	}
}


/* return -1 for error, -2 if type already exists, the new tree head idx otherwise.
 * if -2, then newidx will contain the idx of the existing record.
 */
int avl_insert_attrib(const int head, char *attrib, int *newidx, policy_t *policy)
{
	int newhead, tmpidx, cmpval;
	
	if(head < 0) { /* check for empty subtree */
		if(check_attrib_array(policy) != 0)
			return -1;	
		/* we alloc new memory for attrib since we're not sure whether the
		 * attrib comes from an ATTRIBUTE statement or a rule, and we can't
		 * be sure the memory will be retained if a rule. 
		 */
		*newidx = policy->num_attribs;
		policy->attribs[*newidx].name = (char *) malloc(strlen(attrib)+1);
		if(policy->attribs[*newidx].name == NULL){
			fprintf(stderr, "out of memory\n");
			return -1;
		}
		strcpy(policy->attribs[*newidx].name, attrib);
		policy->attribs[*newidx].num_types = 0;
		policy->attribs[*newidx].types = NULL;
		policy->tree[AVL_ATTRIBS].ptrs[*newidx].left = -1;
		policy->tree[AVL_ATTRIBS].ptrs[*newidx].right = -1;
		policy->tree[AVL_ATTRIBS].ptrs[*newidx].height = 0;
		(policy->num_attribs)++;
		return *newidx;
	}
	/* non-empty tree */
	assert(head < policy->num_attribs);
	cmpval = strcasecmp(attrib, policy->attribs[head].name);
	if(cmpval > 0) {
		tmpidx = avl_insert_attrib(policy->tree[AVL_ATTRIBS].ptrs[head].right,
			attrib, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_ATTRIBS].ptrs[head].right = tmpidx;
		if(avl_check_balance(tmpidx, RIGHT, AVL_ATTRIBS, policy)) {
			if(strcasecmp(attrib, policy->attribs[tmpidx].name) > 0) {
				newhead = avl_srr(head, policy->tree[AVL_ATTRIBS].ptrs);
			}
			else {
				newhead = avl_drr(head, policy->tree[AVL_ATTRIBS].ptrs);
			}
		}
		
		else {
			newhead = head;
		}
	}
	else if(cmpval < 0) {
		tmpidx = avl_insert_attrib(policy->tree[AVL_ATTRIBS].ptrs[head].left,
			attrib, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_ATTRIBS].ptrs[head].left = tmpidx;
		if(avl_check_balance(tmpidx, LEFT, AVL_ATTRIBS, policy)) {
			if(strcasecmp(attrib, policy->attribs[tmpidx].name) < 0) {
				newhead = avl_srl(head, policy->tree[AVL_ATTRIBS].ptrs);
			}
			else {
				newhead = avl_drl(head, policy->tree[AVL_ATTRIBS].ptrs);
			}
		}
		
		else {
			newhead = head;
		}
	}
	else {
		*newidx = head;
		return -2; /* already exists */
	}
	policy->tree[AVL_ATTRIBS].ptrs[newhead].height = avl_max_child(newhead, policy->tree[AVL_ATTRIBS].ptrs) + 1;
	return newhead;
}

/* adds new type entry to first available and returns idx */
const int add_type_entry(char *type, policy_t *policy)
{
	int newidx;
	assert(policy != NULL && type != NULL);
	if(check_type_array(policy) != 0)
		return -1;
	newidx = policy->num_types;
	policy->types[newidx].name = type;
	policy->types[newidx].num_attribs = 0;
	policy->types[newidx].aliases = NULL;	
	policy->types[newidx].attribs = NULL;
	policy->tree[AVL_TYPES].ptrs[newidx].left = -1;
	policy->tree[AVL_TYPES].ptrs[newidx].right = -1;
	policy->tree[AVL_TYPES].ptrs[newidx].height = 0;
	(policy->num_types)++;
	return newidx;
}

/*  return -1 for error, -2 if type already exists, the new tree head idx otherwise
 * if -2, then newidx will contain the idx of the existing record.
 */
int avl_insert_type(const int head, char *type, int *newidx, policy_t *policy)
{
	int newhead, tmpidx, cmpval;
	
	if(head < 0) { /* check for empty subtree */
		*newidx = add_type_entry(type, policy);
		return *newidx;
	}
	/* non-empty list */
	assert(head < policy->num_types);
	cmpval = type_compare(type, head, policy);
	if(cmpval > 0) {
		tmpidx = avl_insert_type(policy->tree[AVL_TYPES].ptrs[head].right,
			type, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_TYPES].ptrs[head].right = tmpidx;
			
		if(avl_check_balance(tmpidx, RIGHT, AVL_TYPES, policy)) {
			if(strcasecmp(type, policy->types[tmpidx].name) > 0) {
				newhead = avl_srr(head, policy->tree[AVL_TYPES].ptrs);
			}
			else {
				newhead = avl_drr(head, policy->tree[AVL_TYPES].ptrs);
			}
		}
		else {
			newhead = head;
		}
	}
	else if(cmpval < 0) {
		tmpidx = avl_insert_type(policy->tree[AVL_TYPES].ptrs[head].left,
			type, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_TYPES].ptrs[head].left = tmpidx;
		if(avl_check_balance(tmpidx, LEFT, AVL_TYPES, policy)) {
			if(strcasecmp(type, policy->types[tmpidx].name ) < 0) {
				newhead = avl_srl(head, policy->tree[AVL_TYPES].ptrs);
			}
			else {
				newhead = avl_drl(head, policy->tree[AVL_TYPES].ptrs);
			}
		}
		else {
			newhead = head;
		}	}
	else /* type already exists */ {
		*newidx = head;
		return -2;
	}
	policy->tree[AVL_TYPES].ptrs[newhead].height = avl_max_child(newhead, policy->tree[AVL_TYPES].ptrs) + 1;
	return newhead;
}


/*  return -1 for error, -2 if class already exists, the new tree head idx otherwise
 * if -2, then newidx will contain the idx of the existing record.
 */
int avl_insert_class(const int head, char *classname, int *newidx, policy_t *policy)
{
	int newhead, tmpidx, cmpval;
	
	if(head < 0) { /* check for empty subtree */
		if(check_class_array(policy) != 0)
		 	return -1;
		*newidx = policy->num_obj_classes;
		policy->obj_classes[*newidx].name = classname;
		policy->obj_classes[*newidx].common_perms = -1;
		policy->obj_classes[*newidx].num_u_perms = 0;
		policy->obj_classes[*newidx].u_perms = NULL;
		policy->tree[AVL_CLASSES].ptrs[*newidx].left = -1;
		policy->tree[AVL_CLASSES].ptrs[*newidx].right = -1;
		policy->tree[AVL_CLASSES].ptrs[*newidx].height = 0;
		(policy->num_obj_classes)++;
		return *newidx;
	}
	/* non-empty list */
	assert(head < policy->num_obj_classes);
	cmpval = strcasecmp(classname, policy->obj_classes[head].name);
	if(cmpval > 0) {
		tmpidx = avl_insert_class(policy->tree[AVL_CLASSES].ptrs[head].right,
			classname, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_CLASSES].ptrs[head].right = tmpidx;
			
		if(avl_check_balance(tmpidx, RIGHT, AVL_CLASSES, policy)) {
			if(strcasecmp(classname, policy->obj_classes[tmpidx].name) > 0) {
				newhead = avl_srr(head, policy->tree[AVL_CLASSES].ptrs);
			}
			else {
				newhead = avl_drr(head, policy->tree[AVL_CLASSES].ptrs);
			}
		}
		else {
			newhead = head;
		}
	}
	else if(cmpval < 0) {
		tmpidx = avl_insert_class(policy->tree[AVL_CLASSES].ptrs[head].left,
			classname, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_CLASSES].ptrs[head].left = tmpidx;
		if(avl_check_balance(tmpidx, LEFT, AVL_CLASSES, policy)) {
			if(strcasecmp(classname, policy->obj_classes[tmpidx].name ) < 0) {
				newhead = avl_srl(head, policy->tree[AVL_CLASSES].ptrs);
			}
			else {
				newhead = avl_drl(head, policy->tree[AVL_CLASSES].ptrs);
			}
		}
		else {
			newhead = head;
		}	}
	else /* type already exists */ {
		*newidx = head;
		return -2;
	}
	policy->tree[AVL_CLASSES].ptrs[newhead].height = avl_max_child(newhead, policy->tree[AVL_CLASSES].ptrs) + 1;
	return newhead;
}

/* return -1 for error, -2 if perm already exists, the new tree head idx otherwise
 * if -2, then newidx will contain the idx of the existing record.
 */
int avl_insert_perm(const int head, char *perm, int *newidx, policy_t *policy)
{
	int newhead, tmpidx, cmpval;
	
	if(head < 0) { /* check for empty subtree */
		if(check_perm_array(policy) != 0)
		 	return -1;
		*newidx = policy->num_perms;
		policy->perms[*newidx] = perm;
		policy->tree[AVL_PERMS].ptrs[*newidx].left = -1;
		policy->tree[AVL_PERMS].ptrs[*newidx].right = -1;
		policy->tree[AVL_PERMS].ptrs[*newidx].height = 0;
		(policy->num_perms)++;
		return *newidx;
	}
	/* non-empty list */
	assert(head < policy->num_perms);
	cmpval = strcasecmp(perm, policy->perms[head]);
	if(cmpval > 0) {
		tmpidx = avl_insert_perm(policy->tree[AVL_PERMS].ptrs[head].right,
			perm, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_PERMS].ptrs[head].right = tmpidx;
			
		if(avl_check_balance(tmpidx, RIGHT, AVL_PERMS, policy)) {
			if(strcasecmp(perm, policy->perms[tmpidx]) > 0) {
				newhead = avl_srr(head, policy->tree[AVL_PERMS].ptrs);
			}
			else {
				newhead = avl_drr(head, policy->tree[AVL_PERMS].ptrs);
			}
		}
		else {
			newhead = head;
		}
	}
	else if(cmpval < 0) {
		tmpidx = avl_insert_perm(policy->tree[AVL_PERMS].ptrs[head].left,
			perm, newidx, policy);
		if(tmpidx < 0)
			return tmpidx; /* error */
		else
			policy->tree[AVL_PERMS].ptrs[head].left = tmpidx;
		if(avl_check_balance(tmpidx, LEFT, AVL_PERMS, policy)) {
			if(strcasecmp(perm, policy->perms[tmpidx]) < 0) {
				newhead = avl_srl(head, policy->tree[AVL_PERMS].ptrs);
			}
			else {
				newhead = avl_drl(head, policy->tree[AVL_PERMS].ptrs);
			}
		}
		else {
			newhead = head;
		}	}
	else /* type already exists */ {
		*newidx = head;
		return -2;
	}
	policy->tree[AVL_PERMS].ptrs[newhead].height = avl_max_child(newhead, policy->tree[AVL_PERMS].ptrs) + 1;
	return newhead;
}




#endif /* USE_SORTING */





