/* Copyright (C) 2002-2003 Tresys Technology, LLC
 * see file 'COPYING' for use and warranty information */

/* 
 * Author: mayerf@tresys.com 
 */

/* seuser_db.c
 *
 * The user database functions for seuser.
 *
 */

#include <stdlib.h>
#include <tcl.h>
#include <tk.h>
#include <assert.h>
#include <time.h>
#include "seuser_db.h"
#include "seuser_tcl.h"

/* apol lib */
#include "../libapol/policy.h"

int copy_db_from_apol(user_db_t *db, policy_t *policy)
{
	user_item_t *user, *ptr;
	ta_item_t *taptr;
	seuser_data_t *data;
	int rt;
	
	data = NULL;
	if(db == NULL || policy == NULL) {
		return -1;
	}
	
	db->users.tail = NULL;
	if(get_first_user_ptr(policy) == NULL) {
		db->users.head = NULL;
		db->num_users = 0;
		return 0;
	}
	db->num_users = policy->rule_cnt[RULE_USER];
	
	for(ptr = get_first_user_ptr(policy); ptr != NULL; ptr = get_next_user_ptr(ptr)) {
		user = (user_item_t *)malloc(sizeof(user_item_t));
		if(user == NULL) {
			fprintf(stderr, "out of memory");
			return -1;
		}
		memset(user, 0, sizeof(user_item_t));		
		user->name = (char *)malloc(strlen(get_user_name_ptr(ptr)) + 1);
		if(user->name == NULL) {
			fprintf(stderr, "out of memory");
			return -1;
		}
		strcpy(user->name, get_user_name_ptr(ptr));
		
		for(taptr = get_user_first_role_ptr(ptr); taptr != NULL; taptr = get_user_next_role_ptr(taptr)) {
			ta_item_t *newta;
			newta = (ta_item_t *) malloc(sizeof(ta_item_t));
			if(newta == NULL) {
				fprintf(stderr, "out of memory");
				return -1;
			}
			newta->type = taptr->type;
			newta->idx = taptr->idx;
			rt = insert_ta_item(newta, &(user->roles));
			if(rt != 0) {
				fprintf(stderr, "problem inserting role in user");
				return -1;
			}
		}
		
		/* CHANGED */				
		if(use_old_login_contexts()) {
			data = (seuser_data_t *)malloc(sizeof(seuser_data_t));
			if(data == NULL) {
				fprintf(stderr, "out of memory");
				return -1;
			}
			data->flags = SEUSER_FLAG_NONE;
			user->data = data;
		} else {
			user->data = NULL;
		}

		rt = append_user(user, &(db->users));
		if(rt != 0) {
			fprintf(stderr, "problem inserting user in database ");
			return -1;
		}	
	}
	return 0;
}

/* check if user exists*/
bool_t seuser_does_user_exist(const char *name, user_db_t *db)
{
	user_item_t *ptr;
	if(name == NULL || db == NULL) {
		return FALSE;
	}
	
	for(ptr = db->users.head; ptr != NULL; ptr = ptr->next) {
		if(strcasecmp(name, ptr->name) == 0) {
			return TRUE;
		}
	}
	return FALSE;	
}



/* check if user exists, and if so return a pointer to its structure */
int seuser_get_user_by_name(const char *name, user_item_t **user, user_db_t *db)
{
	user_item_t *ptr;
	if(user == NULL || name == NULL || db == NULL) {
		return -1;
	}
	
	for(ptr = db->users.head; ptr != NULL; ptr = ptr->next) {
		if(strcasecmp(name, ptr->name) == 0) {
			*user = ptr;
			return 0;
		}
	}
	return -1;	
}



/* following return codes 
 * 
 *  0	is proper
 * -1	unexpected error
 *  1	a user role is not a valid role
 *  2	login role not valid for user
 *  3	login type not valid for login role
 *  4	cron role not valid for user
 *  5	cron type not valid for cron role
 *  6	does not have either login or cron context
 */
int is_proper_user_record(user_item_t *user, policy_t *policy)
{
	seuser_data_t *data;
	ta_item_t *taitem;
	
	/* various pointers */
	if(use_old_login_contexts()) {
		if(user == NULL || policy == NULL || user->name == NULL || user->roles == NULL ||
				user->data == NULL) {
			return -1;
		}
	} else if(user == NULL || policy == NULL || user->name == NULL || user->roles == NULL ) {
			return -1;
	}
		
	/* roles */
	for(taitem = user->roles; taitem != NULL; taitem = taitem->next) {
		if(taitem->type != IDX_ROLE) {
			return -1;
		}
		if(!is_valid_role_idx(taitem->idx, policy)) {
			return 1;
		}
	}

	data = user->data;
	
	/* Skip if user->data is NULL. This indicates that this is not a pre-Oct2002 system */	
	if(data != NULL) {
		if((use_old_login_contexts()) && !(data->flags & (SEUSER_FLAG_LOGIN_CONTEXT | SEUSER_FLAG_CRON_CONTEXT)))
			return 6;
		if(data->flags & SEUSER_FLAG_LOGIN_CONTEXT) {
			if(!does_role_use_type(data->contexts[DC_LOGIN].role, data->contexts[DC_LOGIN].type, policy)) {
				return 3;
			}
			if(!is_role_in_list(data->contexts[DC_LOGIN].role, user->roles)) {
				return 2;
			}
		}
		if(data->flags & SEUSER_FLAG_CRON_CONTEXT) {
			if(!does_role_use_type(data->contexts[DC_CRON].role, data->contexts[DC_CRON].type, policy)) {
				return 4;
			}
			if(!is_role_in_list(data->contexts[DC_CRON].role, user->roles)) {
				return 5;
			}
		}
	}
	return 0;
}


/* rename a user from the database 
 *  0  success
 *  1  user doesn't exist
 * -1  other error
 */
int seuser_rename_user(const char *oldname, const char *newname, user_db_t *db)
{
	user_item_t *ptr;
	char *tmp;
	
	if(oldname == NULL || newname == NULL || db == NULL)
		return -1;
		
	for(ptr = db->users.head; ptr != NULL; ptr = ptr->next) {
		if(strcasecmp(oldname, ptr->name) == 0) {
			break;
		}
	}
	if(ptr == NULL)
		return 1; /* user doens't exist */
	
	tmp = (char *)malloc(strlen(newname)+1);
	if(tmp == NULL) {
		fprintf(stderr, "out of memory");
		return -1;
	}
	strcpy(tmp, newname);
	assert(ptr->name != NULL);
	free(ptr->name);
	ptr->name = tmp;
		
	return 0;
}


/* remove a user from the database 
 *  0  success
 *  1  user doesn't exist
 * -1  other error
 */
int seuser_remove_user(const char *name, user_db_t *db)
{
	user_item_t *ptr, *ptr2;
	
	if(name == NULL || db == NULL)
		return -1;
		
	for(ptr = db->users.head, ptr2 = NULL; ptr != NULL; ptr2 = ptr, ptr = ptr->next) {
		if(strcasecmp(name, ptr->name) == 0) {
			if(ptr2 == NULL) { 
				/* case: head of list */
				assert(ptr == db->users.head);
				db->users.head = ptr->next;
			}
			else {
				assert(ptr2->next == ptr);
				ptr2->next = ptr->next;
				if(ptr == db->users.tail) {
					db->users.tail = ptr2;
				}
			}
			free_user(ptr);
			(db->num_users)--;
			return 0;
		}
	}
	
	return 1;
}

/* add or change a user record to the db 
 *  0  success
 *  1  user already exists in selinux policy and is newuser
 *  2  user doesnt exist and !newuser
 *  3  improperly formed user record
 * -1  other error
 */
int seuser_update_user(user_item_t *user, bool_t newuser, user_db_t *db, policy_t *policy) 
{
	user_item_t *tmpuser;
	int rt;
	
	if(user == NULL || db == NULL || policy == NULL)
		return -1;
	
	/* TODO: Handle new_SELinux_Version global flag variable being ON */	
	if(is_proper_user_record(user, policy) != 0)
		return 3;
	
	rt = seuser_get_user_by_name(user->name, &tmpuser, db);
	if(rt == 0 && newuser) {
		return 1; /* user already exists */
	}
	else if(rt != 0 && !newuser) {
		return 2; /* no user to update */
	}
	if(!newuser) {
		rt = seuser_remove_user(user->name, db);
		assert(rt != 1); /* can't possibly be that the user didn't exist at this point */
		if(rt < 0) {
			return -1;
		}
	}
	
	return(append_user(user, &(db->users)));	
}


/* get the default context for a given user name
 * results provided in malloc'd string names for role and type
 * return values:
 *  0 = success
 * -1 = no such user
 * -2 = invalid role
 * -3 = invalid type
 * -4 = other error
 *  1 = no context defined for the user for the "which" one asked for
 */
int seuser_get_user_context(const char *user, int which, char **role, char **type, user_db_t *db, policy_t *policy)
{
	int rt; 
	user_item_t *uitem;
	char *result;
	seuser_data_t *data;
	
	if(user == NULL || role == NULL || type == NULL || db == NULL || which < DC_LOGIN 
			|| which > DC_CRON || policy == NULL) {
				
		return -4;
	}
	
	rt = seuser_get_user_by_name(user, &uitem, db);
	if(rt != 0) {
		return -1;		
	}
	
	data = uitem->data;
	if(which == DC_LOGIN) {
		if(! (data->flags & SEUSER_FLAG_LOGIN_CONTEXT) ) {
			return 1;
		}
	}
	else /* which == DC_CRON */ {
		if(! (data->flags & SEUSER_FLAG_CRON_CONTEXT) ) {
			return 1;
		}
	}
	
	/* role */
	rt = get_role_name(data->contexts[which].role, &result, policy);
	if(rt != 0)
		return -2;
		
	*role = result;
	
	/* type */
	
	rt = get_type_name(data->contexts[which].type, &result, policy);
	if(rt != 0) {
		free(*role);
		return -3;
	}
	*type = result;
	
	return 0;	
}


/* write out the users file 
 * fp presumed to be opened for write access
 */
int seuser_write_user_file(user_db_t *db, policy_t *policy, FILE *fp)
{
	user_item_t *ptr;
	ta_item_t *taptr;
	char *name;
	int rt;
	time_t ltime;
	
	if(db == NULL || policy == NULL || fp == NULL)
		return -1;
	
	time(&ltime);
	rt = fprintf(fp, "# seuser\n# This file created automatically by seuser on %s\n", ctime(&ltime));	
	if(rt < 0)
		return -1;
	rt = fprintf(fp, "#\n# user file\n\n");
	if(rt < 0)
		return -1;

	for(ptr = db->users.head; ptr != NULL; ptr = ptr->next) {
		rt = fprintf(fp, "user %s roles { ", ptr->name);
		if(rt < 0)
			return -1;
		for(taptr = ptr->roles; taptr != NULL; taptr = taptr->next) {
			assert(taptr->type == IDX_ROLE);
			rt = get_role_name(taptr->idx, &name, policy);
			if(rt != 0) {
				fprintf(stderr, "Problem looking up role name in seuser_write_user_file");
				return -1;
			}
			rt = fprintf(fp, "%s ", name);
			free(name);
			if(rt < 0) 
				return -1;
		}
		rt = fprintf(fp, "} ;\n");
		if(rt < 0)
			return -1;
	}
	return 0;
}


/* write out the default context files 
 * fps's presumed to be opened for w+
 */
int seuser_write_default_context(int which, user_db_t *db, policy_t *policy, FILE *fp)
{
	int rt;
	time_t ltime;
	seuser_data_t *data;
	user_item_t *ptr;
	char *role, *type;
		
	if(which < 0 || which > DC_MAX || db == NULL || policy == NULL || fp == NULL)
		return -1;
		
	time(&ltime);
	rt = fprintf(fp, "# seuser\n# This file created automatically by seuser on %s\n", ctime(&ltime));		
	if(rt < 0) 
		return -1;
	if(which == DC_LOGIN) {
		rt = fprintf(fp, "#\n# default_context file (default user login context)\n\n");
	}
	else {
		rt = fprintf(fp, "#\n# cron_context file (default user cron context)\n\n");
	}
	if(rt < 0)
		return -1;
		
	for(ptr = db->users.head; ptr != NULL; ptr = ptr->next) {
		data = ptr->data;
		assert(data != NULL);
		if((which == DC_LOGIN && (data->flags & SEUSER_FLAG_LOGIN_CONTEXT)) ||
		   (which == DC_CRON  && (data->flags & SEUSER_FLAG_CRON_CONTEXT))) {
			rt = get_role_name(data->contexts[which].role, &role, policy);
			if(rt != 0) {
				fprintf(stderr, "Problem looking up role name in seuser_write_default_context");
				return -1;
			}
			rt = get_type_name(data->contexts[which].type, &type, policy);
			if(rt != 0) {
				free(role);
				fprintf(stderr, "Problem looking up type name in seuser_write_default_context");
				return -1;
			}
			rt = fprintf(fp, "%s:%s:%s\n", ptr->name, role, type);
			free(role);
			free(type);
			if(rt < 0) 
				return -1;
		}
	}
	
	return 0;
}


/* free all memory assoicated with a user database */
int free_user_db(user_db_t *db)
{
	user_item_t *ptr;
	
	if(db == NULL)
		return -1;
		
	for(ptr = db->users.head; ptr != NULL; ptr = ptr->next) {
		free_user(ptr);		
	}
	db->num_users = 0;
	db->users.head = NULL;
	db->users.tail = NULL;
		
	return 0;
}


/* copies a given user record into a new structure 
 *  0  success
 *  1  no copy; user doesn't exist
 * -1  other error
 */

int seuser_copy_user(const char *name, user_item_t **uitem, user_db_t *db)
{
	int rt, i;
	user_item_t *ptr;
	ta_item_t *taptr;
	seuser_data_t *datacur, *datacpy;
	
	if(name == NULL || uitem == NULL || db == NULL)
		return -1;
		
	rt = seuser_get_user_by_name(name, &ptr, db);
	if(rt != 0)
		return 1 /* no such user */;
	assert(ptr != NULL);
		
	*uitem = (user_item_t *)malloc(sizeof(user_item_t));
	if(*uitem == NULL) {
		fprintf(stderr, "out of memory");
		return -1;
	}
	assert(ptr->name != NULL);
	(*uitem)->name = (char *)malloc(strlen(get_user_name_ptr(ptr)) + 1);
	if((*uitem)->name == NULL) {
		fprintf(stderr, "out of memory");
		return -1;
	}
	strcpy((*uitem)->name, get_user_name_ptr(ptr));
	
	
	for(taptr = get_user_first_role_ptr(ptr); taptr != NULL; taptr = get_user_next_role_ptr(taptr)) {
		ta_item_t *newta;
		newta = (ta_item_t *) malloc(sizeof(ta_item_t));
		if(newta == NULL) {
			fprintf(stderr, "out of memory");
			return -1;
		}
		memset(newta, 0, sizeof(ta_item_t));
		newta->type = taptr->type;
		newta->idx = taptr->idx;
		rt = insert_ta_item(newta, &((*uitem)->roles));
		if(rt != 0) {
			fprintf(stderr, "problem inserting role in user");
			return -1;
		}
	}	
	datacur = ptr->data;
	/* assert(datacur != NULL); */
	
	/* CHANGED. If the data item is NULL then this means it is not the pre-Oct2002 login version */
	if(datacur != NULL) {
		datacpy = (seuser_data_t *)malloc(sizeof(seuser_data_t));
		if(datacpy == NULL) {
			fprintf(stderr, "out of memory");
			return -1;
		}
		
		datacpy->flags = datacur->flags;
		for(i = DC_LOGIN; i < DC_MAX; i++) {
			datacpy->contexts[i].role = datacur->contexts[i].role;
			datacpy->contexts[i].type = datacur->contexts[i].type;	
		}
	}
	
	return 0;	
}

