/* Copyright (C) 2002-2003 Tresys Technology, LLC
 * see file 'COPYING' for use and warranty information */

/* 
 * Author: mayerf@tresys.com, dac@tresys.com
 */

/* seuser_tcl.c
 *
 */ 

#include <unistd.h>
#include <tcl.h>
#include <tk.h>
#include <assert.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include "seuser_tcl.h"
#include "seuser_db.h"

/* apol lib */
#include "../libapol/policy.h"
#include "../libapol/apol_tcl.h"
#include "../libapol/util.h"

/* global vars holding config info */
char *policy_conf = NULL;
char *default_context = NULL;
char *cron_context = NULL;
char *user_file = NULL;
char *policy_dir = NULL;

/* global vars for login verions checking */
bool_t login_version_is_set = FALSE;
bool_t system_uses_old_login_contexts = FALSE;
char *login_version_str = NULL;

/* temp file for make output */
char *tmpmakeout = NULL;

/* assumed defines */
#define MAKEFILE		"Makefile"
/* With the Aug 2002 policy mgt patch, we use intall rather than
 * policy.conf as our "refresh" target.  This will make policy.conf
 * and install the changes to both intalled policy forms (policy.conf)
 * and binary policy file.
 */
#define POLICY_CONF_TARGET	"install"
#define INSTALL_TARGET		"load"

/* database */
user_db_t db;
bool_t db_init = FALSE;
bool_t conf_init = FALSE;

/* policy */
extern policy_t *policy; /* from apolicy_parse.y, our policy DB */



/*******************************************************************
 *
 * Support functions
 *
 *******************************************************************/

static int call_make(const char *target, const char *output)
{
	char *make_str;
	int rt;
	
	if(!conf_init)
		return -1;
	assert(policy_dir != NULL);
	assert(target != NULL);
	assert(output != NULL);
	
     	make_str = (char *)malloc(strlen(policy_dir) + strlen(MAKEFILE) + 
     		strlen(target) + strlen(output) + 50);
     	if(make_str == NULL) {
     		fprintf(stderr, "out of memory\n");
     		return -1;
     	}
     	sprintf(make_str, "make -f %s -C %s %s > %s 2>&1", MAKEFILE, policy_dir, target, output);
	
     	rt = system(make_str);	
     	if(rt != 0) {
     		fprintf(stderr, "Error: Make string: %s.", make_str);
     		free(make_str);
     		return -1;
     	} 
	free(make_str);
	
	return 0;
}

/* re-make policy.conf */
static int remake_policy_conf(Tcl_Interp * interp)
{
	int rt;
	if(!conf_init)
		return -1;	
	assert(policy_dir != NULL);

	rt = call_make(POLICY_CONF_TARGET, tmpmakeout);	
     	if(rt != 0) {
     		Tcl_AppendResult(interp, "Error re-making policy.conf file.", (char *) NULL);
     		return TCL_ERROR;
     	}     	
	
	return TCL_OK;
}


/* provide a pointer to the user db is init'd */
user_db_t * get_user_db(void)
{
	/* db_init and db are globals from this file */
	if(!db_init)
		return NULL;
	return &db;
}

/* provide pointer to policy database */
policy_t *get_seuser_policy(void)
{
	return policy;
}

/* "commit" means write access to 3 files, return values are:
 *
 *  0	permission allowed
 *  1	permission denied to users
 *  2 	permission denied to default_context
 *  3	permission denied to cron_context
 * -1	error
 *
 * the file names are stoed in global vars
 */
static int check_commit_perm(void)
{
	int rt;
	if(!conf_init)
		return -1;	
	/* if user_file does exist, check whether we can in fact create it */
	rt = access(user_file, F_OK);
	if(rt == 0) 
		rt = access(user_file, W_OK);
	else
		/* FIX: For now we're assuming (yuk!) that the users file is in the policy_dir 
		 * so if the users files doesn't exist, check that we can indeed create it in the
		 * policy_dir */
		rt = access(policy_dir, W_OK);
	if(rt != 0)
		return 1;

	if(use_old_login_contexts()) {		
		rt = access(default_context, W_OK);
		if(rt != 0)
			return 2;
			
		rt = access(cron_context, W_OK);
		if(rt != 0)
			return 3;
	}
	
	return 0;
}

/* install new policy */
static int reinstall_policy(Tcl_Interp *interp)
{
	int rt;
	if(!conf_init)
		return -1;	
	assert(policy_dir != NULL);

	rt = call_make(INSTALL_TARGET, tmpmakeout);	
     	if(rt != 0) {
     		Tcl_AppendResult(interp, "Error re-making policy.conf file.", (char *) NULL);
     		return TCL_ERROR;
     	}     	
	
	return TCL_OK;
}

/* frees conf info and */
int free_conf_info (void)
{
	if(!conf_init)
		return 0;
	free(policy_conf); policy_conf = NULL;
	free(user_file); user_file = NULL;
	free(policy_dir); policy_dir = NULL;
	
	/* Freeing memory to a variable that doesn't exist will cause an error. 
	   So, here we add our version check */
	if(use_old_login_contexts()) {
		free(default_context); default_context = NULL;
		free(cron_context); cron_context = NULL;
	}
	
	conf_init = FALSE;
	
	return 0;
}

/* get the value for the config var provided from the config file provided */
static char *get_config_var(char *var, FILE *fp)
{
	#define LINE_SZ 1024
	char line[LINE_SZ], t1[LINE_SZ], t2[LINE_SZ], *result = NULL;
		
	if(var == NULL)
		return NULL;
		
	rewind(fp);
	while(fgets(line, LINE_SZ, fp) != NULL) {
		if(line[0] == '#' || sscanf(line, "%s %s", t1, t2) != 2 || strcasecmp(var, t1) != 0) {
			continue;
		}
		else {
			result = (char *)malloc(strlen(t2) + 1);
			if(result == NULL) {
				fprintf(stderr, "out of memory\n");
				return NULL;
			} 
			else {
				strcpy(result, t2);
				return result;
			}
		}
	}                                           
	return NULL;
}

int se_mktmpfile() 
{
	/* get a temp file name if this is first time initializing
	 * we use this file name for our temporary make output */
	/* tmpmakeout is a global variable */
	if(tmpmakeout == NULL) {
		tmpmakeout = tempnam("/tmp", "seuser_tmp.");
	}
	return 0;
}

/* read the seuser config info */
int read_conf_info(Tcl_Interp * interp)
{
	char *dir = NULL, *full_config = NULL, tbuf[1024];
	FILE *fp;
	int rt;
	
	if(conf_init)
		return TCL_OK;
	
	/* we need to find the conf file.  We use the same logic
	 * as apol uses, and generally expect the config file for	
	 * seuser to be stored with the apol stuff.  Thus we use
	 * the unfortunately named function find_tcl_script() in
	 * ./libapol/apol_tcl.c, giveing our conf file name to find 
	 * directory in which we expect the conf file.
	 */
	dir = find_tcl_script(CONFIG_FILE);
	if(dir == NULL) {
		sprintf(tbuf, "Could not locate seuser config file: %s\nEnsure the file is in %s, or that the environ variable %s is set.\n", 
			CONFIG_FILE, APOL_INSTALL_DIR, APOL_ENVIRON_VAR_NAME);
		Tcl_AppendResult(interp, tbuf, (char *) NULL);
		return TCL_ERROR;
	}
     	full_config = (char *)malloc(strlen(dir) + strlen(CONFIG_FILE) + 3);
     	if(full_config == NULL) {
     		Tcl_AppendResult(interp, "out of memory\n", (char *) NULL);
     		free(dir);
     		return TCL_ERROR;
     	}
     	sprintf(full_config, "%s/%s", dir, CONFIG_FILE);
     	free(dir);

	if((fp = fopen(full_config, "r")) == NULL) {
		sprintf(tbuf, "cannot open policy file: %s ", full_config);
		Tcl_AppendResult(interp, tbuf, (char *) NULL);
		free(full_config);
		return TCL_ERROR;
	}	
	free(full_config);
	
	
	policy_conf = get_config_var("policy.conf", fp);
	if(policy_conf == NULL) {
     		Tcl_AppendResult(interp, "Could not find config information for policy.conf\n", (char *) NULL);
     		goto conf_err;
     	}
	rt = access(policy_conf, R_OK);
	if(rt != 0) {
		sprintf(tbuf, "Could not read policy.conf file: %s\n", policy_conf);
     		Tcl_AppendResult(interp, tbuf, (char *) NULL);
     		goto conf_err;
     	}
     			
	policy_dir = get_config_var("policy_dir", fp);
	if(policy_dir == NULL) {
     		Tcl_AppendResult(interp, "Could not find config information for policy_dir\n", (char *) NULL);
     		goto conf_err;
     	}
	rt = access(policy_dir, R_OK);
	if(rt != 0) {
		sprintf(tbuf, "Could not read policy directory: %s\n", policy_dir);
     		Tcl_AppendResult(interp, tbuf, (char *) NULL);
     		goto conf_err;
     	}  
     	   	
     	user_file = get_config_var("user_file", fp);
	if(user_file == NULL) {
     		Tcl_AppendResult(interp, "Could not find config information for user_file\n", (char *) NULL);
     		goto conf_err;
     	}
	/* users file may not exist which is ok, so we won't check read access */   
	
	if(use_old_login_contexts()) {
	     	default_context = get_config_var("default_context", fp);
		if(default_context == NULL) {
	     		Tcl_AppendResult(interp, "Could not find config information for default_context\n", (char *) NULL);
	     		goto conf_err;
	     	}	
	     	rt = access(default_context, R_OK);
		if(rt != 0) {
			sprintf(tbuf, "Could not read default_context file: %s\n", default_context);
	     		Tcl_AppendResult(interp, tbuf, (char *) NULL);
	     		goto conf_err;
	     	} 
	     	
	     	cron_context = get_config_var("cron_context", fp);
		if(cron_context == NULL) {
	     		Tcl_AppendResult(interp, "Could not find config information for cron_context\n", (char *) NULL);
	     		goto conf_err;
	     	}
	     	rt = access(cron_context, R_OK);
		if(rt != 0) {
			sprintf(tbuf, "Could not read cron_context file: %s\n", cron_context);
	     		Tcl_AppendResult(interp, tbuf, (char *) NULL);
	     		goto conf_err;
	     	} 
     	}
     		
	fclose(fp);
	conf_init = TRUE;
	return TCL_OK;
	
conf_err:
	fclose(fp);
	free_conf_info();
	return TCL_ERROR;
}



/* reads user default contexts info from provided file for provided default context */
int get_user_contexts(int which, user_db_t *db, policy_t *policy, FILE *fp)
{
	char line[LINE_SZ], user[LINE_SZ], role[LINE_SZ], type[LINE_SZ];
	bool_t warning = FALSE;
	int rt, tatype, i;
	user_item_t *uitem;
	seuser_data_t *data;

	if(user == NULL || fp == NULL || which < 0 || which > DC_MAX || policy == NULL) {
		return -1;
	}
	rewind(fp);
	while(fgets(line, LINE_SZ, fp) != NULL) {
		/* changing ':' to ' ' for delimeters for sscanf */
		for(i = 0; line[i] != '\0';  i++) {
			if(line[i] == ':') {
				line[i] = ' ';
			}
		}
		if(line[0] == '#' || sscanf(line, "%s %s %s", user, role, type) != 3) {
			if(line[0] != '#') {
				warning = TRUE;
			}
			continue;
		}
		rt = seuser_get_user_by_name(user, &uitem, db);
		if(rt != 0) {
			/* if no matching user in our db, we warn and move on! */
			warning = TRUE;
			continue;
		}
		data = uitem->data;
		if(which == DC_LOGIN) {
			data->flags |= SEUSER_FLAG_LOGIN_CONTEXT;
		}
		else {
			data->flags |= SEUSER_FLAG_CRON_CONTEXT;
		}
		
		rt = get_role_idx(role, policy);
		if(rt < 0) {
			warning = TRUE;
			data->contexts[which].role = -1;
		}
		else {
			data->contexts[which].role = rt;
		}
		rt = get_type_or_attrib_idx(type, &tatype, policy);
		if(rt < 0 || tatype != IDX_TYPE) {
			warning = TRUE;
			data->contexts[which].type = -1;
		}
		else {
			data->contexts[which].type = rt;
		}
	}
	
	if(warning)
		return 1;
	else
		return 0;
	
}

/* read a given default context file and update the database 
 * default_context and cron_context are globals
 */
int read_default_contexts(int which, const char *file, user_db_t *db, policy_t *policy, Tcl_Interp *interp)
{
	FILE *fp;
	char *cntx, tbuf[512];
	int rt;
	bool_t warning = FALSE;
	
	if(db == NULL || which < 0 || which >= DC_MAX || policy == NULL || file == NULL) {
		return -1;
	}
	
	if(which == DC_LOGIN) {
		assert(default_context != NULL);
		cntx = default_context;
	}
	else {
		assert(cron_context != NULL);
		cntx = cron_context;
	}
	if((fp = fopen(cntx, "r")) == NULL) {
		sprintf(tbuf, "cannot open context file: %s", cntx);
		Tcl_AppendResult(interp, tbuf, (char *) NULL);
		return -1;
	}
	
	if((rt = get_user_contexts(which, db, policy, fp)) < 0) {
		Tcl_AppendResult(interp, "problem getting user context", (char *) NULL);
		fclose(fp);
		return -1;
	}
	if(rt == 1) 
		warning = TRUE;

	fclose(fp);
	if(warning) {
		return 1;
	}
	else {
		return 0;	
	}
}

int validate_given_role(char *userName, char *givenRole)
{
	int i, rt = 0;
	char *name = NULL;
	user_item_t *user;
	
	rt = seuser_get_user_by_name(userName, &user, &db);
	if(rt != 0) {
		return -1;		
	}
	
	for(i = 0; get_role_name(i, &name, policy) == 0; i++) {
		if(strcmp(name, givenRole) == 0) {
			free(name);
			return 1;
		}
		free(name);
	}
	
	return 0;
}

/* Check if the default_contexts file exists. If it doesn't then 
 * set the global version string var and global boolean var.
 */
/* Returns 0 if ok, returns -1 if error allocating memory	*/
int set_login_version() 
{	
	int rt;
	if (login_version_is_set) {
		return 0;
	} else {
	}
	/* login_version_str is a global variable! */
	
	rt = access("/etc/security/default_contexts", F_OK);
	if(rt != 0) {
		login_version_str = (char*) malloc(sizeof OLD_LOGIN_STR);
		if(login_version_str == NULL) {
			fprintf(stderr, "out of memory");
			return -1;
		}
		strcpy(login_version_str, OLD_LOGIN_STR);
		system_uses_old_login_contexts = TRUE;
     	} 
     	else {
 		login_version_str = (char*) malloc(sizeof NEW_LOGIN_STR);
		if(login_version_str == NULL) {
			fprintf(stderr, "out of memory");
			return -1;
		}
		strcpy(login_version_str, NEW_LOGIN_STR);    		
		system_uses_old_login_contexts = FALSE;
     	}
     	
     	login_version_is_set = TRUE;
     	
	return 0;
}

/* Public function used to check if the system version is pre-Oct2002. 	*/
/* Returns the value of the global boolean variable. In this way     	*/
/* the global variable is encapsulated behind this interface.		*/
int use_old_login_contexts() 
{		
	return system_uses_old_login_contexts;
}

/* determine user type and return appropriate string.  Caller must 
 * free() type_string
 */
char * determine_user_type(const char* user)
{
	char *result;
	int rt;
	user_item_t *uitem;
	
	/* alloc string..caller must free */
	result = (char *)malloc(SYSUSER_MAX_STR_SZ);
	if(result == NULL) {
		fprintf(stderr, "out of memory\n");
		return NULL;
	}
	
	if(strcmp(user, SPECIAL_USER_SYSTEM) == 0 || strcmp(user, SPECUAL_USER_USER) == 0) {
		strcpy(result, SYSUSER_SPECIAL_STR);
	}
	else {
		rt = seuser_get_user_by_name(user, &uitem, &db);
		if(rt != 0) {
			rt = seuser_get_user_by_name(SPECUAL_USER_USER, &uitem, &db);
			if(rt != 0) 
				strcpy(result, SYSUSER_UNDEFINED_STR);
			else
				strcpy(result, SYSUSER_GENERIC_STR);
		}
		else {
			strcpy(result, SYSUSER_DEFINED_STR);
		}
	}
	return result;
}

/*******************************************************************
 *
 * TCL Commands
 *
 *******************************************************************/

int Seuser_GetVersion(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	Tcl_AppendResult(interp, SEUSER_VERSION_STRING, (char *) NULL);
	return TCL_OK;
}

int Seuser_Get_Login_Version(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) 
{	
	set_login_version();
	Tcl_AppendResult(interp, login_version_str, (char *) NULL);

	return TCL_OK;
}

int Seuser_Use_Old_Login_Contexts(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]) 
{	
	set_login_version();
	if(use_old_login_contexts()) {
		Tcl_AppendResult(interp, "1", (char *) NULL);
	}
	else {
		Tcl_AppendResult(interp, "0", (char *) NULL);
	}
	return TCL_OK;
}

int Seuser_Exit(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	if(argc > 1) {
		Tcl_AppendResult(interp, "Seuser_Exit: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(tmpmakeout != NULL) {
		remove(tmpmakeout);
	}
	return TCL_OK;
}

int Seuser_GetTmpMakeFileName(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	if(argc > 1) {
		Tcl_AppendResult(interp, "Seuser_GetTmpMakeFileName: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(tmpmakeout == NULL) {
		Tcl_AppendResult(interp, "Not initialized: Seuser_InitUserdb must be called first", (char *) NULL);
		return TCL_ERROR;
	}
	Tcl_AppendResult(interp, tmpmakeout, (char *) NULL);
	return TCL_OK;
}

/* initialize User database */
int Seuser_InitUserdb(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	int ac, rt;
	char *av[2];
	
	if(argc > 1) {
		Tcl_AppendResult(interp, "Seuser_InitUserdb: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	
	rt = set_login_version();
	if(rt != 0) {
     		Tcl_AppendResult(interp, "Error setting login version.", (char *) NULL);
     		return TCL_ERROR;
     	}
     	
     	/* read conf info if not already set; policy_conf is a global */
     	if(policy_conf == NULL) {
		if(read_conf_info(interp) != 0) {
			return TCL_ERROR;
		}
	}
	
	se_mktmpfile();
		
	/* open policy.conf */
	av[0] = NULL;
	av[1] = policy_conf;
	ac = 2;
	
	if(Apol_OpenPolicy(NULL, interp, ac, av) != TCL_OK) {
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp, "Problem with policy.conf file.\nMake sure policy is valid using checkpolicy -d\n or that the config file points to correct file", (char *) NULL);
		return TCL_ERROR;
	}	
	
	/* create user database; db is global.  */
	rt = copy_db_from_apol(&db, policy);
	if(rt != 0) {
		Tcl_AppendResult(interp, "Could not copy user database from policy");
		return TCL_ERROR;
	}
	db_init = TRUE;
	
	if(use_old_login_contexts()) {
		rt = read_default_contexts(DC_LOGIN, default_context, &db, policy, interp);
		if(rt < 0) {
			Tcl_AppendResult(interp, "Problem reading login contexts");
			return TCL_ERROR;
		}
		else if(rt == 1) {
			//TODO: warnings
		}
		
		/* get the cron contexts */
		rt = read_default_contexts(DC_CRON, cron_context, &db, policy, interp);
		if(rt < 0) {
			Tcl_AppendResult(interp, "Problem reading cron contexts");
			return TCL_ERROR;
		}
		else if(rt == 1) {
			//TODO: warnings
		}
	}
		
	return TCL_OK;
}


/* frees the current user data base, and uninitialzes the tool configuration 
 * One can then safely call Seuser_InitUserdb again to re-initialze*/
int Seuser_CloseDatabase(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	int rt;
	
	if(argc > 1) {
		Tcl_AppendResult(interp, "Seuser_CloseDatabase: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	
	rt = Apol_ClosePolicy( clientData, interp, argc, argv);
	if(rt != TCL_OK)
		return rt;
	
	if(db_init) {
		rt = free_user_db(&db);
		if(rt != 0) {
			Tcl_AppendResult(interp, "Problem free user database", (char *) NULL);
			return TCL_ERROR;
		}
		db_init = FALSE;
	}
	
	free_conf_info();
	
	return TCL_OK;
}



/* get db list of users */
int Seuser_GetSeUserNames(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	user_item_t *ptr;
	char *name;

	Tcl_ResetResult(interp);
	if(argc > 1) {
		Tcl_AppendResult(interp, "Seuser_GetSeUserNames: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(!db_init) {
		Tcl_AppendResult(interp,"User database not initialized!", (char *) NULL);
		return TCL_ERROR;
	}
	
	for(ptr = db.users.head; get_user_name(ptr, &name) == 0; ptr = ptr->next) {
		Tcl_AppendElement(interp, name);
		free(name);
	}
	
	return TCL_OK;
}

/* Takes a user name, and determines whether the cooresponding SE Linux user
 * record as existing in the database is "valid".  If not valid, and error
 * with appropriate message will be returned.  An error is also returned
 * if the user name is not currently defined in the database
 */
int Seuser_IsUserValid(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	int rt;
	char *user, tmpbuf[256];
	user_item_t *uitem;

	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(!db_init) {
		Tcl_AppendResult(interp,"User database not initialized!", (char *) NULL);
		return TCL_ERROR;
	}	
	assert(policy != NULL);
	
	user = argv[1];
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "User name string too large", (char *) NULL);
		return TCL_ERROR;
	}
	
	rt = seuser_get_user_by_name(user, &uitem, &db);
	if(rt != 0) {
		Tcl_AppendResult(interp,"User does not exist in database!", (char *) NULL);
		return TCL_ERROR;
	}
	rt = is_proper_user_record(uitem, policy);
	
	switch (rt) {
	case 0: 
		return TCL_OK;
		break;
	case 1: 
		sprintf(tmpbuf, "A user role is not a valid role");
		break;
	case 2:
		sprintf(tmpbuf, "The login role is not valid for the user");
		break;
	case 3:
		sprintf(tmpbuf, "The login type is not valid for the login role");
		break;
	case 4:
		sprintf(tmpbuf, "The cron role is not valid for the user");
		break;
	case 5:
		sprintf(tmpbuf, "The cron type is not valid for the cron role");
		break;
	case 6:
		sprintf(tmpbuf, "The user must have either a login or cron context");
		break;		
	default:
		sprintf(tmpbuf, "Unknown error");
		break;
	}
	Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
	return TCL_ERROR;
}


/* remove a user from the database 
 * argv[1] username
 */

int Seuser_RemoveUser(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	int rt;
	char tbuf[256];
	
	if(argc != 2) {
		Tcl_AppendResult(interp, "wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(!db_init) {
		Tcl_AppendResult(interp,"User database not initialized!", (char *) NULL);
		return TCL_ERROR;
	}
	
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "User name string too large", (char *) NULL);
		return TCL_ERROR;
	}
	rt = seuser_remove_user(argv[1], &db);
	if(rt == 1) {
		sprintf(tbuf, "User (%s) does not exist in the database", argv[1]);
		Tcl_AppendResult(interp, tbuf, (char *) NULL);
		return TCL_ERROR;
	}
	else if(rt != 0) {
		Tcl_AppendResult(interp, "Error removing the user from the database");
		return TCL_ERROR;
	}		
	
	return TCL_OK;
}

/* re-make policy.conf file */
int Seuser_RemakePolicyConf(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	if(argc != 1) {
		Tcl_AppendResult(interp, "Seuser_RemakePolicyConf: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	assert(policy_dir != NULL);
	
	/* TODO: Deal with an error response, displaying the make output */
	return(remake_policy_conf(interp));
}

/* re-install policy */
int Seuser_ReinstallPolicy(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
	if(argc != 1) {
		Tcl_AppendResult(interp, "Seuser_ReinstallPolicy: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	assert(policy_dir != NULL);
	
	/* TODO: Deal with an error response, displaying the make output */
	return(reinstall_policy(interp));
}

/* change an existing or add a new user 
 * argv[1]	command ("add" or "change")
 * argv[2]	username
 * argv[3]	roles (a TCL list string)
 * argv[4]	default_login (bool)
 * argv[5]	login_role
 * argv[6]	login_type
 * argv[7]	default_cron (bool)
 * argv[8]	cron_role
 * argv[9]	cron_type
 */
int Seuser_EditUser(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	char **roles, *cmd, *user, *login_role, *login_type, *cron_role, *cron_type, tmpbuf[512];
	int rt, num_roles, i, idx;
	bool_t default_login, default_cron, new_user;
	user_item_t *uitem;
	seuser_data_t *data;
	
	data = NULL;		
	if(argc != 10) {
		Tcl_AppendResult(interp, "Seuser_EditUser: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(!db_init) {
		Tcl_AppendResult(interp,"User database not initialized!", (char *) NULL);
		return TCL_ERROR;
	}
	assert(policy != NULL);
	
	
	/* parse all the parameters */
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "Command name string too large", (char *) NULL);
		return TCL_ERROR;
	}
	cmd = argv[1];
	if(strcmp(cmd, "add") == 0) {
		new_user = TRUE;
	}
	else if(strcmp(cmd, "change") == 0) {
		new_user = FALSE;
	}
	else {
		sprintf(tmpbuf, "Invalid command (%s), must be add or change", cmd);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;
	}
	
	if(!is_valid_str_sz(argv[2])) {
		Tcl_AppendResult(interp, "User name string too large", (char *) NULL);
		return TCL_ERROR;
	}
	user = argv[2];
	if(user == NULL || (strcmp(user, "") == 0)) {
		Tcl_AppendResult(interp, "No user name provided", (char *) NULL);
		return TCL_ERROR;
	}
	
	/* First, make sure the there is at least one role defined for the user. */
	rt = Tcl_SplitList(interp, argv[3], &num_roles, &roles);
	if(rt != TCL_OK) {
		return rt;
	}
	if(num_roles < 1) {
		Tcl_AppendResult(interp, "Users must have at least one role defined for them.", (char *) NULL);
		Tcl_Free((char *)roles);
		return TCL_ERROR;
	}
	
	default_login = getbool(argv[4]);
	default_cron = getbool(argv[7]);
	
	/* If system version is old login style, make sure context information is provided for the user. */
	if((use_old_login_contexts()) && (!default_login && !default_cron)) {
		Tcl_AppendResult(interp, "One or both of the default context must be provided", (char *) NULL);
		return TCL_ERROR;
	}
	if(default_login) {
		if(!is_valid_str_sz(argv[5])) {
			Tcl_AppendResult(interp, "Login role name string too large", (char *) NULL);
			return TCL_ERROR;
		}
		login_role = argv[5];
		if(!is_valid_str_sz(argv[6])) {
			Tcl_AppendResult(interp, "Login type  name string too large", (char *) NULL);
			return TCL_ERROR;
		}
		login_type = argv[6];
	}
	else {
		login_role = NULL;
		login_type = NULL;
	}
	if(default_cron) {
		if(!is_valid_str_sz(argv[8])) {
			Tcl_AppendResult(interp, "Cron role name string too large", (char *) NULL);
			return TCL_ERROR;
		}
		cron_role = argv[8];
		if(!is_valid_str_sz(argv[9])) {
			Tcl_AppendResult(interp, "Cron type name string too large", (char *) NULL);
			return TCL_ERROR;
		}
		cron_type = argv[9];
	}
	else {
		cron_role = NULL;
		cron_type = NULL;
	}
	
	/* assemble the new/changed user record */
	uitem = (user_item_t *)malloc(sizeof(user_item_t));
	memset(uitem, 0, sizeof(user_item_t));		
	if(uitem == NULL) {
		fprintf(stderr, "out of memory");
		Tcl_AppendResult(interp, "out of memory", (char *) NULL);
		Tcl_Free((char *)roles);
		return TCL_ERROR;
	}
		/* name */
	(uitem)->name = (char *)malloc(strlen(user) + 1);
	if((uitem)->name == NULL) {
		fprintf(stderr, "out of memory");
		Tcl_AppendResult(interp, "out of memory", (char *) NULL);
		free_user(uitem);
		Tcl_Free((char *)roles);
		return TCL_ERROR;
	}
	strcpy((uitem)->name, user);
		/* roles */
	for(i = 0; i < num_roles; i++) {
		ta_item_t *newta;
				
		idx = get_role_idx(roles[i], policy);
		if(idx < 0) {
			sprintf(tmpbuf, "Role %d (%s) is not a valid role name", i+1, roles[i]);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			Tcl_Free((char *)roles);
			return TCL_ERROR;
		}
		newta = (ta_item_t *) malloc(sizeof(ta_item_t));
		if(newta == NULL) {
			fprintf(stderr, "out of memory");
			Tcl_AppendResult(interp, "out of memory", (char *) NULL);
			free_user(uitem);
			Tcl_Free((char *)roles);
			return TCL_ERROR;
		}
		memset(newta, 0, sizeof(ta_item_t));
		newta->idx = idx;
		newta->type = IDX_ROLE;
		rt = insert_ta_item(newta, &(uitem->roles));
		if(rt != 0) {
			free_user(uitem);
			Tcl_AppendResult(interp, "problem inserting role into new user record ");
			Tcl_Free((char *)roles);
			return TCL_ERROR;
		}
		
	}
	Tcl_Free((char*) roles);
	
	if(use_old_login_contexts()) {
		data = (seuser_data_t *)malloc(sizeof(seuser_data_t));
		if(data == NULL) {
			fprintf(stderr, "out of memory");
			Tcl_AppendResult(interp, "out of memory", (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;
		}
		data->flags = SEUSER_FLAG_NONE;
	} 
	
	if(default_login) {
		data->flags |= SEUSER_FLAG_LOGIN_CONTEXT;
		idx = get_role_idx(login_role, policy);
		if(idx < 0) {
			sprintf(tmpbuf, "Login role %s is an invalid role", login_role);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;
		}
		if(!is_role_in_list(idx, uitem->roles)) {
			sprintf(tmpbuf, "Login role %s is an invalid role for %s", login_role, user);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;		
		}
		data->contexts[DC_LOGIN].role = idx;
		idx = get_type_idx(login_type, policy);
		if(idx < 0) {
			sprintf(tmpbuf, "Login type %s is an invalid type", login_type);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;
		}
		if(!does_role_use_type(data->contexts[DC_LOGIN].role, idx, policy)) {
			sprintf(tmpbuf, "Login type %s is an invalid type for login role %s", login_type, login_role);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;		
		}
		data->contexts[DC_LOGIN].type = idx;
	}
	else {
		if(use_old_login_contexts()) {
			data->contexts[DC_LOGIN].role = -1;
			data->contexts[DC_LOGIN].type = -1;
		}
	}
		
	if(default_cron) {
		data->flags |= SEUSER_FLAG_CRON_CONTEXT;
		idx = get_role_idx(cron_role, policy);
		if(idx < 0) {
			sprintf(tmpbuf, "Cron role %s is an invalid role", cron_role);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;
		}
		if(!is_role_in_list(idx, uitem->roles)) {
			sprintf(tmpbuf, "Cron role %s is an invalid role for %s", cron_role, user);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;		
		}
		data->contexts[DC_CRON].role = idx;
		idx = get_type_idx(cron_type, policy);
		if(idx < 0) {
			sprintf(tmpbuf, "Cron type %s is an invalid type", cron_type);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;
		}
		if(!does_role_use_type(data->contexts[DC_CRON].role, idx, policy)) {
			sprintf(tmpbuf, "Cron type %s is an type for cron role %s", cron_type, cron_role);
			Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
			free_user(uitem);
			return TCL_ERROR;		
		}
		data->contexts[DC_CRON].type = idx;
	}
	else {
		if(use_old_login_contexts()) {
			data->contexts[DC_CRON].role = -1;
			data->contexts[DC_CRON].type = -1;
		}
	}
	
	if(use_old_login_contexts()) {
		uitem->data = data;
	} else {
		uitem->data = NULL;
	}
	
	/* new add/change the new record; if successful, don't free user record */
	
	rt = seuser_update_user(uitem, new_user, &db, policy);
	switch (rt) {
	case 0: 
		return TCL_OK;
		break;
	case 1: 
		sprintf(tmpbuf, "Cannot add user %s, user already exists", user);
		break;
	case 2:
		sprintf(tmpbuf, "Cannot change user %s, user does not exist", user);
		break;
	case 3:
		sprintf(tmpbuf, "Bug: Improperly formed user record");
		break;
	default:
		sprintf(tmpbuf, "Error trying to add/update record");
		break;
	}
	Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
	free_user(uitem);
	return TCL_ERROR;
}

/* check for commit (write) access */
int Seuser_CheckCommitAccess(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int rt;
	
	if(argc != 1) {
		Tcl_AppendResult(interp, "ClientData: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}

	assert(user_file != NULL);
	if(use_old_login_contexts()) {
		assert(default_context != NULL);
		assert(cron_context != NULL);
	}
	rt = check_commit_perm();
	switch (rt) {
		case 0:	{
			return TCL_OK;
			break;
		}
		case 1: {
			Tcl_AppendResult(interp,"You do not have commit permission (users file)", (char *) NULL);
			return TCL_ERROR;
			break;
		}
		case 2: {
			Tcl_AppendResult(interp,"You do not have commit permission (default_context file)", (char *) NULL);
			return TCL_ERROR;
			break;
		}
		case 3: {
			Tcl_AppendResult(interp,"You do not have commit permission (cron_cotnext file)", (char *) NULL);
			return TCL_ERROR;
			break;
		}
		default: {
			Tcl_AppendResult(interp,"Unexepcted error checking commit permission", (char *) NULL);
			return TCL_ERROR;
			break;
		}
	}
	return TCL_OK;
}

/* commit changes to the user database */
int Seuser_Commit(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	int rt;
	FILE *fp;
	char tbuf[256];
	
	if(argc != 1) {
		Tcl_AppendResult(interp, "ClientData: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(!db_init) {
		Tcl_AppendResult(interp,"User database not initialized!", (char *) NULL);
		return TCL_ERROR;
	}
	assert(policy != NULL);
	assert(user_file != NULL);
	
	/* Skip asserts if pre-Oct2002 system version */
	if(use_old_login_contexts()) {
		assert(default_context != NULL);
		assert(cron_context != NULL);
	}
	
	/* users file */
	if((fp = fopen(user_file, "w+")) == NULL) {
		sprintf(tbuf, "cannot open user file (%s) for write access!", user_file);
		Tcl_AppendResult(interp, tbuf, (char *) NULL);
		return TCL_ERROR;
	}
	rt = seuser_write_user_file(&db, policy, fp);
	if(rt != 0) {
		fclose(fp);
		Tcl_AppendResult(interp, "Problem writing the user file");
		return TCL_ERROR;
	}	
	fclose(fp);
	
	/* Skip this section if pre-Oct2002 system version. */
	if(use_old_login_contexts()) {
		if((fp = fopen(default_context, "w+")) == NULL) {
			sprintf(tbuf, "cannot open default_context file (%s) for write access!", default_context);
			Tcl_AppendResult(interp, tbuf, (char *) NULL);
			return TCL_ERROR;
		}	
		rt = seuser_write_default_context(DC_LOGIN, &db, policy, fp);
		if(rt != 0) {
			fclose(fp);
			Tcl_AppendResult(interp, "Problem writing the default_context file");
			return TCL_ERROR;
		}
		fclose(fp);	
		
		/* cron contexts */
		if((fp = fopen(cron_context, "w+")) == NULL) {
			sprintf(tbuf, "cannot open cron_context file (%s) for write access!", cron_context);
			Tcl_AppendResult(interp, tbuf, (char *) NULL);
			return TCL_ERROR;
		}	
		rt = seuser_write_default_context(DC_CRON, &db, policy, fp);
		if(rt != 0) {
			fclose(fp);
			Tcl_AppendResult(interp, "Problem writing the cron_context file");
			return TCL_ERROR;
		}
		fclose(fp);	
	}
	
	return TCL_OK;
}

/* get roles for a user *
 * argv[1] user name
 */
int Seuser_UserRoles(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	char tmpbuf[256], *name;
	user_item_t *user;
	ta_item_t *ptr;
	int rt;
	
	Tcl_ResetResult(interp);
	if(argc != 2) {
		Tcl_AppendResult(interp, "Seuser_UserRoles: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(!db_init) {
		Tcl_AppendResult(interp,"User database not initialized!", (char *) NULL);
		return TCL_ERROR;
	}
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "User name string too large", (char *) NULL);
		return TCL_ERROR;
	}
	rt = seuser_get_user_by_name(argv[1], &user, &db);
	if(rt != 0) {
		sprintf(tmpbuf, "Invalid user name (%s)", argv[1]);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;		
	}
	
	for(ptr = user->roles; ptr != NULL; ptr = ptr->next) {
		assert(ptr->type == IDX_ROLE);
		rt = get_role_name(ptr->idx, &name, policy);
		if(rt != 0) {
			Tcl_ResetResult(interp);
			Tcl_AppendResult(interp, "error getting role name", (char *) NULL);			
			return TCL_ERROR;
		}
		Tcl_AppendElement(interp, name);
		free(name);
	}
		
	return TCL_OK;
}

/* get default context for a user *
 * argv[1]	user name
 * argv[2]	which (0 = login, 1  = cron)
 * 
 * returns two elements: role name and type name
 */
int Seuser_UserContext(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	char tmpbuf[256], *role, *type;
	int rt, which;
	
	Tcl_ResetResult(interp);
	if(argc != 3) {
		Tcl_AppendResult(interp, "Seuser_UserContext: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(!db_init) {
		Tcl_AppendResult(interp,"User database not initialized!", (char *) NULL);
		return TCL_ERROR;
	}
	rt= Tcl_GetInt(interp, argv[2], &which);
	if(rt == TCL_ERROR)
		return TCL_ERROR;
	if(which < DC_LOGIN || which > DC_CRON) {
		sprintf(tmpbuf, "Which argument has an invalid value (%d)", which);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;
	}
		
	if(!is_valid_str_sz(argv[1])) {
		Tcl_AppendResult(interp, "User name string too large", (char *) NULL);
		return TCL_ERROR;
	}
	rt = seuser_get_user_context(argv[1], which, &role, &type, &db, policy);
	if(rt == -1) {
		sprintf(tmpbuf, "Invalid user name (%s)", argv[1]);
		Tcl_AppendResult(interp, tmpbuf, (char *) NULL);
		return TCL_ERROR;
	}
	else if(rt == -2 || rt == -3) {
		Tcl_AppendElement(interp, "<invalid>");
		Tcl_AppendElement(interp, "<invalid>");
	}
	else if(rt == -4) {
		Tcl_AppendResult(interp, "Unexpected error", (char *) NULL);
		return TCL_ERROR;
	}
	else if(rt == 1) {
		Tcl_AppendElement(interp, "<none>");
		Tcl_AppendElement(interp, "<none>");
	}
	else {
		assert(rt == 0);
		Tcl_AppendElement(interp, role);
		Tcl_AppendElement(interp, type);
		free(role);
		free(type);	
	}
	return TCL_OK;
}


/* get the list of system-defined users */
/* argv[1] 	Indicate if user type is wanted (opt bool)
 *
 * if argv[1] != 0, then the resturn result will be list where there
 * are two elements for each user (username type).  The type will be a string
 * with one of the SYSUSER_TYPE values from the header file.
 *
 * NOTE: if user type info is wanted then user_u and system_u will also be
 * included as system users if they exist in the policy.
 */
int Seuser_GetSysUsers(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
   	struct passwd *line = (struct passwd *)malloc(sizeof(struct passwd));
   	bool_t include_type;
   	char *type;

	Tcl_ResetResult(interp);
	if(argc > 2) {
		Tcl_AppendResult(interp, "Seuser_GetSysUsers: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	if(argc == 2) {
		if(!is_valid_str_sz(argv[1])) {
			Tcl_AppendResult(interp, "Boolean string too large", (char *) NULL);
			return TCL_ERROR;
		}
		include_type = getbool(argv[1]);
	}
	else 
		include_type = FALSE;
		
	while ( (line = getpwent()) != NULL) {
     		Tcl_AppendElement(interp, line->pw_name);
     		if(include_type) {
     			type = determine_user_type(line->pw_name);
     			Tcl_AppendElement(interp, type);
     			free(type);
     		}
   	}
   	free(line);
   	endpwent();
   	
   	/* now add the special users if they exist and type is requested */
	if(include_type) {
		user_item_t *uitem;
		int rt;
		rt = seuser_get_user_by_name(SPECIAL_USER_SYSTEM, &uitem, &db);
		if(rt == 0) {
			Tcl_AppendElement(interp, SPECIAL_USER_SYSTEM);
			Tcl_AppendElement(interp, SYSUSER_SPECIAL_STR);
		}
		rt = seuser_get_user_by_name(SPECUAL_USER_USER, &uitem, &db);
		if(rt == 0) {
			Tcl_AppendElement(interp, SPECUAL_USER_USER);
			Tcl_AppendElement(interp, SYSUSER_SPECIAL_STR);
		}
	}
	return TCL_OK;
}


/* get the list of system-defined groups */
int Seuser_GetSysGroups(ClientData clientData, Tcl_Interp * interp, int argc, char *argv[])
{
   	struct group *line = (struct group *)malloc(sizeof(struct group));

	Tcl_ResetResult(interp);
	if(argc != 1) {
		Tcl_AppendResult(interp, "Seuser_GetSysGroups: wrong # of args", (char *) NULL);
		return TCL_ERROR;
	}
	while ( (line = getgrent()) != NULL) {
     		Tcl_AppendElement(interp, line->gr_name);
   	}
   	free(line);
   	endgrent();

	return TCL_OK;
}


/* Package initialization */
int Seuser_Init(Tcl_Interp *interp) 
{
	Tcl_CreateCommand(interp, "seuser_InitUserdb", Seuser_InitUserdb, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_GetSysUsers", Seuser_GetSysUsers, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_GetSysGroups", Seuser_GetSysGroups, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_GetSeUserNames", Seuser_GetSeUserNames, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_UserRoles", Seuser_UserRoles, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_UserContext", Seuser_UserContext, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_Commit", Seuser_Commit, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_RemoveUser", Seuser_RemoveUser, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_CloseDatabase", Seuser_CloseDatabase, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_EditUser", Seuser_EditUser, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_RemakePolicyConf", Seuser_RemakePolicyConf, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_ReinstallPolicy", Seuser_ReinstallPolicy, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_IsUserValid", Seuser_IsUserValid, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_GetVersion", Seuser_GetVersion, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "euser_Get_Login_Version", Seuser_Get_Login_Version, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_GetTmpMakeFileName", Seuser_GetTmpMakeFileName, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_Use_Old_Login_Contexts", Seuser_Use_Old_Login_Contexts, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_Exit", Seuser_Exit, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateCommand(interp, "seuser_CheckCommitAccess", Seuser_CheckCommitAccess, (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_PkgProvide(interp, "seuser", SEUSER_VERSION_STRING);

	return TCL_OK;
}


