#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <proc_secure.h>
#include <ss.h>
#include <syslog.h>
#include <string.h>
#include <fs_secure.h>
#include <errno.h>

#ifndef PASSWD
#define PASSWD "/usr/bin/passwd"
#endif

#define SHADOW_CONTEXT "/etc/security/shadow_context"
#define PASSWD_CONTEXT "/etc/security/passwd_context"

security_id_t sid_from_file(const char * const file)
{
  security_id_t sid;
  char buf[1024];
  FILE *fp = fopen(file, "r");
  if(!fp || !fgets(buf, sizeof(buf), fp))
  {
    fprintf(stderr, "Can't read %s\n", file);
    exit(1);
  }
  strtok(buf, "\n");
  if(security_context_to_sid(buf, strlen(buf) + 1, &sid))
  {
    fprintf(stderr, "Can't get file context from %s.\n", file);
    exit(1);
  }
  fclose(fp);
  return sid;
}

#define SZ 255

pid_t child;

void sig_handler(int sig, siginfo_t *siginf, void *unused)
{
  kill(child, siginf->si_signo);
}


int main(int argc, char **argv) 
{
  char *p;
  security_id_t shadow_sid, passwd_sid;
#ifndef ADMIN_USE
  security_id_t sid;
  struct passwd *pw;
  security_context_t context;
  uid_t ruid;
  unsigned int len;
#endif
  int ret;
  int status;
  char **passwd_argv;
  int i;
  struct sigaction sa;

  if (argc < 0 || argc > 100) {
	  fprintf(stderr, "%s:  Too many arguments.\n", argv[0]);
	  exit(1);
  }

  passwd_argv = malloc( (argc + 1) * sizeof(char *) );
  if(!passwd_argv)
  {
    fprintf(stderr, "Can't malloc.\n");
    exit(1);
  }
  for(i = 1; i <= argc; i++)
    passwd_argv[i] = argv[i];
  passwd_argv[0] = PASSWD;

#ifndef ADMIN_USE
  if (argc != 1) {
    fprintf(stderr, "%s:  wrapper does not accept arguments\n",
	    argv[0]);
    exit(1);
  }

  ruid = getuid();
  if (ruid < 0) {
    perror(argv[0]);
    exit(1);
  }

  pw = getpwuid(ruid);
  if (!pw) {
    fprintf(stderr, "%s:  no passwd entry for uid %d\n",
	    argv[0], ruid);
    exit(1);
  }

  sid = getsecsid();
  if (sid < 0) {
    perror(argv[0]);
    exit(1);
  }

  context = malloc(SZ);
  if (!context) {
    perror(argv[0]);
    exit(1);
  }

  len = SZ;
  ret = security_sid_to_context(sid, context, &len);
  if (ret) {
    if (len > SZ) {
      context = realloc(context, len);
      if (!context) {
	perror(argv[0]);
	exit(1);
      }
      ret = security_sid_to_context(sid, context, &len);
    }
    if (ret) {
      perror(argv[0]);
      exit(1);
    }
  }

  for (p = context; *p && *p != ':'; p++)
    ;
  *p = 0;
  if (strcmp(pw->pw_name, context) != 0) {
    fprintf(stderr, "%s:  Unix identity %s does not match Flask identity %s\n",
	    argv[0], pw->pw_name, context);
    exit(1);
  }

#endif

  shadow_sid = sid_from_file(SHADOW_CONTEXT);
  passwd_sid = sid_from_file(PASSWD_CONTEXT);
  
  for (p = PASSWD + strlen(PASSWD); *p != '/'; p--)
    ;
  if(*p == '/')
    p++;

  openlog(p, LOG_CONS, LOG_AUTH);

  sa.sa_sigaction = &sig_handler;
  sa.sa_flags = SA_SIGINFO;
  if(sigaction(SIGHUP, &sa, NULL)
    || sigaction(SIGINT, &sa, NULL)
    || sigaction(SIGQUIT, &sa, NULL)
    || sigaction(SIGPIPE, &sa, NULL)
    || sigaction(SIGTERM, &sa, NULL))
  {
    fprintf(stderr, "Can't handle signals.");
    return 1;
  }

  child = fork();
  switch(child)
  {
  case 0:  /* child */
    execv(PASSWD, passwd_argv);
    perror(argv[0]);
    exit(1);
  case -1: /* error */
    perror(argv[0]);
    exit(1);
  }
  while( (ret = wait(&status)) == -1 && errno == EINTR)
    ;
  if(ret != child)
  {
    fprintf(stderr, "Error %d waiting for child.\n", errno);
    ret = 1;
  }
  else
  {
    if(!WIFEXITED(status) || WEXITSTATUS(status) != 0)
      fprintf(stderr, "Child returned an error.\n");
    else
      ret = 0;
  }

  if(chsid("/etc/passwd", passwd_sid))
  {
    fprintf(stderr, "Can't update the sid for /etc/passwd\n");
    syslog(LOG_ERR, "Can't update the sid for /etc/passwd");
    ret = 1;
  }
  if(chsid("/etc/shadow", shadow_sid))
  {
    syslog(LOG_ALERT, "Can't update the sid for /etc/shadow");
    ret = 1;
  }

  /*
   * by default, useradd creates a new group for the new user.  Make
   * certain to update labels.  Label group files with the same contexts
   * as passwd files.
   */
  if(chsid("/etc/group", passwd_sid))
  {
    fprintf(stderr, "Can't update the sid for /etc/group\n");
    syslog(LOG_ERR, "Can't update the sid for /etc/group");
    ret = 1;
  }
  if(chsid("/etc/gshadow", shadow_sid))
  {
    syslog(LOG_ALERT, "Can't update the sid for /etc/gshadow");
    ret = 1;
  }
  return ret;
}
